/*
 * DataTypes.h
 *
 *  Created on: 15.06.2010
 *      Author: troth
 */

#ifndef DATATYPES_H_
#define DATATYPES_H_

#include <Utils/Common.h>


#include <stdint.h>
#include <string>
#include <iostream>
#include <typeinfo>
#include <string>
#include <map>
//#include <boost/shared_array.hpp>

#include <Data/Types_Generators.h>
#include <Data/DataTypes_primitive.h>
#include <Data/DataTypes_dependent.h>


#undef	CLASSNAME
#define	CLASSNAME	DataTypes

namespace grip {

	/**
	 * define types, should always be available as the type itself and the according shared_ptr type with the exception of void.
	 * simple pointer types are not explicitly defined. dt_void_t is also not defined.
	 */

	___GENERATE_ENUMLIST;

	using namespace std;

	/**
	 * @class DataTypes
	 * @brief this class provides all data type related information and operations, including type ids, names and conversion helpers
	 */
	namespace CLASSNAME {

		/** @brief type names are stored here for readability purposes */
		static map<uint32_t, string> sm_typeNames;

		/**
		 * @brief get the enum typeId for the passed type
		 * @return typeId for given template parameter, -1 if type is not supported
		 */
		template<class FTYPE> INLINE int32_t getIdByType() {
			DEBUG_MSG_FUNCTIONCALL("getIdByType()");

			___GENERATE_GETIDBYTYPE_BRANCHING;
		}

		/** @brief get the type name for the given typeId */
		INLINE string getNameById(const int32_t p_typeId) {
			DEBUG_MSG_FUNCTIONCALL("getNameById(" << p_typeId << ")");
			
			/* assign names */
			if (sm_typeNames.empty()) {
				___GENERATE_NAME_ASSIGNMENT;
				
			}
			
			return sm_typeNames[p_typeId];
		}
		

		/** @brief get the type name for the given template parameter */
		template<class FTYPE> INLINE string getNameByType() {
			DEBUG_MSG_FUNCTIONCALL("getNameByType()");

			return getNameById(getIdByType<FTYPE>());
		}


		

		/** @brief get typeId for the given type name */
		INLINE int32_t getIdByName(const string& p_name) {
			DEBUG_MSG_FUNCTIONCALL("getIdByName(" << p_name << ")");

			for (int32_t i = ___DT_TYPEID_MIN; i <= ___DT_TYPEID_MAX; ++i) {
				if (p_name == DataTypes::getNameById(i)) {
					return i;
				}
			}
			return -1;
		}

	}

}

#endif /* DATATYPES_H_ */
