/*
 * UniversalDataElement.h
 *
 *  Created on: 15.06.2010
 *      Author: troth
 */

#ifndef UNIVERSALDATAELEMENT_H_
#define UNIVERSALDATAELEMENT_H_

#include <Data/DataTypes.h>
#include <Utils/Common.h>

#include <iostream>

namespace grip {

	#undef	CLASSNAME
	#define	CLASSNAME	UniversalDataElement

	using namespace std;

	/**
	 * @class UniversalDataElement
	 *
	 * @brief This represents an element of the UniversalContainer class
	 *
	 * Each element stores the data, as well as a typeId and (for pointers only) the (array) size.
	 * CTYPE_INNER is the pointer type for the CUDA device data.
	 */
	template<class CTYPE> class CLASSNAME {
	private:
		CTYPE __m_data;
		uint32_t __m_typeId;
		uint32_t __m_size;

	public:

		INLINE CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("UniversalDataElement()");

			__m_typeId = DataTypes::getIdByType<CTYPE>();
		}


		INLINE CLASSNAME(CTYPE p_data, uint32_t p_size) : __m_data(p_data), __m_size(p_size) {
			DEBUG_MSG_FUNCTIONCALL("UniversalDataElement(" << p_data << ", " << ", " << p_size << ")");

			__m_typeId = DataTypes::getIdByType<CTYPE>();
		}


		INLINE ~CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("~UniversalDataElement()");

		}

		/** @brief get the associated data */
		INLINE CTYPE getData() {
			DEBUG_MSG_FUNCTIONCALL("getData()");

			return __m_data;
		}


		/** @brief returns the data, size included */
		INLINE CTYPE getData(uint32_t& p_size) {
			DEBUG_MSG_FUNCTIONCALL("getData(" << p_size << ")");

			DEBUG_MSG_RUNTIMEINFO("getData, typeID is: " << __m_typeId << " (" << DataTypes::getNameById(__m_typeId) << ")");

			p_size = __m_size;
			return __m_data;
		}


		/** @brief creates NULL element
		 *
		 * This gets called for basic (numeric) types. shared_ptr types are specialized below.
		 */
		INLINE void makeNullElement() {
			DEBUG_MSG_FUNCTIONCALL("makeNullElement()");

			set(0, 0, 0);
		}


		/** @brief get the type id */
		INLINE int32_t getTypeId() {
			DEBUG_MSG_FUNCTIONCALL("getTypeId()");

			return __m_typeId;
		}


		/** @brief get size */
		INLINE uint32_t getSize() {
			DEBUG_MSG_FUNCTIONCALL("getSize()");

			return __m_size;
		}


		/** @brief set data, size and type id at once */
		INLINE void set(CTYPE p_data, uint32_t p_size, uint32_t p_typeId) {
			DEBUG_MSG_FUNCTIONCALL("set(" << p_data << ", " << p_size << ")");

			setData(p_data);
			setSize(p_size);
			setTypeId(p_typeId);
		}


		/** @brief set data */
		INLINE void setData(CTYPE p_data) {
			DEBUG_MSG_FUNCTIONCALL("setData(" << p_data << ")");

			__m_data = p_data;
		}


		/** @brief set size */
		INLINE void setSize(uint32_t p_size) {
			DEBUG_MSG_FUNCTIONCALL("setSize(" << p_size << ")");

			__m_size = p_size;
		}

		INLINE void setTypeId(int32_t p_typeId) {
			DEBUG_MSG_FUNCTIONCALL("setTypeId(" << p_typeId << ")");

			__m_typeId = p_typeId;
		}


		/** @brief check if this element is the NULL element
		 *
		 *  TODO This has to be specialized for the supported types
		 */
		INLINE bool isNullElement() {
			DEBUG_MSG_FUNCTIONCALL("isNullElement()");

			return (__m_data == 0 && __m_size == 0 && __m_typeId == 0);
		}

	};

	// set methods for shared_ptr types


	___GENERATE_MAKENULLELEMENT;
	___GENERATE_ISNULLELEMENT;

}


#endif /* UNIVERSALCONTAINERELEMENT_H_ */
