/*
 * GraphVisualizer.h
 *
 *  Created on: Dec 20, 2010
 *      Author: troth22s
 */

/**
 * Latest changes:
 * 2013-12-16: 
 *  - NOTE: Graph Visualization is currently disabled due to incompatibility issues with recent graphviz libs
 * 2011-11-02:
 * 	- Changed graph drawing library from SDL to SFML (easier and more object oriented ;))
 */


#ifndef GRAPHVISUALIZER_H_
#define GRAPHVISUALIZER_H_

#include <graphviz/gvc.h>

#include <Utils/Common.h>

#include <SFML/Graphics.hpp>

namespace grip {
	
	#undef CLASSNAME
	#define CLASSNAME GraphVisualizer
	
	using namespace std;
	
	enum enum_visualizer_modes {
		ENUM_GRAPHVISUALIZER_UNORDERED, ENUM_GRAPHVISUALIZER_ORDERED, ENUM_GRAPHVISUALIZER_NONE
	};
	
	/**
	 * @class GraphVisualizer
	 * @brief This class uses Graphviz and SFML to generate a visualization based on the GraphWrapper
	 */
	class CLASSNAME {
	private:
		GraphWrapper* __m_graphWrapper;
		string __m_imageName;
		
		sf::RenderWindow __m_graphWindow;
		
	public:
		
		/** @brief standard constructor using /tmp/graph.png as the graphviz render target. GraphWrapper still has to be set. */
		CLASSNAME() {
			/*	DEBUG_MSG_FUNCTIONCALL("GraphVisualizer()");
			 * 
			 *		__m_imageName = "/tmp/graph.png";
			 */
		}
		
		
		/** @brief constructor for setting graphwrapper and an arbitrary graphviz render target path */
		CLASSNAME(GraphWrapper* p_gw, const string& p_imageName) {
			/*		DEBUG_MSG_FUNCTIONCALL("GraphVisualizer(" << p_gw << ", " << p_imageName << ")");
			 * 
			 *		__m_graphWrapper = p_gw;
			 *		__m_imageName = p_imageName;*/
		}
		
		
		/** @brief standard destructor */
		~CLASSNAME() {
			//DEBUG_MSG_FUNCTIONCALL("~GraphVisualizer()");
		}
		
		
		/** @brief set the graphwrapper which should be used */
		void setWrapper(GraphWrapper* p_gw) {
			/*DEBUG_MSG_FUNCTIONCALL("setWrapper(" << p_gw << ")");
			 * 
			 *		__m_graphWrapper = p_gw;*/
		}
		
		
		/** @brief generate graph image with the respective visualization mode for the ordered list of nodes p_orderedNames */
		void generateImage(enum_visualizer_modes p_visMode, vector<string>& p_orderedNames) {
			// 		DEBUG_MSG_FUNCTIONCALL("generateImage()");
			// 
			// 		if (p_visMode == ENUM_GRAPHVISUALIZER_NONE) {
			// 			return;
			// 		}
			// 
			// 		Agraph_t *graph;
			// 		map<string, Agnode_t*> nodes;
			// 		list<Agedge_t*> edges;
			// 		GVC_t *gvc;
			// 
			// 		/* set up graphviz context */
			// 		gvc = gvContext();
			// 
			// 		string outputFlag = "-o" + __m_imageName;
			// 
			// 		/* synthetic command line args */
			// 		char* args[] = { strdup("dot"), strdup("-Gdpi=96"), strdup("-Nfontsize=10"), strdup("-Nstyle=filled"), strdup("-Nfillcolor=#CCCCFF"), strdup("-Estyle=solid"), strdup("-Efontsize=10"), strdup("-Elabel="), strdup("-Earrowhead=open"), strdup("-Nshape=box"), strdup("-Grankdir=TB"), strdup("-Tpng"), strdup(const_cast<char*> (outputFlag.c_str())) };
			// 		gvParseArgs(gvc, sizeof(args) / sizeof(char*), args);
			// 
			// 		std::cout << "parsed" << std::endl;
			// 
			// 		/* create simple digraph */
			// 		graph = agopen(strdup("g"), AGDIGRAPH);
			// 
			// 		if (p_visMode == ENUM_GRAPHVISUALIZER_UNORDERED) {
			// 
			// 			list<string> nodeNames = __m_graphWrapper->getAllNodeNames();
			// 
			// 			/* first add all nodes */
			// 			for (list<string>::iterator it = nodeNames.begin(); it != nodeNames.end(); it++) {
			// 				string nodeType;
			// 				__m_graphWrapper->getNodeTypeByName(*it, nodeType);
			// 				string nodeText = nodeType + "\n(" + *it + ")";
			// 
			// 				char* name = const_cast<char*> (nodeText.c_str());
			// 
			// 				nodes[*it] = agnode(graph, name);
			// 
			// 				if (nodeType == "node_loop_start") {
			// 					agset(nodes[*it], strdup("shape"), strdup("ellipse"));
			// 					agset(nodes[*it], strdup("fillcolor"), strdup("#99FF99"));
			// 				}
			// 
			// 				if (nodeType == "node_loop_end") {
			// 					agset(nodes[*it], strdup("shape"), strdup("ellipse"));
			// 					agset(nodes[*it], strdup("fillcolor"), strdup("#FF9999"));
			// 				}
			// 
			// 			}
			// 
			// 			/* now add all edges */
			// 			for (list<string>::iterator it = nodeNames.begin(); it != nodeNames.end(); it++) {
			// 				list<string> successorNames = __m_graphWrapper->getSuccessorNames(*it);
			// 				for (list<string>::iterator it2 = successorNames.begin(); it2 != successorNames.end(); it2++) {
			// 					edges.push_back(agedge(graph, nodes[*it], nodes[*it2]));
			// 				}
			// 
			// 				for the loop end node, another edge has to be added to the loop start node :)
			// 				string nodeType;
			// 				__m_graphWrapper->getNodeTypeByName(*it, nodeType);
			// 
			// 				if (nodeType == "node_loop_end") {
			// 					for (list<string>::iterator it2 = nodeNames.begin(); it2 != nodeNames.end(); it2++) {
			// 						string otherType;
			// 						__m_graphWrapper->getNodeTypeByName(*it2, otherType);
			// 
			// 						start node found and it belongs to the same loop? draw an edge :)
			// 						if (otherType == "node_loop_start" && ((*it).substr(0, (*it).length() - string("_end").length()) == ((*it2).substr(0, (*it2).length() - string("_start").length())))) {
			// 							Agedge_t* theEdge = agedge(graph, nodes[*it], nodes[*it2]);
			// 							agset(theEdge, strdup("style"), strdup("dotted"));
			// 							string iterations;
			// 							__m_graphWrapper->getDataFromNode(*it2, "iterations", iterations);
			// 							agset(theEdge, strdup("label"), const_cast<char*> ((string("n = ") + iterations).c_str()));
			// 							edges.push_back(theEdge);
			// 						}
			// 					}
			// 				}
			// 
			// 			}
			// 		} else if (p_visMode == ENUM_GRAPHVISUALIZER_ORDERED) {
			// 			/* first add all nodes */
			// 			for (vector<string>::iterator it = p_orderedNames.begin(); it != p_orderedNames.end(); it++) {
			// 				string nodeType;
			// 				__m_graphWrapper->getNodeTypeByName(*it, nodeType);
			// 				string nodeText = nodeType + "\n(" + *it + ")";
			// 
			// 				char* name = const_cast<char*> (nodeText.c_str());
			// 
			// 				nodes[*it] = agnode(graph, name);
			// 
			// 				if (nodeType == "node_loop_start") {
			// 					agset(nodes[*it], strdup("shape"), strdup("ellipse"));
			// 					agset(nodes[*it], strdup("fillcolor"), strdup("#99FF99"));
			// 				}
			// 
			// 				if (nodeType == "node_loop_end") {
			// 					agset(nodes[*it], strdup("shape"), strdup("ellipse"));
			// 					agset(nodes[*it], strdup("fillcolor"), strdup("#FF9999"));
			// 				}
			// 			}
			// 
			// 			for (vector<string>::iterator it = p_orderedNames.begin(); it != p_orderedNames.end(); it++) {
			// 				vector<string>::iterator it2 = it;
			// 				it2++;
			// 				if (it2 != p_orderedNames.end()) {
			// 					edges.push_back(agedge(graph, nodes[*it], nodes[*it2]));
			// 				}
			// 
			// 				string nodeType;
			// 				__m_graphWrapper->getNodeTypeByName(*it, nodeType);
			// 
			// 				if (nodeType == "node_loop_end") {
			// 					for (vector<string>::iterator it2 = p_orderedNames.begin(); it2 != p_orderedNames.end(); it2++) {
			// 						string otherType;
			// 						__m_graphWrapper->getNodeTypeByName(*it2, otherType);
			// 						start node found and it belongs to the same loop? draw an edge :)
			// 						if (otherType == "node_loop_start" && ((*it).substr(0, (*it).length() - string("_end").length()) == ((*it2).substr(0, (*it2).length() - string("_start").length())))) {
			// 							Agedge_t* theEdge = agedge(graph, nodes[*it], nodes[*it2]);
			// 							agset(theEdge, strdup("style"), strdup("dotted"));
			// 							string iterations;
			// 							__m_graphWrapper->getDataFromNode(*it2, "iterations", iterations);
			// 							agset(theEdge, strdup("label"), const_cast<char*> ((string("n = ") + iterations).c_str()));
			// 							edges.push_back(theEdge);
			// 						}
			// 					}
			// 				}
			// 			}
			// 		}
			// 
			// 
			// 		/* compute layout */
			// 		gvLayoutJobs(gvc, graph);
			// 
			// 
			// 		/* Write graph */
			// 		gvRenderJobs(gvc, graph);
			// 
			// 
			// 		/* Free Layout data */
			// 		gvFreeLayout(gvc, graph);
			// 
			// 
			// 		/* Free graph Structures */
			// 		agclose(graph);
			// 
			// 
			// 		/* close output file */
			// 		gvFreeContext(gvc);
			
		}
		
		
		/** @brief draw the generated graph */
		void draw() {
			// 		DEBUG_MSG_FUNCTIONCALL("draw()");
			// 
			// 		sf::Image theImage;
			// 		if (!theImage.LoadFromFile(__m_imageName)) {
			// 			DEBUG_MSG_WARNING("Could not load graph image!");
			// 		}
			// 		theImage.SetSmooth(false);
			// 
			// 		// open window if it's not there yet
			// 		__m_graphWindow.Create(sf::VideoMode(theImage.GetWidth(), theImage.GetHeight(), 32), "grip::GraphVisualizer", sf::Style::Titlebar | sf::Style::Close);
			// 
			// 
			// 		sf::Sprite theSprite;
			// 		theSprite.SetImage(theImage);
			// 
			// 
			// 		__m_graphWindow.Clear();
			// 		__m_graphWindow.Draw(theSprite);
			// 		__m_graphWindow.Display();
			
		}
		
		
		/** @brief call the methods for generating and drawing consecutively */
		void generateAndDraw(enum_visualizer_modes p_visMode, vector<string>& p_orderedNames) {
			// 		DEBUG_MSG_FUNCTIONCALL("generateAndDraw(" << p_visMode << ", ...)");
			// 
			// 		if (p_visMode == ENUM_GRAPHVISUALIZER_NONE) {
			// 			return;
			// 		}
			// 
			// 		generateImage(p_visMode, p_orderedNames);
			// 		draw();
		}
		
	};
	
}

#endif /* GRAPHVISUALIZER_H_ */
