/*
 * ParameterizerGui.h
 *
 *  Created on: Jan 7, 2011
 *      Author: troth22s
 */

#ifndef PARAMETERIZERGUI_H_
#define PARAMETERIZERGUI_H_

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Slider.H>
#include <FL/Fl_Tabs.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Scrollbar.H>
#include <FL/Fl_Pack.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Widget.H>

//#include <OpenThreads/Thread>
//#include <OpenThreads/Mutex>

#include <SFML/System.hpp>



#include <Data/UniversalContainer.h>

#include <Graph/GraphWrapper.h>



#include <Data/DataTypes.h>

#include <iostream>

namespace grip {
	
	#define BOX_SPACING 5
	#define SPACING 4
	
	
	
	void windowEscapeCallback(Fl_Widget*, void*) {
		if (Fl::event()==FL_SHORTCUT && Fl::event_key()==FL_Escape ) { //avoid escape from freezing the sliders
			return; 
		}
		if (Fl::event()==FL_CLOSE) { // ignore window close
			return;	
		}
	}
	
	
	#undef CLASSNAME
	#define CLASSNAME CheckboxCallbackClass
	
	class CLASSNAME {
	private:
		static UniversalContainer* __m_uniCon;
		static sf::Mutex* __m_parameterMutex;
		
	public:
		
		/** @brief set the UniversalContainer to interact with */
		static void setUniversalContainer(UniversalContainer* p_uniCon) {
			DEBUG_MSG_FUNCTIONCALL("setUniversalContainer(" << p_uniCon << ")");
			
			__m_uniCon = p_uniCon;
		}
		
		/** @brief set the mutex which has to be locked */
		static void setParameterMutex(sf::Mutex* p_parameterMutex) {
			__m_parameterMutex = p_parameterMutex;
		}
		
		static void checkboxCallback(Fl_Widget *w, void *data) {
			DEBUG_MSG_FUNCTIONCALL("checkboxCallback(...)");
			
			__m_parameterMutex->Lock();
			
			string* key = (string*)data;
			
			//std::cout << "key: " << *key << std::endl;
			
			bool value = (((Fl_Check_Button*) w)->value());
			
			__m_uniCon->setNodeActivation(*key, value);
			
			__m_parameterMutex->Unlock();
			
		}
		
		
	};
	
	UniversalContainer* CLASSNAME::__m_uniCon = 0;
	sf::Mutex* CLASSNAME::__m_parameterMutex = 0;
	
	
	
	
	#undef CLASSNAME
	#define CLASSNAME SliderCallbackClass
	
	/**
	 * @class SliderCallbackClass
	 * @brief provides static methods for setting the appropriate details
	 * and provide a callback method called when a slider is moved.
	 */
	class CLASSNAME {
	private:
		static UniversalContainer* __m_uniCon;
		static sf::Mutex* __m_parameterMutex;
		
	public:
		/** @brief set the UniversalContainer to interact with */
		static void setUniversalContainer(UniversalContainer* p_uniCon) {
			DEBUG_MSG_FUNCTIONCALL("setUniversalContainer(" << p_uniCon << ")");
			
			__m_uniCon = p_uniCon;
		}
		
		/** @brief set the mutex which has to be locked */
		static void setParameterMutex(sf::Mutex* p_parameterMutex) {
			__m_parameterMutex = p_parameterMutex;
		}
		
		
		/** @brief this is the callback method, which sets the according data in the UniversalContainer */
		template<class FTYPE> static void sliderCallback(Fl_Widget *w, void *data) {
			DEBUG_MSG_FUNCTIONCALL("slider_callback(" << w << ", " << data << ")");
			
			__m_parameterMutex->Lock();
			
			string* key = (string*)data;
			
			__m_uniCon->setObject<FTYPE> (*key, (FTYPE) (((Fl_Value_Slider*) w)->value()), 1);
			
			__m_uniCon->setParameterChanged(*key);
			
			__m_parameterMutex->Unlock();
		}
		
	};
	
	UniversalContainer* CLASSNAME::__m_uniCon = 0;
	sf::Mutex* CLASSNAME::__m_parameterMutex = 0;
	
	
	#undef CLASSNAME
	#define CLASSNAME ParameterizerGuiThread
	
	/**
	 * @class ParameterizerGUIThread
	 * @brief this implements the dynamic parameterizer GUI, consisting of several sliders for parameter modification
	 */
	class CLASSNAME: public sf::Thread {
	private:
		GraphWrapper* __m_graphWrapper;
		UniversalContainer* __m_uniCon;
		Fl_Window* __m_window;
		Fl_Tabs* __m_tabs;
		
		int __m_slIndex;
		
		int __m_slDist;
		int __m_slYStart;
		int __m_slHeight;
		
		map<string, bool> __m_alreadyGuified;
		
		std::map<std::string, Fl_Value_Slider*> __m_sliderMap;
		
		map<string, bool> __m_nodeDeactivationInfo;
		
		bool __m_isReady;
		
		sf::Mutex* __m_parameterMutex;
		
		
		
		/** @brief easy slider creation method with all needed parameters. sliders are automatically positioned! */
		template<class FTYPE> Fl_Value_Slider* createSlider(double p_min, double p_max, double p_step, double p_initval, const string& p_parameter, const string& p_parameterIdentifier, unsigned int p_sliderIndex) {
			DEBUG_MSG_FUNCTIONCALL("createSlider(" << p_min << ", " << p_max << ", " << p_step << " ," << p_initval << ")");
			
			char* labelString = strdup(p_parameter.c_str());
			
			Fl_Value_Slider* slider = new Fl_Value_Slider(0, __m_slYStart + p_sliderIndex * (__m_slHeight + __m_slDist), 0, __m_slHeight, labelString);
			slider->type(FL_HORIZONTAL);
			slider->bounds(p_min, p_max);
			slider->step(p_step);
			slider->value(p_initval);
			slider->align(FL_ALIGN_LEFT);
			//slider->labelsize(36);
			
			string* parameterIdentifier = new string(p_parameterIdentifier);
			
			slider->callback(SliderCallbackClass::sliderCallback<FTYPE>, parameterIdentifier);
			
			//int labelWidth = 0;
			//int labelHeight = 0;
			//	slider->measure_label(labelWidth, labelHeight);
			
			//std::cout << "width, height: " << labelWidth << ", " << labelHeight << std::endl;
			
			
			//	unsigned int newWindowHeight = fmaxf(__m_window->h(), __m_slYStart + p_sliderIndex * (__m_slHeight + __m_slDist));
			//	__m_window->size(450, newWindowHeight);
			
			__m_sliderMap[*parameterIdentifier] = slider;
			
			return slider;
		}
		
		/** @brief used for static parameters */
		void createTextOutput(const string& p_label, const string& p_content, unsigned int p_textIndex) {
			Fl_Output* output = new Fl_Output(180, __m_slYStart + p_textIndex * (__m_slHeight + __m_slDist), 260, __m_slHeight, strdup(p_label.c_str()));
			output->value(strdup(p_content.c_str()));
			output->align(FL_ALIGN_LEFT);
			


			index = 1;


		}
		
		
		void createDeactivationButton(string& p_nodeName, unsigned int p_index, bool p_defaultState, float p_clickable = true) {
			//Fl_Check_Button* deactivationButton = new Fl_Check_Button(180, __m_slYStart + index * (__m_slHeight + __m_slDist), 260, __m_slHeight, "Active");
			Fl_Light_Button* deactivationButton = new Fl_Light_Button(180, __m_slYStart + p_index * (__m_slHeight + __m_slDist), 260, __m_slHeight, "Active");
			deactivationButton->value(p_defaultState);
			deactivationButton->down_color(FL_GREEN);
			deactivationButton->down_box(FL_OVAL_BOX);
			
			string* nodeName = new string(p_nodeName);
			
			deactivationButton->callback(CheckboxCallbackClass::checkboxCallback, nodeName);
			__m_uniCon->setNodeActivation(p_nodeName, p_defaultState);
			
			if (!p_clickable) {
				
				deactivationButton->deactivate();
				
			}
		}
		
		
		/** @brief creates the whole gui, which consists of tabbed sliders/static text areas and (de)activation buttons for each node */
		void createSliders() {
			DEBUG_MSG_FUNCTIONCALL("createSliders()");
			
			// we first have to traverse all nodes and parameters, where each parameter has to be checked for its dynamics
			list<string> allNodeNames = __m_graphWrapper->getAllNodeNames();
			
			int maxIndex = 0;
			
			unsigned int index = 0;
			// first get necessary dimensions
			for (list<string>::iterator it = allNodeNames.begin(); it != allNodeNames.end(); it++) {
				++index;
				index += __m_graphWrapper->getAllDynamicAttributesForNode(*it).size();
				index += __m_graphWrapper->getAllStaticAttributesForNode(*it).size();
				maxIndex = fmaxf(index, maxIndex);
				
				index = 1;
			}
			
			
			
			

			int labelWidth = 0;
			int labelHeight = 0;
			theGroup->measure_label(labelWidth, labelHeight);

			
			list<string> allAppParameters = __m_graphWrapper->getAllApplicationParameterNames();
			
			for (std::list<string>::iterator it = allAppParameters.begin(); it != allAppParameters.end(); ++it) {
				
				std:cout << *it << std::endl;
				
			}
			
			
			maxIndex = fmaxf(maxIndex, allAppParameters.size());
			
			
			
			
			__m_tabs = new Fl_Tabs(0, 0, 450,  __m_slYStart + (maxIndex) * (__m_slHeight + __m_slDist) + BOX_SPACING, "Tabs");
			__m_tabs->box(FL_UP_BOX);
			
			int tabAreaX, tabAreaY, tabAreaW, tabAreaH;
			__m_tabs->client_area(tabAreaX, tabAreaY, tabAreaW, tabAreaH);
			
			
			
			std::cout << "tabArea: " << tabAreaX << ", " << tabAreaY << ", " << tabAreaW << ", " << tabAreaH << std::endl;
			
			unsigned int newWindowWidth = 0;
			
			
			
			//__m_tabs->size((allNodeNames.size() + 1) * 75, 600);
			
			
			index = 0;
			
			
			for (list<string>::iterator it = allNodeNames.begin(); it != allNodeNames.end(); it++) {
				
				index = 0;
				
				int maxLabelWidth = 0;
				
				Fl_Pack* theGroup = new Fl_Pack(tabAreaX, tabAreaY, tabAreaW, tabAreaH, it->c_str());
				theGroup->spacing(SPACING);
				//theGroup->box(FL_DOWN_FRAME);
				
				
				Fl_Box* spacerBox = new Fl_Box(tabAreaX, tabAreaY, tabAreaW, BOX_SPACING);
				spacerBox->box(FL_NO_BOX);
				
				
				
				int labelWidth = 0;
				int labelHeight = 0;
				theGroup->measure_label(labelWidth, labelHeight);
				
				//std::cout << "labelWidth: " << labelWidth << std::endl;
				
				string nodeType;
				__m_graphWrapper->getNodeTypeByName(*it, nodeType);
				// first add deactivation button if node can be deactivated
				std::cout << *it << " deactivatable: " << __m_nodeDeactivationInfo[nodeType] << std::endl;
				
				if (__m_nodeDeactivationInfo[nodeType]) {
					
					string defaultValue_str = __m_graphWrapper->getAttributeValue(*it, "", "defaultstate");
					
					if (defaultValue_str == "active") {
						
						createDeactivationButton(*it, index, true);
						
					} else if (defaultValue_str == "inactive") {
						
						createDeactivationButton(*it, index, false);
						
					} else { 
						
						std::cerr << "Default node state missing or invalid for node " << *it << ", aborting..." << std::endl;
						exit(0);
						
					} 
					
					++index;
				}
				
				
				list<string> allDynamicAttributes = __m_graphWrapper->getAllDynamicAttributesForNode(*it);
				
				//string nodeType;
				//__m_graphWrapper->getNodeTypeByName(*it, nodeType);
				
				for (list<string>::iterator it2 = allDynamicAttributes.begin(); it2 != allDynamicAttributes.end(); it2++) {
					
					string parameterName;
					__m_graphWrapper->getDataFromNode(*it, *it2, parameterName);
					
					string dataType = __m_graphWrapper->getAttributeValue(*it, *it2, "datatype");
					double minValue = atof(__m_graphWrapper->getAttributeValue(*it, *it2, "min").c_str());
					double maxValue = atof(__m_graphWrapper->getAttributeValue(*it, *it2, "max").c_str());
					double step = atof(__m_graphWrapper->getAttributeValue(*it, *it2, "step").c_str());
					double defaultValue = atof(__m_graphWrapper->getAttributeValue(*it, *it2, "defaultvalue").c_str());
					
					if (__m_alreadyGuified.find(parameterName) == __m_alreadyGuified.end()) {
						if (grip::DataTypes::getIdByName(dataType) == grip::ENUM_DT_UINT32_T) {
							cout << "creating uint32_t slider for " << parameterName << endl;
							__m_uniCon->setObject<grip::dt_uint32_t> (parameterName, defaultValue, 1);
							Fl_Value_Slider* theSlider = createSlider<dt_uint32_t> (minValue, maxValue, step, defaultValue, *it2, parameterName, index);
							
							int sliderLabelWidth = 0;
							int sliderLabelHeight = 0;
							theSlider->measure_label(sliderLabelWidth, sliderLabelHeight);
							
							maxLabelWidth = fmaxf(sliderLabelWidth, maxLabelWidth);
							
							
							
							__m_alreadyGuified[parameterName] = true;
							
						} else if (grip::DataTypes::getIdByName(dataType) == grip::ENUM_DT_FLOAT_T) {
							cout << "creating float slider for " << parameterName << endl;
							
							__m_uniCon->setObject<grip::dt_float_t> (parameterName, defaultValue, 1);
							Fl_Value_Slider* theSlider = createSlider<dt_float_t> (minValue, maxValue, step, defaultValue, *it2, parameterName, index);
							
							
							int sliderLabelWidth = 0;
							int sliderLabelHeight = 0;
							theSlider->measure_label(sliderLabelWidth, sliderLabelHeight);
							maxLabelWidth = fmaxf(sliderLabelWidth, maxLabelWidth);
							
							
							__m_alreadyGuified[parameterName] = true;
						}
					} else {
						
						--index;
						
					}
					
					++index;
					
				}
				
				// Also display all static attributes!
				
				list<string> allStaticAttributes = __m_graphWrapper->getAllStaticAttributesForNode(*it);
				
				for (list<string>::iterator it2 = allStaticAttributes.begin(); it2 != allStaticAttributes.end(); it2++) {
					
					string value;
					__m_graphWrapper->getDataFromNode(*it, *it2, value);
					createTextOutput(*it2, value, index);
					
					++index;
				}
				
				spacerBox = new Fl_Box(tabAreaX, __m_slYStart + (index+1) * (__m_slHeight + __m_slDist), tabAreaW, BOX_SPACING);
				spacerBox->box(FL_NO_BOX);
				
				
				
				theGroup->resize(maxLabelWidth + BOX_SPACING, tabAreaY, tabAreaW - (maxLabelWidth + 2 * BOX_SPACING), tabAreaH);
				
				theGroup->end();
				
				// if there are no parameters, do not display this tab
				if (index == 0) {
					
					__m_tabs->remove(theGroup);
					
				} else {
					
					newWindowWidth += labelWidth * 1.2f;
					
				}
				
				
				
				
			}
			
			
			
			
			//		std::cout << "getting all app parameters" << std::endl;
			
			//		list<string> allAppParameters = __m_graphWrapper->getAllApplicationParameterNames();
			
			//		std::cout << "size: " << allAppParameters.size() << std::endl;
			
			if (allAppParameters.size() > 0) {
				
				
				int maxLabelWidth = 0;
				
				
				Fl_Pack* theGroup = new Fl_Pack(tabAreaX, tabAreaY, tabAreaW, tabAreaH, "App");
				theGroup->spacing(4);
				//theGroup->box(FL_DOWN_FRAME);
				
				Fl_Box* spacerBox = new Fl_Box(tabAreaX, tabAreaY, tabAreaW, BOX_SPACING);
				spacerBox->box(FL_NO_BOX);
				
				
				
				
				int labelWidth = 0;
				int labelHeight = 0;
				theGroup->measure_label(labelWidth, labelHeight);
				newWindowWidth += labelWidth;
				
				index = 0;
				
				for (list<string>::iterator it = allAppParameters.begin(); it != allAppParameters.end(); it++) {
					
					string dataType = __m_graphWrapper->getApplicationParameterAttribute(*it, "datatype");
					//			std::cout << dataType << std::endl;
					
					double minValue = atof(__m_graphWrapper->getApplicationParameterAttribute(*it, "min").c_str());
					//			std::cout << minValue << std::endl;
					
					double maxValue = atof(__m_graphWrapper->getApplicationParameterAttribute(*it, "max").c_str());
					//			std::cout << maxValue << std::endl;
					
					double step = atof(__m_graphWrapper->getApplicationParameterAttribute(*it, "step").c_str());
					//			std::cout << step << std::endl;
					
					double defaultValue = atof(__m_graphWrapper->getApplicationParameterAttribute(*it, "defaultvalue").c_str());
					//			std::cout << defaultValue << std::endl;
					
					if (grip::DataTypes::getIdByName(dataType) == grip::ENUM_DT_UINT32_T) {
						cout << "creating uint32_t slider for " << *it << endl;
						__m_uniCon->setObject<grip::dt_uint32_t> (*it, defaultValue, 1);
						
						Fl_Value_Slider* theSlider = createSlider<dt_uint32_t> (minValue, maxValue, step, defaultValue, *it, *it, index);
						
						
						int sliderLabelWidth = 0;
						int sliderLabelHeight = 0;
						theSlider->measure_label(sliderLabelWidth, sliderLabelHeight);
						
						maxLabelWidth = fmaxf(sliderLabelWidth, maxLabelWidth);
						
						
						__m_alreadyGuified[*it] = true;
						
					} else if (grip::DataTypes::getIdByName(dataType) == grip::ENUM_DT_FLOAT_T) {
						cout << "creating float slider for " << *it << endl;
						__m_uniCon->setObject<grip::dt_float_t> (*it, defaultValue, 1);
						Fl_Value_Slider* theSlider = createSlider<dt_float_t> (minValue, maxValue, step, defaultValue, *it, *it, index);
						
						int sliderLabelWidth = 0;
						int sliderLabelHeight = 0;
						theSlider->measure_label(sliderLabelWidth, sliderLabelHeight);
						maxLabelWidth = fmaxf(sliderLabelWidth, maxLabelWidth);
						
						
						__m_alreadyGuified[*it] = true;
					}


				} else {
				
					--index;
					

				}
				
				spacerBox = new Fl_Box(tabAreaX, __m_slYStart + (index+1) * (__m_slHeight + __m_slDist), tabAreaW, BOX_SPACING);
				spacerBox->box(FL_NO_BOX);
				
				theGroup->resize(maxLabelWidth + BOX_SPACING, tabAreaY, tabAreaW - (maxLabelWidth + 2 * BOX_SPACING), tabAreaH);
				
				theGroup->end();
				
				

			} else {
			
				newWindowWidth += labelWidth * 1.2f;
				

			}
			
	
		
			theGroup->end();
						
			// if there are no parameters, do not display this tab
			if (index == 0) {
				
				__m_tabs->remove(theGroup);
				
			} else {
			
				newWindowWidth += labelWidth * 1.2f;
				
			}

	

			
			
			//__m_tabs->resize(0, 0, __m_window->w(), __m_window->h());
			//theGroup->resize(tabAreaX, tabAreaY, __m_window->w() - tabAreaX, __m_window->h() - tabAreaY);
			
			
			unsigned int newWindowHeight = fmaxf(__m_window->h(), __m_slYStart + (maxIndex) * (__m_slHeight + __m_slDist) + BOX_SPACING);
			newWindowWidth = fmaxf(__m_window->w(), newWindowWidth);
			
			//	std::cout << "newWindowWidth: " << newWindowWidth << std::endl;
			//	std::cout << "newWindowHeight: " << newWindowHeight << std::endl;
			
			
			__m_window->size(newWindowWidth, newWindowHeight);
			__m_tabs->size(newWindowWidth, __m_tabs->h());		
			
			__m_tabs->end();
			__m_window->end();
			
			__m_window->redraw();
			
		}
		
	public:
		/** @brief constructor taking GraphWrapper, UniversalContainer and Mutex */
		CLASSNAME(GraphWrapper* p_gw, UniversalContainer* p_uniCon, map<string, bool>& p_nodeDeactivationInfo, sf::Mutex* p_parameterMutex) {
			DEBUG_MSG_FUNCTIONCALL("ParameterizerGuiThread(" << p_gw << ")");
			__m_slIndex = 0;
			
			__m_slDist = 5;
			__m_slYStart = 0;
			__m_slHeight = 25;
			
			__m_window = 0;
			__m_graphWrapper = p_gw;
			__m_uniCon = p_uniCon;
			
			__m_isReady = false;
			
			__m_parameterMutex = p_parameterMutex;
			
			__m_nodeDeactivationInfo = p_nodeDeactivationInfo;
			
			CheckboxCallbackClass::setUniversalContainer(p_uniCon);
			CheckboxCallbackClass::setParameterMutex(p_parameterMutex);
			
			SliderCallbackClass::setUniversalContainer(p_uniCon);
			SliderCallbackClass::setParameterMutex(p_parameterMutex);
		}
		
		/** @brief default destructor */
		~CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("~ParameterizerGuiThread()");
			
		}
		
		
		bool isReady() {
			DEBUG_MSG_FUNCTIONCALL("isReady()");
			
			return __m_isReady;
		}
		
		
		
		
		/** @brief start the thread */
		void Run() {
			DEBUG_MSG_FUNCTIONCALL("Run()");
			
			__m_window = new Fl_Window(450, 0, "grip::ParameterizerGui");
			
			__m_window->callback(windowEscapeCallback);
			
			createSliders();
			
			__m_window->end();
			__m_window->show();
			
			__m_isReady = true;
			
			Fl::run();
			
		}
		
		
		void updateSlider(std::string p_name) {
			
			//		std::cout << "updating slider: " << p_name << std::endl;
			
			__m_parameterMutex->Lock();
			
			Fl_Value_Slider* theSlider = __m_sliderMap[p_name];
			grip::dt_int32_t typeId = __m_uniCon->getTypeId(p_name);
			if (typeId == grip::ENUM_DT_UINT32_T) {
				dt_uint32_t value = __m_uniCon->getObject<grip::dt_uint32_t>(p_name);
				theSlider->value(value);
				//			std::cout << "uint32_t, value: " << value << std::endl;
			} else if (typeId == grip::ENUM_DT_FLOAT_T) {
				dt_float_t value = __m_uniCon->getObject<grip::dt_float_t>(p_name);
				theSlider->value(value);
				//			std::cout << "float_t, value: " << value << std::endl;
				
			}
			
			//		Fl::redraw();
			//		Fl::flush();
			//		Fl::event();
			Fl::check();
			
			__m_parameterMutex->Unlock();
			
		}
		
		
		/** @brief stop the thread */
		void stopWorking() {
			DEBUG_MSG_FUNCTIONCALL("stopWorking()");
			
			__m_window->hide();
			Fl::check();
		}
		
	};
	
}
#endif
