/*
 * UniversalImageBase.h
 *
 *  Created on: 08.07.2010
 *      Author: troth
 */

#ifndef UNIVERSALIMAGEBASE_H_
#define UNIVERSALIMAGEBASE_H_

#include <Utils/Common.h>


namespace grip {

	#undef	CLASSNAME
	#define	CLASSNAME	UniversalImageBase

	using namespace std;

	/**
	 * @class UniversalImageBase
	 * @brief provides basic information for images which need no template stuff, serves as super class for UniversalImage and its sub classes
	 */
	class CLASSNAME {
	private:
		string __sm_orientationNames[2];
		uint32_t __m_res[2];
		uint32_t __m_orientation;


		// syntactic sugar ;)
		#define __m_resx __m_res[ENUM_UI_RES_X]
		#define __m_resy __m_res[ENUM_UI_RES_Y]

	protected:

		/** @brief set horizontal and vertical image resolution */
		INLINE void setRes(uint32_t p_resx, uint32_t p_resy) {
			DEBUG_MSG_FUNCTIONCALL("setRes(" << p_resx << ", " << p_resy << ")");

			__m_resx = p_resx;
			__m_resy = p_resy;
		}

	public:


		enum ui_resolutionIdentifiers {
			ENUM_UI_RES_X = 0,
			ENUM_UI_RES_Y
		};


		enum ui_orientationIdentifiers {
			ENUM_UI_ORI_LR_TB = 0,	// left to right, top to bottom
			ENUM_UI_ORI_LR_BT		// left to right, bottom to top
		};

		/** @brief create new image with orientation ltr ttb */
		INLINE CLASSNAME() : __m_orientation(ENUM_UI_ORI_LR_TB) {
			DEBUG_MSG_FUNCTIONCALL("UniversalImageBase()");

			__sm_orientationNames[0] = "LEFT TO RIGHT, TOP TO BOTTOM";
			__sm_orientationNames[1] = "LEFT TO RIGHT, BOTTOM TO TOP";

		}


		/** @brief get image resolution */
		INLINE void getRes(uint32_t& p_resx, uint32_t& p_resy) {
			DEBUG_MSG_FUNCTIONCALL("getRes(" << p_resx << ", " << p_resy << ")");

			p_resx = __m_resx;
			p_resy = __m_resy;
		}


		/** @brief get image width */
		INLINE uint32_t getResX() {
			DEBUG_MSG_FUNCTIONCALL("getResX()");

			return __m_resx;
		}

		/** @brief get image width */
		INLINE uint32_t getWidth() {
			DEBUG_MSG_FUNCTIONCALL("getResWidth()");

			return __m_resx;
		}


		/** @brief get image height */
		INLINE uint32_t getResY() {
			DEBUG_MSG_FUNCTIONCALL("getResY()");

			return __m_resy;
		}

		/** @brief get image height */
		INLINE uint32_t getHeight() {
			DEBUG_MSG_FUNCTIONCALL("getHeight()");

			return __m_resy;
		}


		/** @brief get image orientation */
		INLINE uint32_t getOrientation() {
			DEBUG_MSG_FUNCTIONCALL("getOrientation()");

			return __m_orientation;
		}


		/** @brief get image orientation description */
		INLINE string getOrientationDescription() {
			DEBUG_MSG_FUNCTIONCALL("getOrientationDescription()");

			return __sm_orientationNames[__m_orientation];
		}


		/** @brief get description for p_orientation */
		INLINE string getOrientationDescription(uint32_t p_orientation) {
			DEBUG_MSG_FUNCTIONCALL("getOrientationDescription(" << p_orientation << ")");

			return __sm_orientationNames[p_orientation];
		}


		/** @brief set image orientation. Does not change data, which is done by changeOrientation in UniversalImage. */
		INLINE void setOrientation(uint32_t p_orientation) {
			DEBUG_MSG_FUNCTIONCALL("setOrientation(" << p_orientation << ")");

			DEBUG_MSG_RUNTIMEINFO("setOrientation, setting orientation to " << getOrientationDescription(p_orientation));
			__m_orientation = p_orientation;
		}
	};

}

#endif /* UNIVERSALIMAGEBASE_H_ */
