/*
 * BitUtils.h
 *
 *  Created on: 29.06.2010
 *      Author: troth
 */

#ifndef BITUTILS_H_
#define BITUTILS_H_

#include <endian.h>

// just a collection of bit utils

#if __BYTE_ORDER == __LITTLE_ENDIAN
	#define BU_SHIFT_TO_GREATER(x,y)		(x << y)
	#define BU_SHIFT_TO_LESSER(x,y)			(x >> y)
	#define BU_GET_BYTE_UINT32(x,y)			BU_SHIFT_TO_LESSER((x & (MASKING_SHIFT_TO_GREATER(0x000000FF, 8*y))), 8*y) // get byte number y in x (origin at LSB)

	#define	BU_MASK_LS_BYTE_FOR_BITS(x)		0x000000FF
	#define BU_MASK_MS_BYTE_FOR_BITS(x)		BU_SHIFT_TO_GREATER(0x000000FF, (x-8)) // get MSB for an x-bit number

	#define BU_MASK_LS_BIT_FOR_BITS(x)		0x00000001
	#define BU_MASK_MS_BIT_FOR_BITS(x)		BU_SHIFT_TO_GREATER(0x00000001, x-1)

	#define BU_MASK_LS_BYTE_UINT32			BU_MASK_LS_BYTE_FOR_BITS(32)
	#define BU_MASK_MS_BYTE_UINT32			BU_MASK_MS_BYTE_FOR_BITS(32)

	#define BU_MASK_LS_BIT_UINT32			BU_MASK_LS_BIT_FOR_BITS(32)
	#define BU_MASK_MS_BIT_UINT32			BU_MASK_MS_BIT_FOR_BITS(32)

#else //probably big endian
	#define BU_SHIFT_TO_GREATER(x,y)		(x >> y)
	#define BU_SHIFT_TO_LESSER(x,y)			(x << y)
	#define BU_GET_BYTE_UINT32(x,y)			BU_SHIFT_TO_LESSER((x & (MASKING_SHIFT_TO_GREATER(0xFF000000, 8*y))), 8*y) // get byte number y in x (origin at LSB)

	#define	BU_MASK_LS_BYTE_FOR_BITS(x)		BU_SHIFT_TO_LESSER(0x000000FF, (x-8))
	#define BU_MASK_MS_BYTE_FOR_BITS(x)		0x000000FF

	#define BU_MASK_LS_BIT_FOR_BITS(x)		BU_SHIFT_TO_LESSER(0x00000001, x-1)
	#define BU_MASK_MS_BIT_FOR_BITS(x)		0x00000001

	#define BU_MASK_LS_BYTE_UINT32			BU_MASK_LS_BYTE_FOR_BITS(32)
	#define BU_MASK_MS_BYTE_UINT32			BU_MASK_MS_BYTE_FOR_BITS(32)

	#define BU_MASK_LS_BIT_UINT32			BU_MASK_LS_BIT_FOR_BITS(32)
	#define BU_MASK_MS_BIT_UINT32			BU_MASK_MS_BIT_FOR_BITS(32)

#endif


// byte masks from  right to left (little endian sense)
#define BU_MASK_B1			0x000000FF
#define BU_MASK_B2			0x0000FF00
#define BU_MASK_B3			0x00FF0000
#define BU_MASK_B4			0xFF000000
#define BU_MASK_FULL32		0xFFFFFFFF

// byte mask offsets from right to left (little endian sense)
#define BU_MASKOFFSET_B1	0
#define BU_MASKOFFSET_B2 	8
#define BU_MASKOFFSET_B3	16
#define BU_MASKOFFSET_B4	24


// color masks for 32bit integers
#define BU_MASK_UINT32_ARGB_ALPHA		BU_MASK_B4
#define BU_MASK_UINT32_ARGB_RED			BU_MASK_B3
#define BU_MASK_UINT32_ARGB_GREEN		BU_MASK_B2
#define BU_MASK_UINT32_ARGB_BLUE		BU_MASK_B1

#define BU_MASKOFFSET_UINT32_ARGB_ALPHA	BU_MASKOFFSET_B4
#define BU_MASKOFFSET_UINT32_ARGB_RED 	BU_MASKOFFSET_B3
#define BU_MASKOFFSET_UINT32_ARGB_GREEN	BU_MASKOFFSET_B2
#define BU_MASKOFFSET_UINT32_ARGB_BLUE	BU_MASKOFFSET_B1


#define BU_MASK_UINT32_RGBA_RED			BU_MASK_B4
#define BU_MASK_UINT32_RGBA_GREEN		BU_MASK_B3
#define BU_MASK_UINT32_RGBA_BLUE		BU_MASK_B2
#define BU_MASK_UINT32_RGBA_ALPHA		BU_MASK_B1

#define BU_MASKOFFSET_UINT32_RGBA_RED	BU_MASKOFFSET_B4
#define BU_MASKOFFSET_UINT32_RGBA_GREEN	BU_MASKOFFSET_B3
#define BU_MASKOFFSET_UINT32_RGBA_BLUE	BU_MASKOFFSET_B2
#define BU_MASKOFFSET_UINT32_RGBA_ALPHA	BU_MASKOFFSET_B1


#define BU_MASK_UINT32_ABGR_ALPHA		BU_MASK_B4
#define BU_MASK_UINT32_ABGR_BLUE		BU_MASK_B3
#define BU_MASK_UINT32_ABGR_GREEN		BU_MASK_B2
#define BU_MASK_UINT32_ABGR_RED			BU_MASK_B1

#define BU_MASKOFFSET_UINT32_ABGR_ALPHA	BU_MASKOFFSET_B4
#define BU_MASKOFFSET_UINT32_ABGR_BLUE	BU_MASKOFFSET_B3
#define BU_MASKOFFSET_UINT32_ABGR_GREEN	BU_MASKOFFSET_B2
#define BU_MASKOFFSET_UINT32_ABGR_RED	BU_MASKOFFSET_B1


#define BU_MASK_UINT32_BGRA_BLUE		BU_MASK_B4
#define BU_MASK_UINT32_BGRA_GREEN		BU_MASK_B3
#define BU_MASK_UINT32_BGRA_RED			BU_MASK_B2
#define BU_MASK_UINT32_BGRA_ALPHA		BU_MASK_B1

#define BU_MASKOFFSET_UINT32_BGRA_BLUE	BU_MASKOFFSET_B4
#define BU_MASKOFFSET_UINT32_BGRA_GREEN	BU_MASKOFFSET_B3
#define BU_MASKOFFSET_UINT32_BGRA_RED 	BU_MASKOFFSET_B2
#define BU_MASKOFFSET_UINT32_BGRA_ALPHA	BU_MASKOFFSET_B1



// color access for 32bit integers
#define BU_COLOR_UINT32_ARGB_ALPHA(x)	((x & BU_MASK_UINT32_ARGB_ALPHA)	>> BU_MASKOFFSET_UINT32_ARGB_ALPHA)
#define BU_COLOR_UINT32_ARGB_RED(x)		((x & BU_MASK_UINT32_ARGB_RED)		>> BU_MASKOFFSET_UINT32_ARGB_RED)
#define BU_COLOR_UINT32_ARGB_GREEN(x)	((x & BU_MASK_UINT32_ARGB_GREEN)	>> BU_MASKOFFSET_UINT32_ARGB_GREEN)
#define BU_COLOR_UINT32_ARGB_BLUE(x)	((x & BU_MASK_UINT32_ARGB_BLUE) 	>> BU_MASKOFFSET_UINT32_ARGB_BLUE)

#define BU_COLOR_UINT32_RGBA_RED(x)		((x & BU_MASK_UINT32_RGBA_RED)		>> BU_MASKOFFSET_UINT32_RGBA_RED)
#define BU_COLOR_UINT32_RGBA_GREEN(x)	((x & BU_MASK_UINT32_RGBA_GREEN)	>> BU_MASKOFFSET_UINT32_RGBA_GREEN)
#define BU_COLOR_UINT32_RGBA_BLUE(x)	((x & BU_MASK_UINT32_RGBA_BLUE)		>> BU_MASKOFFSET_UINT32_RGBA_BLUE)
#define BU_COLOR_UINT32_RGBA_ALPHA(x)	((x & BU_MASK_UINT32_RGBA_ALPHA) 	>> BU_MASKOFFSET_UINT32_RGBA_ALPHA)

#define BU_COLOR_UINT32_ABGR_ALPHA(x)	((x & BU_MASK_UINT32_ABGR_ALPHA)	>> BU_MASKOFFSET_UINT32_ABGR_ALPHA)
#define BU_COLOR_UINT32_ABGR_BLUE(x)	((x & BU_MASK_UINT32_ABGR_BLUE)		>> BU_MASKOFFSET_UINT32_ABGR_BLUE)
#define BU_COLOR_UINT32_ABGR_GREEN(x)	((x & BU_MASK_UINT32_ABGR_GREEN)	>> BU_MASKOFFSET_UINT32_ABGR_GREEN)
#define BU_COLOR_UINT32_ABGR_RED(x)		((x & BU_MASK_UINT32_ABGR_RED)		>> BU_MASKOFFSET_UINT32_ABGR_RED)

#define BU_COLOR_UINT32_BGRA_BLUE(x)	((x & BU_MASK_UINT32_BGRA_BLUE)		>> BU_MASKOFFSET_UINT32_BGRA_BLUE)
#define BU_COLOR_UINT32_BGRA_GREEN(x)	((x & BU_MASK_UINT32_BGRA_GREEN)	>> BU_MASKOFFSET_UINT32_BGRA_GREEN)
#define BU_COLOR_UINT32_BGRA_RED(x)		((x & BU_MASK_UINT32_BGRA_RED)		>> BU_MASKOFFSET_UINT32_BGRA_RED)
#define BU_COLOR_UINT32_BGRA_ALPHA(x)	((x & BU_MASK_UINT32_BGRA_ALPHA)	>> BU_MASKOFFSET_UINT32_BGRA_ALPHA)

#endif /* BITUTILS_H_ */
