/*
 * CudaHelper.h
 *
 *  Created on: Dec 14, 2010
 *      Author: troth22s
 *
 * changelog:
 * 	2011-12-21:
 * 		- Introduced gripCudaMallocAndCopyIfNullOrDirty(...) and gripCudaMallocIfNull(...) as functions
 * 		- Changed the previously used macros to just forward their calls to these functions
 */

#ifndef CUDAHELPER_H_
#define CUDAHELPER_H_

#include <cuda_runtime.h>
#include <helper_cuda.h>

#include <Data/EnumComputeEngine.h>
#include <Utils/Common.h>

#ifndef __CUDACC__
#include <Data/TempBufferRepository.h>
#endif


/* the following macros are only there for backwards compatibility and now forward the calls to according functions which are defined below */

#define GRIP_CUDA_MALLOC_AND_COPY_IF_NOD(TARGET, SOURCE, SIZE, DIRTY) \
	grip::gripCudaMallocAndCopyIfNullOrDirty((void**) &TARGET, SOURCE, SIZE, DIRTY);

/* allocate based on current data status, indicated by pointer */
#define GRIP_CUDA_MALLOC_IF_NULL(TARGET, SIZE) \
	grip::gripCudaMallocIfNull((void**) &TARGET, SIZE);

/* allocate based on current data status, indicated by pointer */
#define GRIP_CUDA_MALLOC_HOST_IF_NULL(TARGET, SIZE) \
	grip::gripCudaMallocHostIfNull((void**) &TARGET, SIZE);



/* allocate and copy based on current data status, indicated by pointer and dirty flag */
namespace grip {

	
#ifndef __CUDACC__
	/* @brief This is used as a unified way to initialize or get CUDA memory from the temp buffer repository */
	grip::TempBufferWrapper* initializeCudaBufferFromRepo(void** p_data, grip::TempBufferRepository* p_tbRepo, unsigned int p_size) {
		grip::TempBufferWrapper* theBuffer = p_tbRepo->getUsableBuffer(p_size, grip::ENUM_COMPENG_CUDA);
		
		if (!theBuffer) {
			
			std::cout << "No usable buffer found" << std::endl;
			//cudaMallocPitch(p_data, p_pitchInBytes, p_widthInBytes, p_heightInRows);
			cudaError_t retval = cudaMalloc(p_data, p_size);
			if (!(retval == cudaSuccess)) {
				
				std::cout << "cudaMalloc failed" << std::endl;
				exit(0);
			}
			
			//*p_data = malloc(p_size);
			
			//std::cout << "New Pointer: " << *p_data << std::endl;
			
			
			
			theBuffer = p_tbRepo->addBuffer(*p_data, p_size, true, grip::ENUM_COMPENG_CUDA);
			
		} else {
			*p_data = theBuffer->getData();
		}
		
		/*cudaError_t retval = cudaMemset(*p_data, 1, p_size);
		
		if (!(retval == cudaSuccess)) {
			std::cout << "cudaMemset failed" << std::endl;
			exit(0);
		}*/
		
		return theBuffer;
	}
#endif


INLINE void gripCudaMallocAndCopyIfNullOrDirty(void** p_target, void* p_source, size_t p_size, grip::enum_computeEngine& p_dirty) {
	DEBUG_MSG_FUNCTIONCALL("gripCudaMallocAndCopyIfNullOrDirty(" << p_target << ", " << p_source << ", " << p_size << ", " << p_dirty << ")");

	if (!*p_target) {
		
		cudaMalloc(p_target, p_size);
				
		cudaMemcpy(*p_target, p_source, p_size, cudaMemcpyHostToDevice);
		p_dirty = ENUM_COMPENG_NONE;

		
	} else if (p_dirty == ENUM_COMPENG_HOST) {
	
		cudaMemcpy(*p_target, p_source, p_size, cudaMemcpyHostToDevice);
		p_dirty = ENUM_COMPENG_NONE;
	}
}

INLINE void gripCudaMallocIfNull(void** p_target, size_t p_size) {
	DEBUG_MSG_FUNCTIONCALL("gripCudaMallocIfNull(" << p_target << ", " << p_size << ")");

	if (!*p_target) {
		cudaMalloc(p_target, p_size);
	}
}

INLINE void gripCudaMallocHostIfNull(void** p_target, size_t p_size) {
	DEBUG_MSG_FUNCTIONCALL("gripCudaMallocIfNull(" << p_target << ", " << p_size << ")");

	if (!*p_target) {
		cudaMallocHost(p_target, p_size);
	}
}

}


#endif /* CUDAHELPER_H_ */
