/*
 * VecUtils.h
 *
 *  Created on: 09.09.2010
 *      Author: troth
 */

#ifndef VECUTILS_H_
#define VECUTILS_H_

#include <Utils/Common.h>

namespace grip {

	#undef CLASSNAME
	#define CLASSNAME VecUtils

	/**
	 * @class VecUtils
	 * @brief provides some template-based vector utility functions
	 */
	namespace CLASSNAME {

		/** @brief get length of a vector pointed to by p_vec with p_numElements elements */
		template<class FTYPE> INLINE FTYPE length(const FTYPE* p_vec, uint32_t p_numElements ) {
			DEBUG_MSG_FUNCTIONCALL("length(" << p_vec << ", " << p_numElements << ")");

			FTYPE sum;
			for (int i = 0; i < p_numElements; ++i) {
				sum += p_vec[i];
			}
			return (FTYPE)sqrt(sum);
		}

		/** @brief normalize a vector pointed to by p_vec with p_numElements elements */
		template<class FTYPE> INLINE void normalize(const FTYPE* p_vec, uint32_t p_numElements) {
			DEBUG_MSG_FUNCTIONCALL("normalize(" << p_vec << ", " << p_numElements << ")");

			FTYPE l = length(p_vec, p_numElements);
			for (int i = 0; i < p_numElements; ++i) {
				p_vec[i] /= l;
			}
		}


	}

}
#endif /* VECUTILS_H_ */
