/* CUDA Node template for use with the PoProcK image processing library */

#include <iostream>
#include <string>

#include <Graph/NodeInterface.h>
#include <Graph/NodeHelper.h>
#include <Utils/TimerMacros.h>

#include "parameters.h"

/** has to be renamed accordingly */

using namespace std;


extern "C" bool can_be_deactivated() {
	return false;	
}

/** this is the main kernel run function. Arguments may be changed according to what the filter does, e.g. to pass image pointers etc. */
extern void run_cuda(Parameters<float*>& p_parm);

extern "C" void run_node(const string& p_nodeName, grip::GraphWrapper& p_gw, grip::UniversalContainer& p_uniCon, grip::TempBufferRepositoryDirectory& p_tbRepoDir) {
	/** just output information ;) */
	//cout << "RUN_NODE (node_subtract) CALLED! (" << p_nodeName << ")" << endl;
	___GRIP_INITIALIZE_TIMER(0);

	___GRIP_START_TIMER(0);

//	___GRIP_END_TIMER_AND_PRINT(0, "node_gauss_cuda", "total time", p_nodeName);
	grip::NodeHelper nh(p_nodeName, p_gw, p_uniCon);

	string summand1Type, summand1Name, summand2Type, summand2Name, targetType, targetName;
	nh.getIOTypeAndIdentifier("input_summand1", summand1Type, summand1Name);
	nh.getIOTypeAndIdentifier("input_summand2", summand2Type, summand2Name);
	nh.getIOTypeAndIdentifier("output", targetType, targetName);

	if (summand1Type != summand2Type || summand1Type != targetType) {
		nh.errorInequalTypes();
	}

	if (summand1Type == "image_float_4") {

		Parameters<float*> parm;

		GRIP_GET_IMAGE_AND_SET_PARAMETERS_FOR_CUDA(grip::dt_UniversalImage_float_4_sptr, float*, summand1Image, summand1Name, parm.summand1Image);
		GRIP_GET_IMAGE_AND_SET_PARAMETERS_FOR_CUDA(grip::dt_UniversalImage_float_4_sptr, float*, summand2Image, summand2Name, parm.summand2Image);

		parm.resx = summand1Image->getResX();
		parm.resy = summand1Image->getResY();

		GRIP_CREATE_TARGET_IMAGE_AND_SET_PARAMETERS_FOR_CUDA(grip::dt_UniversalImage_float_4, float*, parm.resx, parm.resy, targetImage, targetName, parm.targetImage);

		run_cuda(parm);

		p_uniCon.setObject<grip::dt_UniversalImage_float_4_sptr>(targetName, targetImage, 1);

		GRIP_UPDATE_DIRTY_FLAG_AND_DEVICE_PTR(summand1Name, parm.summand1Image, float*);
		GRIP_UPDATE_DIRTY_FLAG_AND_DEVICE_PTR(summand2Name, parm.summand2Image, float*);
		GRIP_UPDATE_DIRTY_FLAG_AND_DEVICE_PTR(targetName, parm.targetImage, float*);

	} else {
		nh.errorUnsupportedFormat();
	}

	___GRIP_END_TIMER_AND_PRINT(0, "node_add_cuda", "total time", p_nodeName);

}
