#include <cuda.h>
//#include <helper_cuda.h>
//#include <helper_functions.h> 

#include <iostream>
//#include <cutil_inline.h>

#include <Utils/CudaHelper.h>
#include <Utils/TimerMacros.h>

#include "parameters.h"

typedef struct __PARAMETERS_FLOAT {
	int resx;
	int resy;
	float* targetImage;
} parm_float;

#undef PARM
#define PARM parm_float_c
__device__ __constant__ parm_float PARM;

__global__ void add(float* p_summand1Image, float* p_summand2Image) {
	/** for each pixel */
	for (int y = blockIdx.x; y < PARM.resy; y += gridDim.x) {
		for (int x = threadIdx.x; x < PARM.resx *4; x += blockDim.x) {
			int pos = y * PARM.resx * 4 + x ;
			PARM.targetImage[pos] = p_summand1Image[pos] + p_summand2Image[pos];
		}
	}
}

void run_cuda(Parameters<float*>& p_parm) {
	int n = (p_parm.resx * p_parm.resy); // number of pixels
	size_t size = n * sizeof(float) * 4; //RGBA float image
	
	GRIP_CUDA_MALLOC_AND_COPY_IF_NOD(p_parm.summand1Image_d, p_parm.summand1Image_h, size, p_parm.summand1Image_dirty);
	GRIP_CUDA_MALLOC_AND_COPY_IF_NOD(p_parm.summand2Image_d, p_parm.summand2Image_h, size, p_parm.summand2Image_dirty);
	GRIP_CUDA_MALLOC_IF_NULL(p_parm.targetImage_d, size);

	parm_float parm_host;
	parm_host.resx = p_parm.resx;
	parm_host.resy = p_parm.resy;
	parm_host.targetImage = p_parm.targetImage_d;
	cudaMemcpyToSymbol(PARM, &parm_host, sizeof(parm_host));

	___GRIP_INITIALIZE_TIMER(1);
	___GRIP_START_TIMER(1);

	add<<< 64, 256 >>>(p_parm.summand1Image_d, p_parm.summand2Image_d);
	//cutilCheckMsg("Kernel execution failed!");

	cudaThreadSynchronize();
	
	___GRIP_END_TIMER_AND_PRINT(1, "node_add_cuda", "kernel execution time", "run_cuda");

	p_parm.targetImage_dirty = grip::ENUM_COMPENG_CUDA;

}
