#include <cuda.h>
#include <iostream>
//#include <cutil_inline.h>

#include <Utils/CudaHelper.h>
#include <helper_cuda.h>
#include <helper_math.h>

#include "parameters.h"

typedef struct __PARAMETERS_FLOAT {
	int resx;
	int resy;
	int alpha;
} parm_float;

#undef PARM
#define PARM parm_float_c
__device__ __constant__ parm_float PARM;

__global__ void convertIfloat4ToIuint84(float* p_colorImage, uchar* p_targetImage) {
	/** for each pixel */
	for (int y = blockIdx.x; y < PARM.resy; y += gridDim.x) {
		for (int x = threadIdx.x; x < PARM.resx; x += blockDim.x) {
			int pos = y * PARM.resx * 4 + 4*x ;

			// convert range and flip upside down!
			p_targetImage[pos] = fmaxf(0.f, fminf(p_colorImage[(PARM.resy - y - 1) * PARM.resx * 4 + 4*x] * 255.99f, 255.f));
			p_targetImage[pos+1] = fmaxf(0.f, fminf(p_colorImage[(PARM.resy - y - 1) * PARM.resx * 4 + 4*x + 1] * 255.99f, 255.f));
			p_targetImage[pos+2] = fmaxf(0.f, fminf(p_colorImage[(PARM.resy - y - 1) * PARM.resx * 4 + 4*x + 2] * 255.99f, 255.f));
			p_targetImage[pos+3] = PARM.alpha;//fmaxf(0.f, fminf(p_colorImage[pos+2] * 255.99f, 255));
		}
	}
}

void run_cuda(Parameters& p_parm) {
	int n = (p_parm.resx * p_parm.resy); // number of pixels
	size_t size = n * sizeof(float) * 4; //RGBA float image

	GRIP_CUDA_MALLOC_AND_COPY_IF_NOD(p_parm.colorImage_d, p_parm.colorImage_h, size, p_parm.colorImage_dirty);
//	GRIP_CUDA_MALLOC_AND_COPY_IF_NOD(p_parm.subtrahendImage_d, p_parm.subtrahendImage_h, size, p_parm.subtrahendImage_dirty);

	GRIP_CUDA_MALLOC_IF_NULL(p_parm.targetImage_d, size);
	
	parm_float parm_host;
	parm_host.resx = p_parm.resx;
	parm_host.resy = p_parm.resy;
	parm_host.alpha = p_parm.alpha * 255.f;
	cudaMemcpyToSymbol(PARM, &parm_host, sizeof(parm_host));

	convertIfloat4ToIuint84<<< 64, 256 >>>(p_parm.colorImage_d, p_parm.targetImage_d);

//	cutilCheckMsg("Kernel execution failed!");
	
	p_parm.targetImage_dirty = grip::ENUM_COMPENG_CUDA;

}
