/* CUDA Node template for use with the PoProcK image processing library */
#include <iostream>
#include <string>

#include <Utils/CudaHelper.h>
#include <Graph/NodeInterface.h>
#include <Graph/NodeHelper.h>
#include <Data/TempBufferRepository.h>

#include "parameters.h"

/** has to be renamed accordingly */

using namespace std;


/** this is the main kernel run function. Arguments may be changed according to what the filter does, e.g. to pass image pointers etc. */
extern void run_cuda(Parameters<float*>& p_parm);

extern "C" bool can_be_deactivated() {
	return true;	
}

extern "C" void run_node(const string& p_nodeName, grip::GraphWrapper& p_gw, grip::UniversalContainer& p_uniCon, grip::TempBufferRepositoryDirectory& p_tbRepoDir) {
	/** just output information ;) */
	//cout << "RUN_NODE (node_subtract) CALLED! (" << p_nodeName << ")" << endl;

	grip::NodeHelper nh(p_nodeName, p_gw, p_uniCon);
	string colorType, colorName, targetType, targetName;
	nh.getIOTypeAndIdentifier("input", colorType, colorName);
	nh.getIOTypeAndIdentifier("output", targetType, targetName);
	
	Parameters<float*> parm;
	
	GRIP_GET_IMAGE_AND_SET_PARAMETERS_FOR_CUDA(grip::dt_UniversalImage_float_4_sptr, float*, colorImage, colorName, parm.colorImage);
	parm.resx = colorImage->getResX();
	parm.resy = colorImage->getResY();
	GRIP_CREATE_TARGET_IMAGE_AND_SET_PARAMETERS_FOR_CUDA(grip::dt_UniversalImage_float_4, float*, parm.resx, parm.resy, targetImage, targetName, parm.targetImage);
	
	int n = (parm.resx * parm.resy); // number of pixels
	size_t size = n * sizeof(float) * 4; //RGBA float image
	
	GRIP_CUDA_MALLOC_AND_COPY_IF_NOD(parm.colorImage_d, parm.colorImage_h, size, parm.colorImage_dirty);
	GRIP_CUDA_MALLOC_IF_NULL(parm.targetImage_d, size);

	if (!p_uniCon.getNodeActivation(p_nodeName)) {
			
		cudaMemcpy(parm.targetImage_d, parm.colorImage_d, parm.resx * parm.resy * sizeof(float) * 4, cudaMemcpyDeviceToDevice);
			
	} else {
	

		//GRIP_GET_STATIC_OR_DYNAMIC_ATTRIBUTE("worldLuminance", parm.worldLuminance, grip::dt_float_t);

		GRIP_GET_STATIC_OR_DYNAMIC_ATTRIBUTE("alpha", parm.alpha, grip::dt_float_t);
		GRIP_GET_STATIC_OR_DYNAMIC_ATTRIBUTE("worldLuminanceWeight", parm.worldLuminanceWeight, grip::dt_float_t);

		run_cuda(parm);

	}
			
	p_uniCon.setObject<grip::dt_UniversalImage_float_4_sptr>(targetName, targetImage, 1);
	parm.targetImage_dirty = grip::ENUM_COMPENG_CUDA;
	GRIP_UPDATE_DIRTY_FLAG_AND_DEVICE_PTR(colorName, parm.colorImage, float*);
	GRIP_UPDATE_DIRTY_FLAG_AND_DEVICE_PTR(targetName, parm.targetImage, float*);

}
