/* CUDA Node template for use with the PoProcK image processing library */

#include <iostream>
#include <string>
#include <cuda_runtime.h>
//#include <cutil_inline.h>

#include <Graph/NodeInterface.h>
#include <Graph/NodeHelper.h>
#include <Data/EnumComputeEngine.h>
#include <Utils/TimerMacros.h>
#include <Data/TempBufferRepository.h>
#include <Utils/CudaHelper.h>

/** has to be renamed accordingly */

using namespace std;

extern "C" bool can_be_deactivated() {
	return false;	
}

extern "C" void run_node(const string& p_nodeName, grip::GraphWrapper& p_gw, grip::UniversalContainer& p_uniCon, grip::TempBufferRepositoryDirectory& p_tbRepoDir) {
	/** just output information ;) */
	//cout << "RUN_NODE (node_subtract) CALLED! (" << p_nodeName << ")" << endl;
	___GRIP_INITIALIZE_TIMER(0);

	___GRIP_START_TIMER(0);

	//	___GRIP_END_TIMER_AND_PRINT(0, "node_gauss_cuda", "total time", p_nodeName);
	
	grip::NodeHelper nh(p_nodeName, p_gw, p_uniCon);

	string inputType, inputName;
	nh.getIOTypeAndIdentifier("input_image", inputType, inputName);

	void* hostPtr = 0;
	size_t size = 0;

	if (inputType == "image_float_4") {
		grip::dt_UniversalImage_float_4_sptr inputImage = p_uniCon.getObject<grip::dt_UniversalImage_float_4_sptr>(inputName);
		hostPtr = inputImage->getDataPointer().get();
		size = inputImage->getTotalSizeInBytes();

//		std::cout << "before: " << hostPtr << std::endl;
		if (hostPtr == 0) {
			GRIP_CUDA_MALLOC_HOST_IF_NULL(hostPtr, size);
			inputImage->bindDataNoDelete((float*)hostPtr);
		}
//		std::cout << "after: " << hostPtr << std::endl;
	} else if (inputType == "image_uint8_4") {
		grip::dt_UniversalImage_uint8_4_sptr inputImage = p_uniCon.getObject<grip::dt_UniversalImage_uint8_4_sptr>(inputName);
		hostPtr = inputImage->getDataPointer().get();
		size = inputImage->getTotalSizeInBytes();

//		std::cout << "before: " << hostPtr << std::endl;
		if (hostPtr == 0) {
			GRIP_CUDA_MALLOC_HOST_IF_NULL(hostPtr, size);
			inputImage->bindDataNoDelete((unsigned char*)hostPtr);
		}
//		std::cout << "after: " << hostPtr << std::endl;
	} else {
		nh.errorUnsupportedFormat();
	}

	void* devicePtr = p_uniCon.getDevicePtr(inputName);

	if (p_uniCon.isDirty(inputName) == grip::ENUM_COMPENG_HOST) {
		std::cout << "*** Warning: Overwriting dirty host data!" << std::endl;
	}



	if (devicePtr != 0 && hostPtr != 0) {
		cudaMemcpy(hostPtr, devicePtr, size, cudaMemcpyDeviceToHost);
		p_uniCon.setDirty(inputName, grip::ENUM_COMPENG_NONE);
	}

	___GRIP_END_TIMER_AND_PRINT(0, "node_sync_cuda", "total time", p_nodeName);

}
