#include <cuda.h>
#include <iostream>
#include <math.h>
#include <sys/time.h>

#include <Utils/CudaHelper.h>

#include "parameters.h"

typedef struct __PARAMETERS_FLOAT {
	int resx;
	int resy;
	float aperture;
	float focal_length;
	float sensor_height;
	float sensor_width;
	float factor_speed;
	float* targetImage;

} parm_float;

#undef PARM
#define PARM parm_subtract_float_c
__device__ __constant__ parm_float PARM;





// "__global__" =  Deklaration eines Kernels. 
__global__ void vignetting(float* p_colorImage) {
	
	float centery = PARM.resy / 2.f;
	float centerx = PARM.resx / 2.f;
	float temp = (centery * centery) + (centerx * centerx);
	float max_distance = sqrtf(temp);
	float sensor_diagonal = sqrtf((PARM.sensor_height) * (PARM.sensor_height) + (PARM.sensor_width) * (PARM.sensor_width));
	float max_screen_angle = (atanf(sensor_diagonal / (PARM.focal_length)));
	//float aperture_2 = (PARM.aperture) * (PARM.aperture) / PARM.factor_speed;
	float aperture_2 = 1.f;


	/** for each pixel */
	for (int y = blockIdx.x; y < PARM.resy; y += gridDim.x) {
		for (int x = threadIdx.x; x < PARM.resx; x += blockDim.x) {
			int pos = y * PARM.resx * 4 + 4*x ;
			float current_distance = sqrtf((y - centery) * (y - centery) + (x - centerx) * (x - centerx));
			float current_angle = max_screen_angle * current_distance / max_distance;			
			PARM.targetImage[pos]   = (p_colorImage[pos]) * powf(cosf(current_angle), 4) / aperture_2; 
			PARM.targetImage[pos+1] = (p_colorImage[pos+1]) * powf(cosf(current_angle), 4) / aperture_2;
			PARM.targetImage[pos+2] = (p_colorImage[pos+2]) * powf(cosf(current_angle), 4) / aperture_2;
		}
	}
}

void run_cuda(Parameters<float*>& p_parm) {

    int n = (p_parm.resx * p_parm.resy); // number of pixels
	size_t size = n * sizeof(float) * 4; //RGBA float image
		
	// Die Zustände des dirty-Flag : NONE, HOST, DEVICE
	// dirty-Flag bezeichnet, wie die Daten im Hostspeicher und Grafikspeicher verändert vorliegen.
	//GRIP_CUDA_MALLOC_AND_COPY_IF_NOD(p_parm.colorImage_d, p_parm.colorImage_h, size, p_parm.colorImage_dirty);
//	GRIP_CUDA_MALLOC_AND_COPY_IF_NOD(p_parm.subtrahendImage_d, p_parm.subtrahendImage_h, size, p_parm.subtrahendImage_dirty);

	//GRIP_CUDA_MALLOC_IF_NULL(p_parm.targetImage_d, size);

	parm_float parm_host;
	parm_host.factor_speed = p_parm.factor_speed;
	parm_host.focal_length = p_parm.focal_length;
	parm_host.sensor_height = p_parm.sensor_height;
	parm_host.sensor_width = p_parm.sensor_width;
	parm_host.aperture = p_parm.aperture;
	parm_host.resx = p_parm.resx;
	parm_host.resy = p_parm.resy;
	parm_host.targetImage = p_parm.targetImage_d;
	cudaMemcpyToSymbol(PARM, &parm_host, sizeof(parm_host));

	// 64 = Anzahl von Blocks, 256 = Anzahl von Threads per Block

	vignetting<<< 64, 256 >>>(p_parm.colorImage_d);

	//brightness<<< 64, 256 >>>(p_parm.colorImage_d);
	//cutilCheckMsg("Kernel execution failed!");	
}
