#ifndef _ASSIMP_LOADER_H_
#define _ASSIMP_LOADER_H_

// #include "imba/mesh.h"
// #include "imba/float3.h"
// #include "imba/scene.h"
#include "path.h"

#include <iostream>

#include <float.h>
#include <string>
#include <vector>
#include <unordered_map>
#include <fstream>
#include <optixu/optixu_aabb.h>
#include <optixu/optixpp_namespace.h>
#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>
#include <optixu/optixu_vector_types.h>

// assimp include files. These three are usually needed.
#include "assimp/Importer.hpp"	//OO version Header!
#include "assimp/postprocess.h"
#include "assimp/scene.h"
#include "assimp/DefaultLogger.hpp"
#include "assimp/LogStream.hpp"
#include "assimp/cimport.h"

using namespace optix;


struct MatParams {
    //    optix::float3 emissive;
    //    optix::float3 reflectivity;
    optix::float3 kd;
    optix::float3 ks;
    float phong_exp;
    float alpha;
    optix::float3 tf;
    float ni;
    //    int    illum;
    optix::TextureSampler ambient_map;
    optix::TextureSampler diffuse_map;
    optix::TextureSampler specular_map;
    optix::TextureSampler dissolve_map;
    optix::TextureSampler bump_map;
};

    
class AssimpLoader {
public:
    AssimpLoader() {
    	_material = new optix::Material[ENUM_NUM_MAT_TYPES];
    	_intersect_program = 0;
    	_bbox_program = 0;
    }

    bool load_file(optix::Context context, const Path& path, optix::Group geometryGroup, const optix::Matrix4x4 &transform );
	

	optix::Aabb getSceneBBox() {
		return scene_bbox_;
	}

private:
	std::string _pathname;
    std::string _filename;
    optix::Context _context;
    optix::GeometryGroup _geometrygroup;
    optix::Buffer _vbuffer;
    optix::Buffer _nbuffer;
    optix::Buffer _tbuffer;
    optix::Material *_material;
    optix::Program _intersect_program;
    optix::Program _bbox_program;
    optix::Buffer _light_buffer;
    bool _have_default_material;
    bool _force_load_material_params;
    const char *_ASBuilder;
    const char *_ASTraverser;
    bool _large_geom;
    std::vector<MatParams> _material_params;
    Path *_path;


    enum enum_material_type {

    ENUM_MAT_TYPE_UNIVERSAL = 0,
    ENUM_MAT_TYPE_DIFFUSE = 1,
    ENUM_MAT_TYPE_MIRROR,
    ENUM_MAT_TYPE_GLOSSY,
    ENUM_MAT_TYPE_GLASS,

    ENUM_NUM_MAT_TYPES
};


	// This is material information
	struct AssimpMaterial {
		float diffuse[4];
		float ambient[4];
		float specular[4];
		float emissive[4];
		float shininess;
		int texCount;
	};

	// Create an instance of the Importer class
	Assimp::Importer importer;

	const aiScene* scene_ = NULL;

  optix::Aabb scene_bbox_;
  optix::float3 scene_center_;

  bool read_assimp(optix::Context context, optix::Group& geometryGroup, optix::Material* materials, float scale, const optix::Matrix4x4 &transform );

  void computeBBox_recursive (const  aiNode* nd, optix::Aabb &aBB,	aiMatrix4x4* trafo);
  void computeBBox();

  void createMaterial();



	 std::string generatePtxPath ( const std::string &base ) {
        std::cout << "cuda_compile_ptx_1_generated_" + base + ".ptx" << std::endl;
        return "cuda_compile_ptx_1_generated_" + base + ".ptx";
    }


};

#endif
