/*
 * Distribution2D_env_device.h
 *
 *  Created on: 29.10.2012
 *      Author: troth
 */

#ifndef __DISTRIBUTION2D_ENV_DEVICE_CU__
#define __DISTRIBUTION2D_ENV_DEVICE_CU__

#include <optix_world.h>

rtBuffer<float, 2> buffer_envFunc;
rtBuffer<float, 2> buffer_envCdf;
rtBuffer<float, 1> buffer_envFuncInt;

rtBuffer<float, 1> buffer_envMargFunc;
rtBuffer<float, 1> buffer_envMargCdf;
rtDeclareVariable ( float, envMargFuncInt, , );

rtDeclareVariable ( unsigned int, envWidth, , );
rtDeclareVariable ( unsigned int, envHeight, , );

__device__ float sampleContinuousMarginal ( float p_v, float &p_pdf ) {
    //find surrounding cdf segments and offsets with binary search (original version from PBRT uses std::lower_bound)
    unsigned int n = envHeight;
    unsigned int offset = n / 2;
    uint leftIndex = offset - 1;
    uint centerIndex = offset;
    bool lGe = buffer_envMargCdf[leftIndex] >= p_v;
    bool rLt = buffer_envMargCdf[centerIndex] < p_v;
    unsigned int curIvalSize = n / 4;
    unsigned int i = log2f ( n );

    while ( ( lGe || rLt ) && i > 0 ) {
        --i;

        if ( lGe ) {
            offset -= curIvalSize;

        } else if ( rLt ) {
            offset += curIvalSize;
        }

        curIvalSize /= 2;
        leftIndex = offset - 1;
        centerIndex = offset;
        lGe = buffer_envMargCdf[leftIndex] >= p_v;
        rLt = buffer_envMargCdf[centerIndex] < p_v;
    }

    //compute offset along cdf segment
    offset = max ( 0, offset - 1 );
    leftIndex = offset - 1;
    centerIndex = offset;
    uint rightIndex = offset + 1;
    float du = ( p_v - buffer_envMargCdf[centerIndex] ) / ( buffer_envMargCdf[rightIndex] - buffer_envMargCdf[centerIndex] );
    //compute pdf for sampled offset
    p_pdf = buffer_envMargFunc[centerIndex] / envMargFuncInt;
    //return x \in [0,1) corresponding to a sample
    return ( offset + du ) / envHeight; // @TODO or has it to be buffer_envCdf.size()?
}

__device__ float sampleContinuous1D ( unsigned int p_index, float p_u, float &p_pdf ) {
    //find surrounding cdf segments and offsets with binary search (original version from PBRT uses std::lower_bound)
    unsigned int n = envWidth;
    unsigned int offset = n / 2;
    uint2 leftIndex = make_uint2 ( offset - 1, p_index );
    uint2 centerIndex = make_uint2 ( offset, p_index );
    bool lGe = buffer_envCdf[leftIndex] >= p_u;
    bool rLt = buffer_envCdf[centerIndex] < p_u;
    unsigned int curIvalSize = n / 4;
    unsigned int i = log2f ( n );

    while ( ( lGe || rLt ) && i > 0 ) {
        --i;

        if ( lGe ) {
            offset -= curIvalSize;

        } else if ( rLt ) {
            offset += curIvalSize;
        }

        curIvalSize /= 2;
        leftIndex = make_uint2 ( offset - 1, p_index );
        centerIndex = make_uint2 ( offset, p_index );
        lGe = buffer_envCdf[leftIndex] >= p_u;
        rLt = buffer_envCdf[centerIndex] < p_u;
    }

    //compute offset along cdf segment
    offset = max ( 0, offset - 1 );
    leftIndex = make_uint2 ( offset - 1, p_index );
    centerIndex = make_uint2 ( offset, p_index );
    uint2 rightIndex = make_uint2 ( offset + 1, p_index );
    float du = ( p_u - buffer_envCdf[centerIndex] ) / ( buffer_envCdf[rightIndex] - buffer_envCdf[centerIndex] );
    //compute pdf for sampled offset
    p_pdf = buffer_envFunc[centerIndex] / buffer_envFuncInt[p_index];
    //return x \in [0,1) corresponding to a sample
    return ( offset + du ) / envWidth;; // @TODO or has it to be buffer_envCdf.size()?
}


__device__ void sampleContinuous ( float p_u0, float p_u1, float uv[2], float &pdf ) {
    float pdfs[2];
    pdfs[0] = 1.f;
    pdfs[1] = 1.f;
    uv[1] = 0.f;
    uv[1] = sampleContinuousMarginal ( p_u1, pdfs[1] );
    int v = fminf ( envHeight - 1, fmaxf ( 0, uv[1] * envHeight ) );
    uv[0] = sampleContinuous1D ( v, p_u0, pdfs[0] );
    pdf = pdfs[0] * pdfs[1];
}


__device__ float envPdf ( float p_u, float p_v ) {
    int iu = fminf ( envWidth - 1, fmaxf ( 0, p_u * envWidth ) );
    int iv = fminf ( envHeight - 1, fmaxf ( 0, p_v * envHeight ) );

    if ( buffer_envFuncInt[iv] * envMargFuncInt == 0.f ) {
        return 0.f;
    }

    uint2 bufferIndex = make_uint2 ( iu, iv );
    return ( buffer_envFunc[bufferIndex] * buffer_envMargFunc[iv] ) / ( buffer_envFuncInt[iv] * envMargFuncInt );
}

#endif
