/*
 * KeyboardController.h
 *
 *  Created on: Jul 4, 2011
 *      Author: troth22s
 */

#ifndef KEYBOARDCONTROLLER_H_
#define KEYBOARDCONTROLLER_H_

#include <SFML/Window/Keyboard.hpp>
#include <SFML/Window/Event.hpp>
#include <map>
#include <string>
#include <iostream>


enum KeyboardFunctionalities {
    ENUM_KBDFUNC_FUNC_SPD_X_INC = 0, ENUM_KBDFUNC_FUNC_SPD_X_DEC, ENUM_KBDFUNC_FUNC_SPD_Y_INC, ENUM_KBDFUNC_FUNC_SPD_Y_DEC, ENUM_KBDFUNC_FUNC_SPD_Z_INC, ENUM_KBDFUNC_FUNC_SPD_Z_DEC, ENUM_KBDFUNC_FUNC_NUM_SAMPLES_INC, ENUM_KBDFUNC_FUNC_NUM_SAMPLES_DEC, ENUM_KBDFUNC_FUNC_RES_FACTOR_INC, ENUM_KBDFUNC_FUNC_RES_FACTOR_DEC, ENUM_KBDFUNC_FUNC_STOP_OBJECTS, ENUM_KBDFUNC_FUNC_RENDER_FULL_RES, ENUM_KBDFUNC_FUNC_SHOWSTATS, ENUM_KBDFUNC_FUNC_QUIT,

    ENUM_KBDFUNC_FUNC_SIZE
};

enum ModificationType {
    ENUM_KBDFUNC_MODTYPES_MUL_DIV = 0, ENUM_KBDFUNC_MODTYPES_ADD_SUB, ENUM_KBDFUNC_MODTYPES_TOGGLE,

    ENUM_KBDFUNC_MODTYPES_SIZE
};

/**
 * @class KeyboardController
 * @brief This class serves the purpose of supporting keyboard interaction with an SDL application.
 *
 * This class works purely with float values. They have to be recast to the desired type by the application.
 */
class KeyboardController {

  private:
    typedef void ( KeyboardController::*func_modify_value_t ) ( sf::Keyboard::Key );

    // mapped by function
    std::map<KeyboardFunctionalities, sf::Keyboard::Key> __m_funcKeyMap;
    //	std::map<KeyboardFunctionalities, int> __m_contKeyMap;

    // mapped by keys
    std::map<sf::Keyboard::Key, bool> __m_keyChecked;
	std::map<sf::Keyboard::Key, bool> __m_keyPressed;
	std::map<sf::Keyboard::Key, float> __m_valDelta;
	std::map<sf::Keyboard::Key, float *> __m_vars;
	std::map<sf::Keyboard::Key, func_modify_value_t> __m_functions;
	std::map<sf::Keyboard::Key, float> __m_minValues;
	std::map<sf::Keyboard::Key, float> __m_maxValues;
	std::map<sf::Keyboard::Key, bool> __m_oncePerKeyPress;
	std::map<sf::Keyboard::Key, bool> __m_justChanged;

    /** @brief add or subtract some delta to/from the bound variable */
	void addSubValue ( sf::Keyboard::Key p_keyIdentifier );

    /** @brief multiply or divide the bound variable with/by some delta */
	void mulDivValue ( sf::Keyboard::Key p_keyIdentifier );

    /** @brief toggle the bound variable between its min and max value */
	void toggleValue ( sf::Keyboard::Key p_keyIdentifier );

  public:

    KeyboardController();

    ~KeyboardController();

    /** @brief map one of the functionalities to an arbitrary key */
    void mapFunctionality ( sf::Keyboard::Key p_keyIdentifier, KeyboardFunctionalities p_functionality, ModificationType p_modtype, bool p_oncePerKeyPress, float *p_var, float p_min, float p_max, float p_delta );

    void resetInput();


    void processRealTimeInput ();


    /** @brief process keyboard input by checking sfml keystates for all functions */
  //  void processEventBasedInput ( const sf::Keyboard &p_input, sf::Event &event );


    /* @brief check if a key is currently pressed */
    int keyPressed ( sf::Keyboard::Key p_keyIdentifier );

    int keyForFunctionalityPressed ( KeyboardFunctionalities p_functionality );

    /* @brief check if the key for a continuously executed functionality is currently pressed */
    int functionalityActive ( KeyboardFunctionalities p_functionality );

    int justChanged ( KeyboardFunctionalities p_functionality );
};

#endif /* KEYBOARDCONTROLLER_H_ */
