#ifndef __OUTPUT_WINDOW_H__
#define __OUTPUT_WINDOW_H__

#include <cuda_runtime.h>
#include <SFML/Graphics.hpp>
#include <RenderParameters.h>
#include <SixAxis.h>
#include <KeyboardController.h>
#include <BenchmarkValues.h>
#include <tr1/memory>
#include <string.h>

#include "TrialConfig.h"

using namespace std;

/* @brief This class is responsible for managing the SFML window containing rendering OutputWindow
 * 		  and keyboard/mouse/controller interaction
 */
class OutputWindow {

  private:

    tr1::shared_ptr<sf::RenderWindow> __m_window;
    tr1::shared_ptr<sf::Texture> __m_texture;
    tr1::shared_ptr<sf::Sprite> __m_sprite;
	sf::Font __m_font;
	
    sf::Uint8 *__m_imageData;
    SixAxis __m_sixAxis;
	

    unsigned int __m_width;
    unsigned int __m_height;
    float __m_displayFactor;

    float __m_focalPointX, __m_focalPointY;

    float __m_quit, __m_showStats;

    KeyboardController __m_kbdController;


  public:

    OutputWindow ( unsigned int p_width, unsigned int p_height, float p_displayFactor);


    ~OutputWindow();


    void eventLoop ( bool &focusChanged );


    void drawImage();


    void drawStats ( BenchmarkValues &bmv, RenderParameters *renderParameters, unsigned int numPreviousFrames, float eye_x, float eye_y, float eye_z, float direction_x, float direction_y, float direction_z );


   void saveImage ( TrialConfig p_tc, unsigned int numPreviousFrames, float4* p_buf );

    SixAxis *getSixAxis();


    void setImageData ( void *p_imageData );


    sf::RenderWindow *getWindow();

    sf::Image *getImage();

    sf::Sprite *getSprite();

    sf::Uint8 *getImageData();


    float getFocalPointX();


    float getFocalPointY();


    bool getFifarInfo_addKeyFrame() {
        return __m_sixAxis.getStoreKeyFrame();
    }

    bool getFifarInfo_writeToDisk() {
        return __m_sixAxis.getWriteKeyFramesToDisk();
    }

    KeyboardController *getKbdController();

    bool isRunning();
};


#endif