#ifndef __POSTPROCESSOR_H__
#define __POSTPROCESSOR_H__

// grip
#include <Graph/GraphRunner.h>
#include <Data/DataUtils.h>

// own includes
#include <GripBuffers.h>
#include <BufferTypes.h>

#include <cuda.h>

using namespace grip;

class PostProcessor {

  private:

    GraphRunner __m_grunner;
    std::string __m_pathToGraph;
    std::vector<BufferInfo> *__m_bufInfo;
    GripBuffers __m_gripBuffers;
    std::vector<CUdeviceptr> *__m_devicePointers;



  public:



    PostProcessor ( const std::string &p_pathToGraph, std::vector<BufferInfo> *p_bufInfo, std::vector<CUdeviceptr> *p_devicePointers );


    ~PostProcessor();



    void createGripBuffers();




    void setupGripBuffers();


    void setDevicePointers ( std::vector<CUdeviceptr> *p_devicePointers );



    void markGripBuffersDirty();


    void run();


    template<class FTYPE> FTYPE getParameter ( const std::string &p_parameterName ) {
        return __m_grunner.getParameter<FTYPE> ( p_parameterName );
    }


    template<class FTYPE> void setParameter ( const std::string &p_parameterName, FTYPE p_value, uint32_t p_size ) {
        __m_grunner.setParameter ( p_parameterName, p_value, p_size );
        __m_grunner.getGuiThread()->updateSlider ( p_parameterName );
    }


    bool parameterSetChanged ( const std::string &p_parameterSetName );
};


#endif