#ifndef __RENDERMANAGER_H__
#define __RENDERMANAGER_H__

#include <cuda_runtime.h>

#include <RenderProcess.h>
#include <BufferTypes.h>

#include <trdUtils/Fifar.h>

#include "TrialConfig.h"

#include <fstream>

class CmdLineParams {
  public:
    unsigned int width;
    unsigned int height;
    unsigned int firstGPU;
    unsigned int numGPUs;
    float displayFactor;
    std::string texturePath;
    std::string modelPath;
    std::string upVector;
    float envGamma;
    std::string camFile;
    std::string evalConfig;
};

class RenderManager {

  private:
    CmdLineParams __m_cmdLineParams;
    RenderProcess __m_renderProcess;
    RenderParameters __m_renderParameters;
    OutputBuffers __m_outputBuffers;
    std::vector<BufferInfo> __m_bufInfo;
    std::vector<CUdeviceptr> __m_accumulatedDevicePointers;
    Camera __m_camera;
    trd::Fifar* __m_fifar;

    void parseCommandLineParameters ( int p_argc, char **p_argv );

    void createBufferDescriptions();


  public:
    TrialConfig tc;

    RenderManager ( int argc, char **argv, trd::Fifar* p_fifar );
    ~RenderManager();


    /* these are used for GrIP */
	void saveLightfield();
    void createAccumulatedDataDevicePointers();
    void initializeRenderProcess();
    void loadScene();
    void setupCamera();
    void updateCameraAndParameters();
    void refocusCamera ( float p_depthSampleX, float p_depthSampleY );
    void validateAndCompile();
    void render ( std::ofstream& logfile, unsigned int p_oldKeyFrameIndex, unsigned int &p_numPreviousFrames, unsigned int &p_frame, unsigned int p_maxSamples, float p_lfWriteProb, unsigned int p_calcReconstruction );
    void setMinDepthMode();
    std::vector<CUdeviceptr> *getAccumulatedDevicePointers();
    std::vector<BufferInfo> *getBufferInfo();
    CmdLineParams getCmdLineParams();
    RenderProcess getRenderProcess();
    RenderParameters *getRenderParameters();
    Camera *getCamera();
    float4* writeOutbufToHost();


};


#endif