#ifndef __RENDER_PARAMETERS_H__
#define __RENDER_PARAMETERS_H__


#include <vector>
#include <stdint.h>
#include <string>

using namespace std;

enum enum_parameterNames_int {
    ENUM_RENDERPARAMETER_MIN_DEPTH = 0,
    ENUM_RENDERPARAMETER_MAX_DEPTH = 1,
    ENUM_RENDERPARAMETER_RES_X,
    ENUM_RENDERPARAMETER_RES_Y,
    ENUM_RENDERPARAMETER_FIRST_GPU,
    ENUM_RENDERPARAMETER_NUM_GPUS,
    ENUM_RENDERPARAMETER_MAX_SOBOL_DEPTH,
    ENUM_RENDERPARAMETER_USE_LD_SAMPLING_IMAGE,
    ENUM_RENDERPARAMETER_USE_LD_SAMPLING_LENS,
    ENUM_RENDERPARAMETER_USE_LD_SAMPLING_SUN,
    ENUM_RENDERPARAMETER_USE_LD_SAMPLING_ENVIRONMENT,
    ENUM_RENDERPARAMETER_USE_LD_SAMPLING_INTEGRATOR,
    ENUM_RENDERPARAMETER_SAMPLES_PER_LAUNCH,
    ENUM_RENDERPARAMETER_PROJECTION_TYPE,
    ENUM_RENDERPARAMETER_OUT_DIRECT,
    ENUM_RENDERPARAMETER_OUT_INDIRECT,
    ENUM_RENDERPARAMETER_OUT_TEXTURE,
    ENUM_RENDERPARAMETER_OUT_GBUFFER,
    ENUM_RENDERPARAMETER_OUT_RAYDIR,
    ENUM_NUM_RENDERPARAMETERS_INT
};


enum enum_parameterNames_float {
    ENUM_RENDERPARAMETER_SCENE_EPSILON = 0,
    ENUM_RENDERPARAMETER_SKY_BRIGHTNESS = 1,
    ENUM_RENDERPARAMETER_APERTURE_RADIUS,
    ENUM_RENDERPARAMETER_ENV_ROT_PHI,
    ENUM_RENDERPARAMETER_ENV_GAMMA,
    ENUM_RENDERPARAMETER_LIGHT_PROB_DIFFUSE,
    ENUM_RENDERPARAMETER_LIGHT_PROB_SPECULAR,
    ENUM_RENDERPARAMETER_SUN_AXIS_PHI,
    ENUM_RENDERPARAMETER_SUN_AXIS_THETA,
    ENUM_RENDERPARAMETER_SUN_BRIGHTNESS,
    ENUM_RENDERPARAMETER_SUN_SCALE,
    ENUM_RENDERPARAMETER_HFOV,

    ENUM_NUM_RENDERPARAMETERS_FLOAT
};


class RenderParameters {
  private:

    std::vector<int32_t> __m_intParameters;
    std::vector<float> __m_floatParameters;

    std::vector<string> __m_intParameterNamesOptix;
    std::vector<string> __m_floatParameterNamesOptix;
    std::vector<string> __m_intParameterNamesGrip;
    std::vector<string> __m_floatParameterNamesGrip;

  public:

    /** @brief RenderParameters are assigned names for retrieval from GrIP and setup in OptiX here. On GrIP's side this means that the parameter name used here has to be identical to the one set in GrIP's graph definition. On OptiX's side this means that names in variable definitions have to match the names used here. */
    RenderParameters();


    ~RenderParameters();


    void setParameterInt ( enum_parameterNames_int p_parameterName, float p_parameterValue );


    void setParameterFloat ( enum_parameterNames_float p_parameterName, float p_parameterValue );


    int getParameterInt ( enum_parameterNames_int p_parameterName );


    float getParameterFloat ( enum_parameterNames_float p_parameterName );


    std::string &getParameterNameGrip ( enum_parameterNames_int p_parameterName );


    std::string &getParameterNameGrip ( enum_parameterNames_float p_parameterName );


    std::string &getParameterNameOptix ( enum_parameterNames_int p_parameterName );


    std::string &getParameterNameOptix ( enum_parameterNames_float p_parameterName );





};


#endif