#ifndef __RENDERPROCESS_H__
#define __RENDERPROCESS_H__

#include <ContextWrapper.h>
#include <BufferInfo.h>
#include <OutputBuffers.h>
#include <BufferTypes.h>
#include <RenderParameters.h>
#include <Scene.h>

#include "TrialConfig.h"


class RenderProcess {

  private:
    ContextWrapper __m_contextWrapper;
    bool __m_isInitialized;

    RenderParameters *__m_renderParameters;
    OutputBuffers *__m_outputBuffers;
    std::vector<BufferInfo> *__m_bufInfo;

    Scene __m_scene;


    void preComputeSobolVectors();


    void createOptixBuffer ( enum_BufferTypes p_bufferType );


    void setDevicePointerForOptixBuffer ( enum_BufferTypes p_bufferType );


    void setOptixBufferInContext ( enum_BufferTypes p_bufferType );





    void createBuffers();




  public:

    RenderProcess();


    RenderProcess ( RenderParameters *p_renderParameters, OutputBuffers *p_outputBuffers, std::vector<BufferInfo> *p_bufInfo );


    ~RenderProcess();


    void initialize ( RenderParameters *p_renderParameters, OutputBuffers *p_outputBuffers, std::vector<BufferInfo> *p_bufInfo );


    void updateRenderParameters();


    void loadScene ( TrialConfig p_tc, const std::string &p_envPath, const std::string &p_upVector );


    void setCamera ( Camera &p_camera );


    ContextWrapper &getContextWrapper();


    Scene &getScene();

};

#endif