#ifndef __SCENE_H__
#define __SCENE_H__

#include <optixu/optixpp_namespace.h>
#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>
#include <optixu/optixu_aabb_namespace.h>

#include <Distribution2D.h>

#include <iostream>

#include <TrialConfig.h>

class Scene {

  private:
    optix::Context __m_context;
    optix::Group __m_sceneGroup;
    optix::Group __m_lightGroup;
    optix::Group __m_topGroup;
    optix::Transform __m_sceneTransform;
    optix::float3 __m_sceneSize;
    float __m_maxSize;

    optix::float3 __m_directionalLight_direction;
    optix::float3 __m_directionalLight_fluxRGB;

    bool __m_isInitialized;

    optix::Aabb normalizedSceneBB;
    optix::Aabb uniformSceneBB;

  public:

    Scene();


    Scene ( optix::Context p_context );


    void initialize ( optix::Context p_context );

    ~Scene();


    void setGroups();


    void loadModel ( const std::string &p_modelPath, const std::string &p_upVector );


    void createLights(TrialConfig p_tc);


    void setSun ( optix::Context &p_context, std::string &p_brightnessName, std::string &p_scaleName, float p_phi, float p_theta, float p_sunBrightness, float p_sunScale );


    void setEnvironmentMap ( const std::string &p_texturePath, float p_envGamma );


    optix::Group getLightGroup();

    optix::Group getSceneGroup();

    optix::Transform getSceneTransform();

    optix::float3 getSceneSize();

    float getMaxSceneExtent();

};


#endif