#ifndef __SIXAXIS_H__
#define __SIXAXIS_H__


#include <SFML/Window/Event.hpp>

#include <vector>


using namespace std;

enum sfml_sixaxis_axes {

    ENUM_SIXAXIS_LEFT_ANALOG_HORIZONTAL = 0,
    ENUM_SIXAXIS_LEFT_ANALOG_VERTICAL = 1,

    ENUM_SIXAXIS_RIGHT_ANALOG_HORIZONTAL = 4,
    ENUM_SIXAXIS_RIGHT_ANALOG_VERTICAL = 5

};

enum sfml_sixaxis_axis_muliplicator {

    ENUM_SIXAXIS_LEFT_ANALOG_LEFT = -1,
    ENUM_SIXAXIS_LEFT_ANALOG_RIGHT = 1,

    ENUM_SIXAXIS_LEFT_ANALOG_UP = -1,
    ENUM_SIXAXIS_LEFT_ANALOG_DOWN = 1,

    ENUM_SIXAXIS_RIGHT_ANALOG_LEFT = -1,
    ENUM_SIXAXIS_RIGHT_ANALOG_RIGHT = 1,

    ENUM_SIXAXIS_RIGHT_ANALOG_UP = -1,
    ENUM_SIXAXIS_RIGHT_ANALOG_DOWN = 1

};

// enum sfml_sixaxis_buttons {

//     ENUM_SIXAXIS_SELECT = 0,
//     ENUM_SIXAXIS_L3 = 1,
//     ENUM_SIXAXIS_R3 = 2,
//     ENUM_SIXAXIS_START = 3,
//     ENUM_SIXAXIS_DPAD_UP = 4,
//     ENUM_SIXAXIS_DPAD_RIGHT = 5,
//     ENUM_SIXAXIS_DPAD_DOWN = 6,
//     ENUM_SIXAXIS_DPAD_LEFT = 7,
//     ENUM_SIXAXIS_L2 = 8,
//     ENUM_SIXAXIS_R2 = 9,
//     ENUM_SIXAXIS_L1 = 10,
//     ENUM_SIXAXIS_R1 = 11,
//     ENUM_SIXAXIS_TRIANGLE = 12,
//     ENUM_SIXAXIS_CIRCLE = 13,
//     ENUM_SIXAXIS_X = 14,
//     ENUM_SIXAXIS_SQUARE = 15,
//     ENUM_SIXAXIS_MENU = 16,

//     ENUM_SIXAXIS_NUM_BUTTONS

// };


enum sfml_sixaxis_buttons {

    ENUM_SIXAXIS_SELECT = 8,
    ENUM_SIXAXIS_L3 = 11,
    ENUM_SIXAXIS_R3 = 12,
    ENUM_SIXAXIS_START = 9,
    ENUM_SIXAXIS_DPAD_UP = 13,
    ENUM_SIXAXIS_DPAD_RIGHT = 16,
    ENUM_SIXAXIS_DPAD_DOWN = 14,
    ENUM_SIXAXIS_DPAD_LEFT = 15,
    ENUM_SIXAXIS_L2 = 6,
    ENUM_SIXAXIS_R2 = 7,
    ENUM_SIXAXIS_L1 = 4,
    ENUM_SIXAXIS_R1 = 5,
    ENUM_SIXAXIS_TRIANGLE = 2,
    ENUM_SIXAXIS_CIRCLE = 1,
    ENUM_SIXAXIS_X = 0,
    ENUM_SIXAXIS_SQUARE = 3,
    ENUM_SIXAXIS_MENU = 10,

    ENUM_SIXAXIS_NUM_BUTTONS

};

enum enum_controller_vals {
    ENUM_CONTROLLER_VAL_FORWARD = 0,
    ENUM_CONTROLLER_VAL_UP = 1,
    ENUM_CONTROLLER_VAL_STRAFE,
    ENUM_CONTROLLER_VAL_PHI,
    ENUM_CONTROLLER_VAL_THETA,
    ENUM_CONTROLLER_VAL_ROLL,

    ENUM_NUM_CONTROLLER_VALS
};



class SixAxis {

  private:

    float __m_controllerSensitivity_movement;
    float __m_controllerSensitivity_rotation;

    vector<bool> __m_buttonsActive;




    float __m_controllerVals[ENUM_NUM_CONTROLLER_VALS];

    bool __m_storeKeyFrame;

    bool __m_writeKeyFramesToDisk;



    void handleAnalogSticks ( sf::Event &p_event );


    void handleButtonPress ( sf::Event &p_event );


    void handleButtonRelease ( sf::Event &p_event );




  public:


    SixAxis();


    ~SixAxis();


    bool isButtonPressed ( sfml_sixaxis_buttons p_button );


    void setButtonActive ( sfml_sixaxis_buttons p_button );

    void resetTriggers();

    bool setButtonInactive ( sfml_sixaxis_buttons p_button );

    void handleInteraction ( sf::Event &p_event );



    void setControllerSensitivity_movement ( float p_value );

    void setControllerSensitivity_rotation ( float p_value );


    float getValue ( enum_controller_vals p_controllerVal );

    bool getStoreKeyFrame();

    bool getWriteKeyFramesToDisk();


    bool movement();
};


#endif