/*
 * CubeLight.cu
 *
 *  Created on: Jun 29, 2011
 *      Author: troth22s
 */

#ifndef SPHERELIGHT_CU_
#define SPHERELIGHT_CU_

#ifndef __CUDACC__
#include <optixu/optixpp_namespace.h>
#include <PrimitiveCreators.h>
#endif

#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>

#include <optixu/optixu_vector_types.h>

#include <random.h>

#include <LightSource.cu>

#include <constants.h>

class LightSourceSphere: public LightSource {
  private:

    virtual float calcArea() {
        return FOURPI * getSize() * getSize();
    }

  public:

    /* @brief returns a random surface point */
    __host__ __device__ optix::float3 getRandomSurfacePoint ( float p_u1, float p_u2 ) {
        float t = p_u1 * TWOPI;
        float z = p_u2 * 2.f - 1.f;
        optix::float3 pointDirection = optix::make_float3 ( sqrtf ( 1.f - z * z ) * cos ( t ), sqrtf ( 1 - z * z ) * sin ( t ), z );
        pointDirection = pointDirection * getSize();
        optix::float3 pointOnSurface = getPosition() + pointDirection;
        return pointOnSurface;
    }



    /** host specific code is set here (virtual functions etc.) */
    #ifndef __CUDACC__
  private:
    static optix::Geometry __sm_unitSphere;

  public:
    LightSourceSphere ( optix::Context p_context ) :
        LightSource ( p_context ) {
        if ( !__sm_unitSphere ) {
            __sm_unitSphere = createSphere ( p_context, 0, 0, 0, 1 );
        }
    }

    optix::Geometry getUnitGeometry() {
        return __sm_unitSphere;
    }
    #endif

};


/** host specific static variables are set here */
#ifndef __CUDACC__
optix::Geometry LightSourceSphere::__sm_unitSphere = NULL;
#endif

#endif /* CUBELIGHT_CU_ */
