/*
 * common_structs.h
 *
 *  Created on: May 11, 2011
 *      Author: troth22s
 */

#ifndef COMMON_STRUCTS_H_
#define COMMON_STRUCTS_H_

#include <optixu/optixu_vector_types.h>

/* sampling dimensions used for requesting samples from the sobol sequence */
enum enum_dimensions {
    ENUM_DIMENSION_IMAGE_U = 0,
    ENUM_DIMENSION_IMAGE_V = 1,
    ENUM_DIMENSION_LENS_U = 2,
    ENUM_DIMENSION_LENS_V = 3,

    ENUM_DIMENSION_BASE_NUM = 4,

    ENUM_DIMENSION_BSDF_TRANSPARENCY = 0,
    ENUM_DIMENSION_BSDF_FRESNEL = 1,
    ENUM_DIMENSION_BSDF_REFLECTIVITY = 2,

    ENUM_DIMENSION_BSDF_CHOOSE = 3,
    ENUM_DIMENSION_BSDF_U = 4,
    ENUM_DIMENSION_BSDF_V = 5,

    ENUM_DIMENSION_LIGHT_SOURCE_CHOOSE = 6,

    ENUM_DIMENSION_ENV_MIS_CHOOSE = 7,
    ENUM_DIMENSION_ENV_U = 8,
    ENUM_DIMENSION_ENV_V = 9,

    ENUM_DIMENSION_SUN_MIS_CHOOSE = 10,
    ENUM_DIMENSION_SUN_U = 11,
    ENUM_DIMENSION_SUN_V = 12,

    ENUM_DIMENSION_TERMINATE = 13,
    ENUM_DIMENSION_BOUNCE_NUM = 14

};

enum RayType {
    ENUM_RAYTYPE_PT = 0u, ENUM_RAYTYPE_OCCLUSION, ENUM_RAYTYPE_SHADOW, ENUM_RAYTYPE_SIZE
};


enum GeoType {
    ENUM_GEOTYPE_SPHERE = 0u, ENUM_GEOTYPE_BOX, ENUM_GEOTYPE_SIZE
};


enum EntryPoints {
	ENUM_ENTRYPOINT_PT_LF = 0,
	ENUM_ENTRYPOINT_REFOCUS = 1,
	ENUM_ENTRYPOINT_PT,
	
	ENUM_NUM_ENTRYPOINTS
};

/* occlusion rays can not be used to measure how they hit a light source */
struct PerRayData_occlusion {
    bool occluded;
    unsigned int seed;
    optix::float3 throughput;
};


/* shadow rays should be used when sampling intersectable or anisotropic (?) light sources */
struct PerRayData_shadow {
    float emittance;
    float hitDist;
    bool occluded;
    unsigned int queryIdentifier;
    unsigned int seed;
    optix::float3 throughput;
};


/* normal path tracing rays carrying all the necessary information */
struct PerRayData_pt {
    optix::float3 radiance;
    optix::float3 origin;
    optix::float3 direction;
    optix::float3 normal;
    optix::float3 normal_geometric;
    optix::float3 texCol;
    float dotProduct;
    float pdf;
    float hitDist;
    optix::float3 f;
    bool directIllumination;
    unsigned int seed;
    unsigned int uniformSeed;
    int depth;
    bool done;
    unsigned int realFrame;
    unsigned int numCastRays;
    bool background;
	float shininess;
	float kd;
    unsigned int frame;

};


#endif /* COMMON_STRUCTS_H_ */
