/*
 * random.h
 *
 *  Created on: May 16, 2011
 *      Author: troth22s
 */

#ifndef RANDOM_H_
#define RANDOM_H_

#include <optixu_math_namespace.h>
#include <stdint.h>

template<unsigned int N> __host__ __device__ __inline__ unsigned int tea ( unsigned int val0, unsigned int val1 ) {
    unsigned int v0 = val0;
    unsigned int v1 = val1;
    unsigned int s0 = 0;

    for ( unsigned int n = 0; n < N; n++ ) {
        s0 += 0x9e3779b9;
        v0 += ( ( v1 << 4 ) + 0xa341316c ) ^ ( v1 + s0 ) ^ ( ( v1 >> 5 ) + 0xc8013ea4 );
        v1 += ( ( v0 << 4 ) + 0xad90777d ) ^ ( v0 + s0 ) ^ ( ( v0 >> 5 ) + 0x7e95761e );
    }

    return v0;
}

__host__ __device__ __inline__ unsigned int rot_seed ( unsigned int seed, unsigned int frame ) {
    return seed ^ frame;
}

// Generate random unsigned int in [0, 2^24)
__host__ __device__ __inline__ unsigned int lcg ( unsigned int &prev ) {
    const unsigned int LCG_A = 1664525u;
    const unsigned int LCG_C = 1013904223u;
    prev = ( LCG_A * prev + LCG_C );
    return prev & 0x00FFFFFF;
}

__host__ __device__ __inline__ unsigned int lcg2 ( unsigned int &prev ) {
    prev = ( prev * 8121 + 28411 ) % 134456;
    return prev;
}

// Generate random float in [0, 1)
__host__ __device__ __inline__ float rnd ( unsigned int &prev ) {
    return ( ( float ) lcg ( prev ) / ( float ) 0x01000000 );
}




__device__ __inline__ void reverseBits ( unsigned int &p_n ) {
    p_n = ( p_n << 16 ) | ( p_n >> 16 );
    p_n = ( ( p_n & 0x00ff00ff ) << 8 ) | ( ( p_n & 0xff00ff00 ) >> 8 );
    p_n = ( ( p_n & 0x0f0f0f0f ) << 4 ) | ( ( p_n & 0xf0f0f0f0 ) >> 4 );
    p_n = ( ( p_n & 0x33333333 ) << 2 ) | ( ( p_n & 0xcccccccc ) >> 2 );
    p_n = ( ( p_n & 0x55555555 ) << 1 ) | ( ( p_n & 0xaaaaaaaa ) >> 1 );
}


/*#define LD_SCRAMBLE_OFFSET_IMAGE_1            0x5faa1a78
#define LD_SCRAMBLE_OFFSET_IMAGE_2          0x003e7b82
#define LD_SCRAMBLE_OFFSET_LENS_1           0x7c4798bf
#define LD_SCRAMBLE_OFFSET_LENS_2           0xc6f244be
#define LD_SCRAMBLE_OFFSET_SUN_1            0xedc5f922
#define LD_SCRAMBLE_OFFSET_SUN_2            0x0dbdd22d
#define LD_SCRAMBLE_OFFSET_ENV_1            0x5503f713
#define LD_SCRAMBLE_OFFSET_ENV_2            0x71533f65
#define LD_SCRAMBLE_OFFSET_INTEGRATOR_1     0xbb19e05b
#define LD_SCRAMBLE_OFFSET_INTEGRATOR_2 0xb844b1b7*/


__device__ __inline__ float VanDerCorput ( unsigned int p_n, unsigned int p_scramble ) {
    reverseBits ( p_n );
    p_n ^= p_scramble;
    return ( ( p_n >> 8 ) & 0xffffff ) / float ( 1 << 24 );
}


__device__ __inline__ float Sobol2 ( unsigned int p_n, unsigned int p_scramble ) {
    for ( uint32_t v = 1u << 31; p_n != 0; p_n >>= 1, v ^= v >> 1 ) {
        if ( p_n & 0x1 ) {
            p_scramble ^= v;
        }
    }

    return ( ( p_scramble >> 8 ) & 0xffffff ) / float ( 1 << 24 );
}


__device__ __inline__ float LarPil ( unsigned int p_n, unsigned int p_scramble ) {
    for ( uint32_t v = 1u << 31; p_n != 0; p_n >>= 1, v |= v >> 1 ) {
        if ( p_n & 0x1 ) {
            p_scramble ^= v;
        }
    }

    return ( float ) ( p_scramble ) / ( float ) ( 0x100000000LL );
}





// taken from ompf.org
// jenkins!
__device__ __inline__ uint hash ( uint aNumber ) {
    aNumber = ( aNumber + 0x7ed55d16 ) + ( aNumber << 12 );
    aNumber = ( aNumber ^ 0xc761c23c ) ^ ( aNumber >> 19 );
    aNumber = ( aNumber + 0x165667b1 ) + ( aNumber << 5 );
    aNumber = ( aNumber + 0xd3a2646c ) ^ ( aNumber << 9 );
    aNumber = ( aNumber + 0xfd7046c5 ) + ( aNumber << 3 );
    aNumber = ( aNumber ^ 0xb55a4f09 ) ^ ( aNumber >> 16 );
    return aNumber;
}



#endif /* RANDOM_H_ */
