#pragma once

#include <pshHelper.h>
#include <algorithm>

class Data {
private:
	uint32_t* __m_data = 0;
	uint64_t* __m_valid = 0;
	uint32_t __m_numValidEntries;
	
	uint64_t __m_dimSize[3];
	uint64_t __m_dimLimits[3];
	uint64_t __m_size;

	std::vector< uint64_t > __m_validIndices;

public:
	
	
	uint64_t computeLinearIndex(uint64_t *p_index) {
		return PSHHelper::computeLinearIndex(p_index, __m_dimSize);
	}

	void compute3dIndex(uint64_t* p_nDimIndex, uint64_t p_index) {		
		PSHHelper::compute3dIndex(p_nDimIndex, p_index, __m_dimSize);
	}



	void erase() {
		
		uint64_t sizeValid = ceilf(__m_size / 64.f);

		memset(__m_data, 0, __m_size * sizeof(uint32_t));
		memset(__m_valid, 0, sizeValid * sizeof(uint64_t));

		__m_validIndices.clear();
		__m_numValidEntries = 0;

	}

	uint64_t getValidIndex(uint64_t p_i) {
		return __m_validIndices[p_i];
	}

	Data(uint64_t p_numDimensions, uint64_t p_dimSize[3]) {

		__m_size = 1;

		for (int i = 0; i < 3; ++i) {
			__m_size *= p_dimSize[i];
		}
		
		// __m_dimSize = new int[3];

		memcpy(__m_dimSize, p_dimSize, sizeof(uint64_t) * 3);
		
		int sizeValid = ceilf(__m_size / 64.f);

		__m_data = new uint32_t[__m_size];
		__m_valid = new uint64_t[sizeValid];
			
		erase();		

		// compute dimensional borders
		uint64_t totalSize = 1;

		for (int i = 0; i < 3; ++i) {
			totalSize *= __m_dimSize[i];
		}

		// __m_dimLimits = new int[3];
		uint64_t curLimit = 1;

		for (int i = 0; i < 3; ++i) {
			__m_dimLimits[i] = curLimit;
			curLimit *= __m_dimSize[i];
			std::cout << "dimLimit[" << i << "]: " << __m_dimLimits[i] << std::endl;
		}

	}



	void setValid(uint64_t p_index) {

		uint64_t index = p_index / 64;
		__m_valid[index] |= 1 << (64 - p_index % 64);

	}



	bool isValid(uint64_t p_index) {

		uint64_t index = p_index / 64;
		int localValid = __m_valid[index];

		return localValid & (1 << (64 - p_index % 64));

	}


	void setValid(uint64_t *p_index) {

		uint64_t index = computeLinearIndex(p_index);
		setValid(index);

	}



	bool isValid(uint64_t *p_index) {

		uint64_t index = computeLinearIndex(p_index);
		return isValid(index);

	}

	void initialize(float p_density) {
		
		uint64_t numValidEntries = 0;
		std::cout << "Initializing data" << std::endl;
		for (uint64_t i = 0; i < __m_size * p_density; ++i) {

			if (i % 10000 == 0) std::cout << "\r" << i << " / " << (int)(__m_size * p_density);
			float u = (float) rand() / RAND_MAX;
			uint64_t index = (uint64_t)(u * (__m_size - 1));
			if (!isValid(index)) {
				__m_data[index] = index;
				++numValidEntries;
				setValid(index);
				__m_validIndices.push_back(index);
			}

		}

		std::sort(__m_validIndices.begin(), __m_validIndices.end());
		// for (int i = 0; i < __m_validIndices.size(); ++i) {
		// 	std::cout << __m_validIndices[i] << std::endl;
		// }
		// for (int i = 0; i < __m_size; ++i) {
		// 	if (i % 10000 == 0) std::cout << "\r" << i << " / " << __m_size;
		// 	float u = (float) rand() / RAND_MAX;

		// 	if (u < p_density) {
		// 		__m_data[i] = i;
		// 		++numValidEntries;
		// 		setValid(i);
		// 	} 

		// }

		__m_numValidEntries = numValidEntries;

	}


	int getNumValidEntries() {

		return __m_numValidEntries;

	}


	void printData() {

		std::cout << "Data::printData(): " << std::endl;

		for (uint64_t i = 0; i < __m_size; ++i) {
			
			if (isValid(i)) {
				std::cout << __m_data[i];
			} else {
				std::cout << "-";
			}

			std::cout << "\t";
			
			for (uint64_t j = 1; j < 3; ++j) {

				if ((i + 1) % __m_dimLimits[j] == 0) {
						std::cout << std::endl;
				}

			}

		}

		std::cout << std::endl;

	}


	uint64_t getSize() {
		return __m_size;
	}


	void printValid() {

		std::cout << "Data::printValid(): " << std::endl;

		for (uint64_t i = 0; i < __m_size; ++i) {
			std::cout << isValid(i) << "\t";

			for (uint64_t j = 1; j < 3; ++j) {

				if ((i + 1) % __m_dimLimits[j] == 0) {
						std::cout << std::endl;
				}

			}

		}

		std::cout << std::endl;
	}


};