#pragma once

#include <pshHelper.h>
#include <Data.h>
#include <OffsetTable.h>


class HashTable {

private:
	Data* __m_data = 0;
	OffsetTable *__m_offTab = 0;

	uint64_t __m_dimSize[3];

	uint64_t __m_numValidEntries = 0;

	uint64_t __m_mbar = 1;
	uint64_t __m_m = 1;

	uint64_t __m_rbar = 1;
	uint64_t __m_r = 1;

	uint64_t __m_offsetScale = 1;

	uint64_t computeLinearIndex(uint64_t *p_index) {
		return PSHHelper::computeLinearIndex(p_index, __m_dimSize);
	}

	void compute3dIndex(uint64_t* p_Index, uint64_t p_index) {		
		PSHHelper::compute3dIndex(p_Index, p_index, __m_dimSize);
	}



	uint64_t determineTableSize(uint64_t p_n, uint64_t p_d, float p_mul) {

		return ceil(powf(p_mul * p_n, 1.f / p_d));

	}


	void optimizeTableSize(uint64_t& p_mbar, uint64_t& p_rbar) {
		uint64_t cf = 1;

		while((( cf = PSHHelper::commonFactor(__m_mbar, __m_rbar)) != 1) 
		      || p_mbar % p_rbar == 1 
		      || p_mbar % p_rbar == p_rbar - 1) {

			if (cf != 1) {
				// std::cout << "COMMON FACTOR: " << cf << std::endl;
			}  else {
				std::cout << "mbar % rbar: " << (p_mbar % p_rbar) << std::endl;
			}
			
			++p_rbar;

		}

		__m_r = powf(__m_rbar, 3);
	}


public:

	float getCompressionRatio(float p_rawDataSize, float p_offsetSize) {
		return (double)((__m_m) + (__m_r)) / (__m_data->getSize());
	}


	#define RAW_DATA_SIZE 12
	#define OFFSET_SIZE 4



	void allocateTables() {
		__m_numValidEntries = __m_data->getNumValidEntries();

		__m_offTab = new OffsetTable();

		std::cout << "Valid entries: " << __m_numValidEntries << std::endl;
		std::cout << "Determining table sizes" << std::endl;
		__m_mbar = fmaxf(2, determineTableSize(__m_numValidEntries, 3, 1.01f));
		__m_m = powf(__m_mbar, 3);

		__m_rbar = fmaxf(2, __m_offTab->determineTableSize_fast(__m_numValidEntries, 3));
		__m_r = powf(__m_rbar, 3);
		std::cout << "Done!" << std::endl;

		std::cout << "HashTable::HashTable(...) Table size before optimization:" << std::endl;
		std::cout << "mbar, m, rbar, r: " << __m_mbar << ", " << __m_m << ", " << __m_rbar << ", " << __m_r << std::endl;
		std::cout << "compression ratio: " << getCompressionRatio(RAW_DATA_SIZE, OFFSET_SIZE) << std::endl;

		// ensure that mbar and rbar are coprime
		std::cout << "Optimizing table size" << std::endl;
		optimizeTableSize(__m_mbar, __m_rbar);
		std::cout << "done!" << std::endl;

		std::cout << "HashTable::HashTable(...) Table size after optimization:" << std::endl;
		std::cout << "mbar, m, rbar, r: " << __m_mbar << ", " << __m_m << ", " << __m_rbar << ", " << __m_r << std::endl;
		std::cout << "compression ratio: " << getCompressionRatio(RAW_DATA_SIZE, OFFSET_SIZE) << std::endl;
	}


		
#define PRIME1 1
#define PRIME2 1


	uint64_t h0(uint64_t p_index) {
		return p_index % __m_m;
	}


	uint64_t h1(uint64_t p_index) {
		return p_index % __m_r;
	}


	uint64_t h(uint64_t p_index) {
		// return h0(p_index) + __m_offTab.get(h1[p_index]);
	}


	int determineOffsetDistribution(std::vector< std::vector<uint64_t> >& p_offsetBuckets, uint64_t* p_entryCount) {
		std::cout << "Determining offset distribution..." << std::endl;	
 		
		int maxEntries = 0;

 		for (uint64_t i = 0; i < __m_data->getNumValidEntries(); ++i) {

 			uint64_t validIndex = __m_data->getValidIndex(i);
			uint64_t offsetIndex = h1(validIndex);

			++(p_entryCount[offsetIndex]);
			if (p_entryCount[offsetIndex] > maxEntries) {
				maxEntries = p_entryCount[offsetIndex];
			}

			p_offsetBuckets[offsetIndex].push_back(validIndex); // this is wrong! 

		}

		return maxEntries;
	}


	void buildEntryCountBuckets(std::vector< std::vector<uint64_t> >& p_offsetBuckets, std::vector< std::vector<uint64_t> >& p_entryCountBuckets) {
		for (uint64_t i = 0; i < p_offsetBuckets.size(); ++i) {

			uint64_t bucketIndex = p_offsetBuckets[i].size();
			for (uint64_t j = 0; j < p_offsetBuckets[i].size(); ++j) {
				p_entryCountBuckets[bucketIndex].push_back(p_offsetBuckets[i][j]); // push data indices
			}

		}
	}


	HashTable(Data* p_data) {
		
		__m_data = p_data;
	
		allocateTables();
				
		// determine how many hash table entries fall into each offset entry
		uint64_t* entryCount = new uint64_t[__m_r];
		memset(entryCount, 0, sizeof(uint64_t) * __m_r);
		std::vector< std::vector<uint64_t> > offsetBuckets(__m_r);
		int maxEntries = determineOffsetDistribution(offsetBuckets, entryCount);

		// @TODO maxEntries should be limited by some extent to avoid overcomplicated hash computation! (maybe limit variance?) 

		std::cout << "Building entryCountBuckets..." << std::endl;
		// use maxEntries for number of buckets in bucket sort
		std::vector< std::vector<uint64_t> > entryCountBuckets(maxEntries + 1);

		std::cout << "Maximum entries: " << maxEntries << std::endl;
		buildEntryCountBuckets(offsetBuckets, entryCountBuckets);
		std::cout << "DONE!" << std::endl;

		float avgSize = 0.f;
		for (uint64_t i = 0; i < offsetBuckets.size(); ++i) {
			avgSize += (float)offsetBuckets[i].size() / offsetBuckets.size();
		}
		std::cout << "Average entries" << avgSize << std::endl;

		
		uint64_t* hashTable = new uint64_t[__m_m];
		bool* occupiedTable = new bool[__m_m];
		uint64_t* offsetTable = new uint64_t[__m_r];

		memset(hashTable, 0, sizeof(uint64_t) * __m_m);
		memset(occupiedTable, 0, sizeof(bool) * __m_m);
		memset(offsetTable, 0, sizeof(uint64_t) * __m_r);

		bool offsetValid = true;
		// now hash ordered by buckets
		for (uint64_t i = entryCountBuckets.size() - 1; i > 0; --i) {
			std::cout << i << " / " << 0 << std::endl;
			for (uint64_t j = 0; j < entryCountBuckets[i].size(); j += i) {
				// std::cout << j << " / " << entryCountBuckets[i].size() << std::endl;
				// tuples of i points have to be hashed and assigned an offset one after another

				uint64_t* curPoints = &(entryCountBuckets[i][j]);
				
				PRINT_DEBUG("Startoffset: ");
				uint64_t startOffset = (uint64_t)(((float)rand() / RAND_MAX) * __m_m) % __m_m;
				PRINT_DEBUG(startOffset);
				PRINT_DEBUG(std::endl);

				uint64_t curOffset = startOffset;
				
				offsetValid = true;
				std::vector< uint64_t > linHashIndices;
				std::vector< uint64_t > linOffsetIndices;
 				
 				do {
 					PRINT_DEBUG(std::endl);
 					offsetValid = true;
					for (uint64_t k = 0; k < i; ++k) { //verify for each point if the offset is valid

						// PRINT_DEBUG("current points: " << i << std::endl);
						uint64_t curPoint = curPoints[k];
						uint64_t hashIndex = (curPoint + curOffset) % __m_m;
						uint64_t offsetIndex = (PRIME2 * curPoint) % __m_r;


						if (occupiedTable[hashIndex]) { // offset does not work because cell is already occupied; 
							// std::cout << hashIndex << " ALREADY OCCUPIED" << std::endl;
							offsetValid = false;
							break;

						} else {
							// std::cout << "VALID" << std::endl;
							offsetValid = true;
						}

					}


					if (offsetValid) {
						for (uint64_t k = 0; k < i; ++k) {
							uint64_t curPoint = curPoints[k];

							uint64_t hashIndex = (curPoint + curOffset) % __m_m;

							// uint64_t offsetIndex = (PRIME2 * curPoint) % __m_r;
							occupiedTable[hashIndex] = true;
						}
					} else { 
						curOffset = (curOffset + 1) % __m_m;
					}

					PRINT_DEBUG(std::endl);
					// if (j % (entryCountBuckets[i].size() / 1000 + 1) == 0) {
					// 	std::cout << i << ", " << j << " / " << entryCountBuckets[i].size() << ": " << curOffset << " / " << __m_m << std::endl;
					// }
					// if (curOffset % 10000 == 0) {
					// 	std::cout << "offset: " << curOffset << ", startOffset: " << startOffset << ", __m_m: " << __m_m << std::endl;
					// }
					// if (curOffset == startOffset && !offsetValid) {
					// 	PRINT_DEBUG("FAILED TO FIND VALID OFFSET" << std::endl);
					// 	std::cout << "FAILED TO FIND VALID OFFSET" << std::endl;
					// 	break;
					// }
				} while (!offsetValid && !(!offsetValid && curOffset == startOffset));

				if (!offsetValid) {
					std::cout << "Offset not valid! Next loop!" << std::endl;
					break;
					// if offset is valid, we can work with what we have

				}

			}	

			if (!offsetValid) {
				break;
			}
		}




	}


	~HashTable() {

	}


	

};
