#pragma once

#include <pshHelper.h>


class Data {
private:
	uint32_t* __m_data = 0;
	uint32_t* __m_valid = 0;
	uint32_t __m_numValidEntries;
	int __m_numDimensions;
	int* __m_dimSize;
	int* __m_dimLimits;
	int __m_size;

public:
	
	
	int computeLinearIndex(int *p_index) {
		PSHHelper::computeLinearIndex(p_index, __m_numDimensions, __m_dimSize);
	}

	void computeNDimensionalIndex(int* p_nDimIndex, int p_index) {		
		PSHHelper::computeNDimensionalIndex(p_nDimIndex, p_index, __m_numDimensions, __m_dimSize);
	}



	void erase() {
		
		int sizeValid = ceilf(__m_size / 32.f);

		memset(__m_data, 0, __m_size * sizeof(uint32_t));
		memset(__m_valid, 0, sizeValid * sizeof(uint32_t));

		__m_numValidEntries = 0;

	}



	Data(int p_numDimensions, int* p_dimSize) {

		__m_size = 1;

		for (int i = 0; i < p_numDimensions; ++i) {
			__m_size *= p_dimSize[i];
		}
		
		__m_numDimensions = p_numDimensions;
		__m_dimSize = new int[p_numDimensions];

		for (int i = 0; i < p_numDimensions; ++i) {
			__m_dimSize[i] = p_dimSize[i];
		}

		int sizeValid = ceilf(__m_size / 32.f);

		__m_data = new uint32_t[__m_size];
		__m_valid = new uint32_t[sizeValid];
			
		erase();		

		// compute dimensional borders
		int totalSize = 1;

		for (int i = 0; i < __m_numDimensions; ++i) {
			totalSize *= __m_dimSize[i];
		}

		__m_dimLimits = new int[__m_numDimensions];
		int curLimit = 1;

		for (int i = 0; i < __m_numDimensions; ++i) {
			__m_dimLimits[i] = curLimit;
			curLimit *= __m_dimSize[i];
			std::cout << "dimLimit[" << i << "]: " << __m_dimLimits[i] << std::endl;
		}

	}


	int getNumDimensions() {

		return __m_numDimensions;

	}


	void setValid(int p_index) {

		int index = p_index / 32;
		__m_valid[index] |= 1 << (31 - p_index % 32);

	}



	bool isValid(int p_index) {

		int index = p_index / 32;
		int localValid = __m_valid[index];

		return localValid & (1 << (31 - p_index % 32));

	}


	void setValid(int *p_index) {

		int index = computeLinearIndex(p_index);
		setValid(index);

	}



	bool isValid(int *p_index) {

		int index = computeLinearIndex(p_index);
		return isValid(index);

	}



	


	void initialize(float p_density) {
		
		int numValidEntries = 0;

		for (int i = 0; i < __m_size; ++i) {
			float u = (float) rand() / RAND_MAX;

			if (u < p_density) {
				__m_data[i] = i;
				++numValidEntries;
				setValid(i);
			} 

		}

		__m_numValidEntries = numValidEntries;

	}


	int getNumValidEntries() {

		return __m_numValidEntries;

	}


	void printData() {

		std::cout << "Data::printData(): " << std::endl;

		for (int i = 0; i < __m_size; ++i) {
			
			if (isValid(i)) {
				std::cout << __m_data[i];
			} else {
				std::cout << "-";
			}

			std::cout << "\t";
			
			for (int j = 1; j < __m_numDimensions; ++j) {

				if ((i + 1) % __m_dimLimits[j] == 0) {
						std::cout << std::endl;
				}

			}

		}

		std::cout << std::endl;

	}


	int getSize() {
		return __m_size;
	}


	void printValid() {

		std::cout << "Data::printValid(): " << std::endl;

		for (int i = 0; i < __m_size; ++i) {
			std::cout << isValid(i) << "\t";

			for (int j = 1; j < __m_numDimensions; ++j) {

				if ((i + 1) % __m_dimLimits[j] == 0) {
						std::cout << std::endl;
				}

			}

		}

		std::cout << std::endl;
	}


};