#pragma once

#include <pshHelper.h>
#include <Data.h>
#include <OffsetTable.h>


class HashTable {

private:
	Data* __m_data = 0;
	OffsetTable *__m_offTab = 0;

	int* __m_dimSize = 0;

	int __m_numDimensions = 1;
	int __m_numValidEntries = 0;

	int __m_mbar = 1;
	int __m_m = 1;

	int __m_rbar = 1;
	int __m_r = 1;

	int __m_offsetScale;

	int computeLinearIndex(int *p_index) {
		PSHHelper::computeLinearIndex(p_index, __m_numDimensions, __m_dimSize);
	}

	void computeNDimensionalIndex(int* p_nDimIndex, int p_index) {		
		PSHHelper::computeNDimensionalIndex(p_nDimIndex, p_index, __m_numDimensions, __m_dimSize);
	}



	int determineTableSize(int p_n, int p_d, float p_mul) {

		return ceilf(powf(p_mul * p_n, 1.f / p_d));

	}


	void optimizeTableSize(int& p_mbar, int& p_rbar) {
		int cf = 1;

		while((( cf = PSHHelper::commonFactor(__m_mbar, __m_rbar)) != 1) 
		      || p_mbar % p_rbar == 1 
		      || p_mbar % p_rbar == p_rbar - 1) {

			if (cf != 1) {
				// std::cout << "COMMON FACTOR: " << cf << std::endl;
			}  else {
				std::cout << "mbar % rbar: " << (p_mbar % p_rbar) << std::endl;
			}
			
			++p_rbar;

		}

		__m_r = powf(__m_rbar, __m_numDimensions);
	}


public:

	float getCompressionRatio(float p_rawDataSize, float p_offsetSize) {
		return ((__m_m * p_rawDataSize) + (__m_r * p_offsetSize * __m_numDimensions)) / (__m_data->getSize() * p_rawDataSize);
	}


	#define RAW_DATA_SIZE 12
	#define OFFSET_SIZE 4

	HashTable(Data* p_data) {
		
		__m_data = p_data;
	
		__m_numDimensions = __m_data->getNumDimensions();
		__m_numValidEntries = __m_data->getNumValidEntries();

		__m_offTab = new OffsetTable();

		__m_mbar = fmaxf(2, determineTableSize(__m_numValidEntries, __m_numDimensions, 1.01f));
		__m_m = powf(__m_mbar, __m_numDimensions);

		__m_rbar = fmaxf(2, __m_offTab->determineTableSize_fast(__m_numValidEntries, __m_numDimensions));
		__m_r = powf(__m_rbar, __m_numDimensions);

		std::cout << "HashTable::HashTable(...) Table size before optimization:" << std::endl;
		std::cout << "mbar, m, rbar, r: " << __m_mbar << ", " << __m_m << ", " << __m_rbar << ", " << __m_r << std::endl;
		std::cout << "compression ratio: " << getCompressionRatio(RAW_DATA_SIZE, OFFSET_SIZE) << std::endl;

		// ensure that mbar and rbar are coprime
		optimizeTableSize(__m_mbar, __m_rbar);

		std::cout << "HashTable::HashTable(...) Table size after optimization:" << std::endl;
		std::cout << "mbar, m, rbar, r: " << __m_mbar << ", " << __m_m << ", " << __m_rbar << ", " << __m_r << std::endl;
		std::cout << "compression ratio: " << getCompressionRatio(RAW_DATA_SIZE, OFFSET_SIZE) << std::endl;

		__m_offsetScale = ceilf(__m_mbar / 255.f); // Paper tells us to scale by ceilf(m/255.f), but this doesn't make sense for n-dimensional indices

		// for simplicity, we assume the hash table to be of equal size for all dimensions; this can (and should) be changed later
		__m_dimSize = new int[__m_numDimensions];
		int dimSizeOffset[__m_numDimensions];
		for (int i = 0; i < __m_numDimensions; ++i) {
			__m_dimSize[i] = __m_mbar;
			dimSizeOffset[i] = __m_rbar;
		}

		// determine how many hash table entries fall into each offset entry
		int* entryCount = new int[__m_r];
		for (int i = 0; i < __m_r; ++i) {
			entryCount[i] = 0;
		}

		int maxEntries = 0;

		for (int i = 0; i < __m_data->getSize(); ++i) {
			if (!__m_data->isValid(i)) {
				continue;
			}
			int nDimIndex[__m_numDimensions];
			__m_data->computeNDimensionalIndex(nDimIndex, i);

			int offsetIndex [__m_numDimensions];
			for (int j = 0; j < __m_numDimensions; ++j) {
				offsetIndex[j] = nDimIndex[j] % __m_rbar;
			}

			int linIndex = PSHHelper::computeLinearIndex(offsetIndex, __m_numDimensions, dimSizeOffset);
			// std::cout << "LinIndex: " << linIndex << std::endl;
			++(entryCount[linIndex]);
			if (entryCount[linIndex] > maxEntries) {
				maxEntries = entryCount[linIndex];
			}
			// std::cout << "nDimIndex for " << i << ": ";
			// for (int j = 0; j < __m_numDimensions; ++j) {
			// 	std::cout << nDimIndex[j] << "; ";
			// }
			// std::cout << std::endl;
		}

#ifdef DEBUG
		PSHHelper::printNDim(__m_r, entryCount, __m_numDimensions, dimSizeOffset);
		std::cout << std::endl;
#endif

		// use maxEntries for number of buckets in bucket sort
		std::vector< std::vector<int> > offsetBuckets(__m_r);

		for (int i = 0; i < __m_data->getSize(); ++i) {
			// determine hash
			if (!__m_data->isValid(i)) {
				continue;
			}
		

			// void computeNDimensionalIndex(int* p_nDimIndex, int p_index) {	
			int nDimIndex[__m_numDimensions];
			
			__m_data->computeNDimensionalIndex(nDimIndex, i);

			int offsetIndex[__m_numDimensions];
			for (int j = 0; j < __m_numDimensions; ++j) {
				offsetIndex[j] = nDimIndex[j] % __m_rbar;
			}

			int linIndex = PSHHelper::computeLinearIndex(offsetIndex, __m_numDimensions, dimSizeOffset);

			offsetBuckets[linIndex].push_back(i); // this is wrong! We have to sort by offset index first!


			// std::cout << entryCount[i] << ": +" << i << std::endl;
		}

		std::vector< std::vector<int> > entryCountBuckets(maxEntries + 1);
		for (int i = 0; i < offsetBuckets.size(); ++i) {

			int bucketIndex = offsetBuckets[i].size();
			for (int j = 0; j < offsetBuckets[i].size(); ++j) {
				entryCountBuckets[bucketIndex].push_back(offsetBuckets[i][j]);
			}

		}

		// exit(0);


#ifdef DEBUG		
		for (int i = 0; i <= maxEntries; ++i) {
			std::cout << i << ": ";
			for (int j = 0; j < entryCountBuckets[i].size(); ++j) {
				std::cout << entryCountBuckets[i][j];
				if (i > 0) {
					std::cout << (((j+1) % i == 0) ? "; " : "~");
				}

			}
			std::cout << std::endl;
		}
#endif

		
		int* hashTable = new int[__m_m];
		int* occupiedTable = new int[__m_m];
		int* offsetTable = new int[__m_r];
		
		

		memset(hashTable, 0, sizeof(int) * __m_m);
		memset(occupiedTable, 0, sizeof(int) * __m_m);
		memset(offsetTable, 0, sizeof(int) * __m_r);

		bool offsetValid = true;
		// now hash ordered by buckets
		for (int i = entryCountBuckets.size() - 1; i > 0; --i) {
			std::cout << i << " / " << 0 << std::endl;
			for (int j = 0; j < entryCountBuckets[i].size(); j += i) {
				std::cout << j << " / " << entryCountBuckets[i].size() << std::endl;
				// tuples of i points have to be hashed and assigned an offset one after another

				int* curPoints = &(entryCountBuckets[i][j]);

				int startOffset[__m_numDimensions];
				
				PRINT_DEBUG("Startoffset: ");
				for (int k = 0; k < __m_numDimensions; ++k) {
					startOffset[k] = (int)(((float)rand() / RAND_MAX) * 255.f * __m_offsetScale) % __m_mbar;
					PRINT_DEBUG(startOffset[k] << ", ");
				}
				PRINT_DEBUG(std::endl);

				int curOffset[__m_numDimensions];
				memcpy(curOffset, startOffset, sizeof(int) * __m_numDimensions);

				offsetValid = true;
				std::vector< int > linHashIndices;
				std::vector< int > linOffsetIndices;
 				
				int numStepsPerDim = __m_mbar / __m_offsetScale;
				int stepsPerDim[__m_numDimensions];
				for (int k = 0; k < __m_numDimensions; ++k) {
					stepsPerDim[k] = 0;
				}

#ifdef DEBUG
				std::cout << "Steps per dim: ";
				for (int k = 0; k < __m_numDimensions; ++k) {
					// startOffset[k] = (int)(((float)rand() / RAND_MAX) * 255.f * __m_offsetScale) % __m_mbar;
					std::cout << stepsPerDim[k] << ", ";
				}
				std::cout << std::endl;
#endif

				// PSHHelper::printNDim(__m_m, occupiedTable, __m_numDimensions, __m_dimSize);

 				do {
 					PRINT_DEBUG(std::endl);
 					offsetValid = true;
					for (int k = 0; k < i; ++k) { //verify for each point if the offset is valid

						PRINT_DEBUG("current points: " << i << std::endl);

						int curPoint = curPoints[k];
						int nDimDataIndex[__m_numDimensions];
						int nDimHashIndex[__m_numDimensions];
						int nDimOffsetIndex[__m_numDimensions];
						__m_data->computeNDimensionalIndex(nDimDataIndex, curPoint);
						
						for (int l = 0; l < __m_numDimensions; ++l) { // compute hash including offset

							nDimHashIndex[l] = nDimDataIndex[l] % __m_mbar;
							nDimHashIndex[l] += curOffset[l];
							nDimHashIndex[l] %= __m_mbar;

							nDimOffsetIndex[l] = nDimDataIndex[l] % __m_rbar;

							PRINT_DEBUG("dataIndex: " << nDimDataIndex[l] << "; ");
						}
						PRINT_DEBUG(std::endl);

						int linHashIndex = computeLinearIndex(nDimHashIndex);
						int nDimHashIndex_test[__m_numDimensions];
						computeNDimensionalIndex(nDimHashIndex_test, linHashIndex);
#ifdef DEBUG
						std::cout << "Occupying " << linHashIndex << " for point ";
						for (int l = 0; l < __m_numDimensions; ++l) {
							std::cout << nDimHashIndex[l] << "//" << nDimHashIndex_test[l] << "; ";
						}
						std::cout << std::endl;
						std::cout << "Offset ";
						for (int l = 0; l < __m_numDimensions; ++l) {
							std::cout << nDimOffsetIndex[l] << "; ";
						}


						std::cout << std::endl;
#endif

						if (occupiedTable[linHashIndex]) { // offset does not work because cell is already occupied; 
							// std::cout << "ALREADY OCCUPIED" << std::endl;
							offsetValid = false;

							// std::cout << "CLEANING UP" << std::endl;
							
							// clean up occupiedTable for all previous points
							for (int l = k - 1; l >= 0; --l) {
								int curPoint_ = curPoints[l];
								int nDimHashIndex_[__m_numDimensions];
								__m_data->computeNDimensionalIndex(nDimHashIndex_, curPoint_);
								
								// compute hash including offset for all dimensions
								for (int m = 0; m < __m_numDimensions; ++m) {
									nDimHashIndex_[m] %= __m_mbar;
									nDimHashIndex_[m] += curOffset[m];
									nDimHashIndex_[m] %= __m_mbar;
								}

								int linHashIndex_ = computeLinearIndex(nDimHashIndex_);
								// std::cout << "linIndex: " << linHashIndex << std::endl;
								occupiedTable[linHashIndex_] = false;

							}
							
							linHashIndices.clear();
							break;

						} else {
							// std::cout << "VALID" << std::endl;
							offsetValid = true;
							linHashIndices.push_back(linHashIndex);
							occupiedTable[linHashIndex] = true;
						}

					}



					if (!offsetValid) { // no valid offset found -> modify offset; note that __m_offsetScale is currently one-dimensional => square table
						for (int k = 0; k < i; ++k) { //verify for each point if the offset is valid
							int curPoint = curPoints[k];
							int nDimDataIndex[__m_numDimensions];
							int nDimHashIndex[__m_numDimensions];
							__m_data->computeNDimensionalIndex(nDimDataIndex, curPoint);
							
							// std::cout << "OFFSET: ";
							// for (int l = 0; l < __m_numDimensions; ++l) {
							// 	std::cout << curOffset[l] << ";";
							// }
							// std::cout << std::endl;
							// std::cout << "POINT: ";
							for (int l = 0; l < __m_numDimensions; ++l) { // compute hash including offset

								nDimHashIndex[l] = nDimDataIndex[l] % __m_mbar;
								nDimHashIndex[l] += curOffset[l];
								nDimHashIndex[l] %= __m_mbar;

								// nDimOffsetIndex[l] = nDimDataIndex[l] % __m_rbar;
								// std::cout << nDimHashIndex[l] << "; ";

							}
							PRINT_DEBUG("offset invalid" << std::endl);
							// std::cout << std::endl;

						}

						bool incNext = true;

						for (int k = 0; k < __m_numDimensions; ++k) {
							if (incNext) {
								
								if (stepsPerDim[k] >= numStepsPerDim) {								
									curOffset[k] = startOffset[k];
									stepsPerDim[k] = 0;
									incNext = true;
								} else {
									stepsPerDim[k]++;
									curOffset[k] += __m_offsetScale;
									curOffset[k] %= __m_mbar;
									incNext = false;
								}
							}
							// std::cout << curOffset[k] << ", ";
						}
						// std::cout << std::endl;
					}

					bool failed = true;
					PRINT_DEBUG("stepsPerDim: ");
					for (int k = 0; k < __m_numDimensions; ++k) {
						PRINT_DEBUG(stepsPerDim[k] << ", ");
						failed = failed && (stepsPerDim[k] == 0);
					}
					PRINT_DEBUG(std::endl);
					if (failed && !offsetValid) {
						PRINT_DEBUG("FAILED TO FIND VALID OFFSET" << std::endl);
						break;
					}
				} while (!offsetValid);


					// PSHHelper::printNDim(__m_m, occupiedTable, __m_numDimensions, __m_dimSize);

				if (!offsetValid) {
						// PSHHelper::printNDim(__m_m, occupiedTable, __m_numDimensions, __m_dimSize);
					break;
					// if offset is valid, we can work with what we have

				}

			}	


			if (!offsetValid) {
				break;
			}
		}




	}


	~HashTable() {

	}


	

};
