#include <Camera.h>


#include <optixu/optixpp_namespace.h>
#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>

#include <assert.h>


Camera::Camera() {
    __m_isInitialized = false;
}


Camera::Camera ( optix::float3 p_eye, optix::float3 p_up, optix::float3 p_look_at, float p_hfov, optix::float2 p_res ) {
    initialize ( &p_eye.x, &p_up.x, &p_look_at.x, p_hfov, &p_res.x );
}


void Camera::initialize ( float *p_eye_x, float *p_up_x, float *p_look_at_x, float p_hfov, float *p_res_x ) {
    optix::float3 p_eye = * ( optix::float3 * ) p_eye_x;
    optix::float3 p_up = * ( optix::float3 * ) p_up_x;
    optix::float3 p_look_at = * ( optix::float3 * ) p_look_at_x;
    optix::float2 p_res = * ( optix::float2 * ) p_res_x;
    assert ( !__m_isInitialized );
    __m_eye = p_eye;
    __m_up = p_up;
    __m_look_at = p_look_at;
    __m_direction = optix::normalize ( p_look_at - p_eye );
    __m_hfov = p_hfov;
    __m_aspect_ratio = p_res.x / p_res.y;
    __m_local_x = optix::cross ( __m_direction, __m_up );
    __m_isInitialized = true;
}


void Camera::setEye ( optix::float3 p_eye ) {
    __m_eye = p_eye;
    __m_direction = optix::normalize ( __m_look_at - p_eye );
}


void Camera::setUp ( optix::float3 p_up ) {
    __m_up = p_up;
    __m_local_x = optix::cross ( __m_direction, __m_up );
}


void Camera::setLookAt ( optix::float3 p_look_at ) {
    __m_look_at = p_look_at;
    __m_direction = optix::normalize ( p_look_at - __m_eye );
    __m_local_x = optix::cross ( __m_direction, __m_up );
}


void Camera::setEye( float p_x, float p_y, float p_z) {
    setEye(optix::make_float3(p_x, p_y, p_z));
}


void Camera::setUp( float p_x, float p_y, float p_z) {
    setUp(optix::make_float3(p_x, p_y, p_z));
}


void Camera::setLookAt( float p_x, float p_y, float p_z) {
    setLookAt(optix::make_float3(p_x, p_y, p_z));

}



void Camera::setHfov ( float p_hfov ) {
    __m_hfov = p_hfov;
}



optix::float3 Camera::getEye() 			{ return __m_eye; }
optix::float3 Camera::getUp() 			{ return __m_up; }
optix::float3 Camera::getLookAt() 		{ return __m_look_at; }
optix::float3 Camera::getDirection() 	{ return __m_direction; }
float Camera::getHfov() 				{ return __m_hfov; }
float Camera::getAspectRatio() 			{ return __m_aspect_ratio; }



void Camera::transform ( float p_theta, float p_phi, float p_roll, optix::float3 p_move, float p_time ) {
    if ( p_theta != 0.f || p_phi != 0.f || p_roll != 0.f ) {
        // rotate theta horizontally and phi vertically, calculate new direction and up vector
        optix::Matrix4x4 rotMat = optix::Matrix4x4::rotate ( p_phi * p_time, __m_up );
        rotMat = rotMat * optix::Matrix4x4::rotate ( p_theta * p_time, __m_local_x );
        rotMat = rotMat * optix::Matrix4x4::rotate ( p_roll * p_time, __m_direction );
        optix::float4 direction4 = optix::make_float4 ( __m_direction );
        optix::float4 up4 = optix::make_float4 ( __m_up );
        //direction4 = direction4 * rotMat;
        //up4 = up4 * rotMat;
        direction4 = direction4 * rotMat;
        up4 = up4 * rotMat;
        __m_direction = optix::normalize ( optix::make_float3 ( direction4 ) );
        __m_up = optix::normalize ( optix::make_float3 ( up4 ) );
        __m_local_x = optix::cross ( __m_direction, __m_up );
    }

    __m_eye = __m_eye + __m_direction * p_move.z * p_time + __m_local_x * p_move.x * p_time + __m_up * p_move.y * p_time;
    __m_look_at = __m_eye + __m_direction * p_time;
}


void Camera::transform ( SixAxis *p_sixAxis, float p_time ) {
    transform (
        p_sixAxis->getValue ( ENUM_CONTROLLER_VAL_THETA ) * M_PIf / 180.f,
        p_sixAxis->getValue ( ENUM_CONTROLLER_VAL_PHI ) * M_PIf / 180.f,
        p_sixAxis->getValue ( ENUM_CONTROLLER_VAL_ROLL ) * M_PIf / 180.f,
        optix::make_float3 (
            p_sixAxis->getValue ( ENUM_CONTROLLER_VAL_STRAFE ),
            p_sixAxis->getValue ( ENUM_CONTROLLER_VAL_UP ),
            p_sixAxis->getValue ( ENUM_CONTROLLER_VAL_FORWARD )
        ),
        p_time
    );
}




Camera::~Camera() {
}
