#include <Distribution2D.h>

#include <optixu/optixpp_namespace.h>
#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>
#include <optix_cuda_interop.h>
#include <vector>

#include <string.h>
#include <iostream>

#include <assert.h>



bool Distribution2D::buffersMapped() {
    return __m_func && __m_cdf && __m_funcInt && __m_margFunc && __m_margCdf;
}

void Distribution2D::computeIntegralForAllXi ( unsigned int p_index ) {
    assert ( buffersMapped() );
//		std::cout << "p_index: " << p_index << ", " << "width: " << width << ", height: " << height << std::endl;
//		std::cout << __m_cdf << std::endl;
    __m_cdf[p_index * ( __m_nu + 1 )] = 0.f;

    for ( int i = 1; i < __m_nu + 1; ++i ) {
//			std::cout << i << std::endl;
        __m_cdf[p_index * ( __m_nu + 1 ) + i] = __m_cdf[p_index * ( __m_nu + 1 ) + i - 1] + __m_func[p_index * __m_nu + i - 1] / __m_nu;
//			std::cout << __m_cdf[p_index * (width+1) + i] << "; ";
    }
}

void Distribution2D::transformIntegralIntoCDF ( unsigned int p_index ) {
    assert ( buffersMapped() );
    __m_funcInt[p_index] = __m_cdf[p_index * ( __m_nu + 1 ) + __m_nu];

//		std::cout << __m_funcInt[p_index] << std::endl;
    for ( int i = 1; i < __m_nu + 1; ++i ) {
        __m_cdf[p_index * ( __m_nu + 1 ) + i] /= __m_funcInt[p_index];
//			std::cout << __m_cdf[p_index * (__m_nu + 1) + i] << "; ";
    }
}

void Distribution2D::mapBuffers() {
    __m_func = ( float * ) __m_func_buffer->map();
    __m_cdf = ( float * ) __m_cdf_buffer->map();
    __m_funcInt = ( float * ) __m_funcInt_buffer->map();
    __m_margFunc = ( float * ) __m_margFunc_buffer->map();
    __m_margCdf = ( float * ) __m_margCdf_buffer->map();
}

void Distribution2D::unmapBuffers() {
    __m_func_buffer->unmap();
    __m_cdf_buffer->unmap();
    __m_funcInt_buffer->unmap();
    __m_margFunc_buffer->unmap();
    __m_margCdf_buffer->unmap();
    __m_func = 0;
    __m_cdf = 0;
    __m_funcInt = 0;
    __m_margFunc = 0;
    __m_margCdf = 0;
}


Distribution2D::Distribution2D ( optix::Context p_context, const float *p_func, unsigned int p_nu, unsigned int p_nv ) {
    __m_func = 0;
    __m_cdf = 0;
    __m_funcInt = 0;
    __m_margFunc = 0;
    __m_margCdf = 0;
    __m_context = p_context;
    std::cout << "allocating buffers with p_nv = " << p_nv << " and p_nu = " << p_nu << std::endl;
    __m_func_buffer = __m_context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT, p_nu, p_nv );
    __m_cdf_buffer = __m_context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT, p_nu + 1, p_nv );
    __m_funcInt_buffer = __m_context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT, p_nv );
    __m_margFunc_buffer = __m_context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT, p_nv );
    __m_margCdf_buffer = __m_context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT, p_nv + 1 );
    __m_nu = p_nu;
    __m_nv = p_nv;
    mapBuffers();

    for ( int v = 0; v < p_nv; ++v ) {
        memcpy ( & ( __m_func[v * p_nu] ), & ( p_func[v * p_nu] ), p_nu * sizeof ( float ) );
        computeIntegralForAllXi ( v );
        transformIntegralIntoCDF ( v );
    }

    memcpy ( __m_margFunc, __m_funcInt, p_nv * sizeof ( float ) );
    __m_margCdf[0] = 0.f;

    for ( int i = 1; i < p_nv + 1; ++i ) {
        __m_margCdf[i] = __m_margCdf[i - 1] + __m_margFunc[i - 1] / p_nv;
    }

    __m_margFuncInt = __m_margCdf[p_nv];

    for ( int i = 1; i < p_nv + 1; ++i ) {
        __m_margCdf[i] /= __m_margFuncInt;
    }

    unmapBuffers();
    __m_context["buffer_envFunc"]->set ( __m_func_buffer );
    __m_context["buffer_envCdf"]->set ( __m_cdf_buffer );
    __m_context["buffer_envFuncInt"]->set ( __m_funcInt_buffer );
    __m_context["buffer_envMargFunc"]->set ( __m_margFunc_buffer );
    __m_context["buffer_envMargCdf"]->set ( __m_margCdf_buffer );
    __m_context["envMargFuncInt"]->setFloat ( __m_margFuncInt );
    __m_context["envWidth"]->setUint ( p_nu );
    __m_context["envHeight"]->setUint ( p_nv );
}


