/*
 * Copyright (c) 2008 - 2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and proprietary
 * rights in and to this software, related documentation and any modifications thereto.
 * Any use, reproduction, disclosure or distribution of this software and related
 * documentation without an express license agreement from NVIDIA Corporation is strictly
 * prohibited.
 *
 * TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED *AS IS*
 * AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS OR IMPLIED,
 * INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY
 * SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES WHATSOEVER (INCLUDING, WITHOUT
 * LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF
 * BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS) ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGES
 */
#include <ImageLoader.h>

#include <optixu/optixu_math_namespace.h>


#include <string.h>
#include <stdlib.h>
#include <map>
#include <string>
#include <iostream>
#include <FreeImagePlus.h>


//-----------------------------------------------------------------------------
//
//  Utility functions
//
//-----------------------------------------------------------------------------


namespace {

    std::map<std::string, optix::TextureSampler> textureSamplers;

}



SUTILAPI optix::TextureSampler loadTexture ( optix::Context context,
	const std::string &filename,
	const optix::float3 &default_color ) {
    std::string key = filename;// + default_color.x;
    key += default_color.x;
    key += default_color.y;
    key += default_color.z;
	
    if ( textureSamplers.find ( key ) == textureSamplers.end() ) {
		
		optix::TextureSampler sampler = context->createTextureSampler();
        sampler->setWrapMode ( 0, RT_WRAP_REPEAT );
        sampler->setWrapMode ( 1, RT_WRAP_REPEAT );
        sampler->setWrapMode ( 2, RT_WRAP_REPEAT );
        sampler->setIndexingMode ( RT_TEXTURE_INDEX_NORMALIZED_COORDINATES );
        sampler->setReadMode ( RT_TEXTURE_READ_NORMALIZED_FLOAT );
        sampler->setMaxAnisotropy ( 1.0f );
        sampler->setMipLevelCount ( 1u );
        sampler->setArraySize ( 1u );

		fipImage input;
		input.load( filename.c_str() );
		std::cout << filename << std::endl;
		std::cout << input.getBitsPerPixel() << std::endl;
 		input.convertTo32Bits();
		
		if ( input.isValid()) {
			
			unsigned int texWidth = input.getWidth();
			unsigned int texHeight = input.getHeight();
			optix::Buffer buffer = context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_UNSIGNED_BYTE4, texWidth, texHeight );

			
			unsigned char *buffer_data = static_cast<unsigned char *> ( buffer->map() );
			unsigned char *hostBuffer = ( unsigned char* ) input.accessPixels();
				
			for ( int i = 0; i < texHeight; ++i ) {
				for ( int j = 0; j < texWidth; ++j ) {
															
						buffer_data[i * texWidth * 4 + j * 4 + 0] = hostBuffer[i * texWidth * 4 + j * 4 + 2];
						buffer_data[i * texWidth * 4 + j * 4 + 1] = hostBuffer[i * texWidth * 4 + j * 4 + 1];
						buffer_data[i * texWidth * 4 + j * 4 + 2] = hostBuffer[i * texWidth * 4 + j * 4 + 0];
						buffer_data[i * texWidth * 4 + j * 4 + 3] = hostBuffer[i * texWidth * 4 + j * 4 + 3];

					
			
				}
			}

			
			buffer->unmap();
			sampler->setBuffer ( 0u, 0u, buffer );
			sampler->setFilteringModes ( RT_FILTER_LINEAR, RT_FILTER_LINEAR, RT_FILTER_NONE );
			textureSamplers[key] = sampler;

		
			
            return sampler;
			

        } else {
// 			std::cout << "2.2" << std::endl;
			
			
            // Create buffer with single texel set to default_color
            optix::Buffer buffer = context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_UNSIGNED_BYTE4, 1u, 1u );
            unsigned char *buffer_data = static_cast<unsigned char *> ( buffer->map() );
            buffer_data[0] = ( unsigned char ) optix::clamp ( ( int ) ( default_color.x * 255.0f ), 0, 255 );
            buffer_data[1] = ( unsigned char ) optix::clamp ( ( int ) ( default_color.y * 255.0f ), 0, 255 );
            buffer_data[2] = ( unsigned char ) optix::clamp ( ( int ) ( default_color.z * 255.0f ), 0, 255 );
            buffer_data[3] = 255;
            buffer->unmap();
            sampler->setBuffer ( 0u, 0u, buffer );
				
			
            // Although it would be possible to use nearest filtering here, we chose linear
            // to be consistent with the textures that have been loaded from a file. This
            // allows OptiX to perform some optimizations.
            sampler->setFilteringModes ( RT_FILTER_LINEAR, RT_FILTER_LINEAR, RT_FILTER_NONE );
            textureSamplers[key] = sampler;
            return sampler;
        }

    } else { // return already existing sampler
        return textureSamplers[key];
    }
}

SUTILAPI optix::TextureSampler loadEnvironmentTexture ( optix::Context context,
	const std::string &filename,
	float dc_r, float dc_g, float dc_b ) {
    std::string key = filename;// + default_color.x;
    key += dc_r;
    key += dc_g;
    key += dc_b;

    if ( textureSamplers.find ( key ) == textureSamplers.end() ) {
//         ilInit();
//         ILuint imageId;
//         ilGenImages ( 1, &imageId );
//         ilBindImage ( imageId );
//         ILboolean success = ilLoadImage ( ( ILstring ) ( filename.c_str() ) );
        optix::TextureSampler sampler = context->createTextureSampler();
        sampler->setWrapMode ( 0, RT_WRAP_REPEAT );
        sampler->setWrapMode ( 1, RT_WRAP_REPEAT );
        sampler->setWrapMode ( 2, RT_WRAP_REPEAT );
        sampler->setIndexingMode ( RT_TEXTURE_INDEX_NORMALIZED_COORDINATES );
        sampler->setReadMode ( RT_TEXTURE_READ_NORMALIZED_FLOAT );
        sampler->setMaxAnisotropy ( 1.0f );
        sampler->setMipLevelCount ( 1u );
        sampler->setArraySize ( 1u );
		
		if ( filename != "" ) {

			fipImage input;
			
			input.load( filename.c_str() );
			input.convertToRGBF();
			input.flipVertical();
		
            unsigned int texWidth = input.getWidth();
            unsigned int texHeight = input.getHeight();

			optix::Buffer buffer;
				
			buffer = context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT4, texWidth, texHeight );
			
            float *buffer_data = static_cast<float *> ( buffer->map() );

			float maxVal = 0.f;
			
			
			for (int y = 0; y < texHeight; ++y) {
				
				float* scanLine = (float*) input.getScanLine(y);
				
				for (int x = 0; x < texWidth; ++x) {

					buffer_data[y * texWidth * 4 + x * 4 + 3] = 1.f;
					
					for (int i = 0; i < 3; ++i) {

						buffer_data[y * texWidth * 4 + x * 4 + i] = scanLine[x * 3 + i];
													
					}
				}
					
			}
		
						
			
			buffer->unmap();

			 
            sampler->setBuffer ( 0u, 0u, buffer );
//             ilDeleteImages ( 1, &imageId );
            textureSamplers[key] = sampler;
            return sampler;

        } else {
            // Create buffer with single texel set to default_color
            optix::Buffer buffer = context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT4, 1u, 1u );
            float *buffer_data = static_cast<float *> ( buffer->map() );
            buffer_data[0] = dc_r;
            buffer_data[1] = dc_g;
            buffer_data[2] = dc_b;
            buffer_data[3] = 1.0f;
            buffer->unmap();
            sampler->setBuffer ( 0u, 0u, buffer );
            // Although it would be possible to use nearest filtering here, we chose linear
            // to be consistent with the textures that have been loaded from a file. This
            // allows OptiX to perform some optimizations.
            sampler->setFilteringModes ( RT_FILTER_LINEAR, RT_FILTER_LINEAR, RT_FILTER_NONE );
            textureSamplers[key] = sampler;
            return sampler;
        }

    } else {
        return textureSamplers[key];
    }

    /*if(IsHDR)
     *	return loadHDRTexture(context, filename, default_color);
     *	else
     *	return loadPPMEnvironmentTexture(context, filename, default_color);
     */
}




