/*
 * KeyboardController.h
 *
 *  Created on: Jul 4, 2011
 *      Author: troth22s
 */

#include <KeyboardController.h>


// #include <SFML/Window/Input.hpp>
#include <SFML/Window/Event.hpp>
#include <map>
#include <string>
#include <iostream>





/** @brief add or subtract some delta to/from the bound variable */
void KeyboardController::addSubValue ( sf::Keyboard::Key p_keyIdentifier ) {
    //		std::cout << "addSubValue(" << p_keyIdentifier << ")" << std::endl;;
    float *theVar = __m_vars[p_keyIdentifier];
    //		std::cout << "addSubValue old: " << *theVar << ", new: ";
    float addedValue = ( *theVar ) + __m_valDelta[p_keyIdentifier];
    *theVar = addedValue > __m_maxValues[p_keyIdentifier] ? __m_maxValues[p_keyIdentifier] : addedValue < __m_minValues[p_keyIdentifier] ? __m_minValues[p_keyIdentifier] : addedValue;
    //		std::cout << *theVar << std::endl;
}

/** @brief multiply or divide the bound variable with/by some delta */
void KeyboardController::mulDivValue ( sf::Keyboard::Key p_keyIdentifier ) {
    //		std::cout << "mulDivValue(" << p_keyIdentifier << ")" << std::endl;;
    float *theVar = __m_vars[p_keyIdentifier];
    //		std::cout << "mulDivValue old: " << *theVar << ", new: ";
    float mulValue = ( *theVar ) * __m_valDelta[p_keyIdentifier];
    *theVar = mulValue > __m_maxValues[p_keyIdentifier] ? __m_maxValues[p_keyIdentifier] : mulValue < __m_minValues[p_keyIdentifier] ? __m_minValues[p_keyIdentifier] : mulValue;
    //		std::cout << *theVar << std::endl;
}

/** @brief toggle the bound variable between its min and max value */
void KeyboardController::toggleValue ( sf::Keyboard::Key p_keyIdentifier ) {
    //		std::cout << "toggleValue(" << p_keyIdentifier << ")" << std::endl;;
    float *theVar = __m_vars[p_keyIdentifier];
    std::cout << "toggleValue old: " << *theVar << ", new: ";
    *theVar = *theVar == __m_maxValues[p_keyIdentifier] ? __m_minValues[p_keyIdentifier] : __m_maxValues[p_keyIdentifier];
    std::cout << *theVar << std::endl;
}


KeyboardController::KeyboardController() {
}

KeyboardController::~KeyboardController() {
}

/** @brief map one of the functionalities to an arbitrary key */
void KeyboardController::mapFunctionality ( sf::Keyboard::Key p_keyIdentifier, KeyboardFunctionalities p_functionality, ModificationType p_modtype, bool p_oncePerKeyPress, float *p_var, float p_min, float p_max, float p_delta ) {
    if ( __m_funcKeyMap.find ( p_functionality ) != __m_funcKeyMap.end() ) {
        std::cout << "WARNING: functionality already assigned to key " << __m_funcKeyMap[p_functionality] << std::endl;
    }

    __m_funcKeyMap[p_functionality] = p_keyIdentifier;
    __m_keyPressed[p_keyIdentifier] = false;
    __m_valDelta[p_keyIdentifier] = p_delta;
    __m_minValues[p_keyIdentifier] = p_min;
    __m_maxValues[p_keyIdentifier] = p_max;
    __m_vars[p_keyIdentifier] = p_var;
    __m_oncePerKeyPress[p_keyIdentifier] = p_oncePerKeyPress;

    switch ( p_modtype ) {
        case ENUM_KBDFUNC_MODTYPES_ADD_SUB: {
                __m_functions[p_keyIdentifier] = &KeyboardController::addSubValue;
                break;
            }

        case ENUM_KBDFUNC_MODTYPES_MUL_DIV: {
                __m_functions[p_keyIdentifier] = &KeyboardController::mulDivValue;
                break;
            }

        case ENUM_KBDFUNC_MODTYPES_TOGGLE: {
                __m_functions[p_keyIdentifier] = &KeyboardController::toggleValue;
                break;
            }

        default: {
                break;
            }
    }
}

void KeyboardController::resetInput() {
    for ( std::map<sf::Keyboard::Key, bool>::iterator it = __m_justChanged.begin(); it != __m_justChanged.end(); ++it ) {
        it->second = false;
    }

    for ( std::map<sf::Keyboard::Key, bool>::iterator it = __m_keyChecked.begin(); it != __m_keyChecked.end(); ++it ) {
        it->second = false;
    }
}


void KeyboardController::processRealTimeInput () {
    for ( std::map<sf::Keyboard::Key, bool>::iterator it = __m_oncePerKeyPress.begin(); it != __m_oncePerKeyPress.end(); ++it ) {
		if ( sf::Keyboard::isKeyPressed ( it->first ) ) {
			__m_keyPressed[it->first] = true;
			( this->*__m_functions[it->first] ) ( it->first );

		} else {
			__m_keyPressed[it->first] = false;
		}
	}
}

/** @brief process keyboard input by checking sfml keystates for all functions */
/*void KeyboardController::processEventBasedInput ( const sf::Keyboard &p_input, sf::Event &event ) {
    //		for (std::map<KeyboardFunctionalities, sf::Keyboard::Key>::iterator it = __m_funcKeyMap.begin(); it != __m_funcKeyMap.end(); ++it) {
    if ( __m_keyPressed.find ( event.KeyEvent.code ) == __m_keyPressed.end() || __m_keyChecked[event.KeyEvent.code] ) {
        return;
    }

    __m_keyChecked[event.Key.Code] = true;

    if ( __m_oncePerKeyPress[event.Key.Code] ) {
        if ( event.Type == sf::Event::KeyPressed ) {
            if ( !__m_keyPressed[event.Key.Code] ) {
//					std::cout << "trying to call function for " << (event.Key.Code) << std::endl;
                ( this->*__m_functions[event.Key.Code] ) ( event.KeyEvent.Code );
                __m_justChanged[event.Key.Code] = true;
            }

            __m_keyPressed[event.Key.Code] = true;

        } else if ( event.Type == sf::Event::KeyReleased ) {
            if ( __m_keyPressed[event.Key.Code] ) {
                __m_justChanged[event.Key.Code] = true;
            }

            __m_keyPressed[event.Key.Code] = false;
        }
    }
}*/


/* @brief check if a key is currently pressed */
int KeyboardController::keyPressed ( sf::Keyboard::Key p_keyIdentifier ) {
    return __m_keyPressed[p_keyIdentifier];
}

int KeyboardController::keyForFunctionalityPressed ( KeyboardFunctionalities p_functionality ) {
    return __m_keyPressed[__m_funcKeyMap[p_functionality]];
}

/* @brief check if the key for a continuously executed functionality is currently pressed */
int KeyboardController::functionalityActive ( KeyboardFunctionalities p_functionality ) {
    return __m_keyPressed[__m_funcKeyMap[p_functionality]] && !__m_oncePerKeyPress[__m_funcKeyMap[p_functionality]];
}

int KeyboardController::justChanged ( KeyboardFunctionalities p_functionality ) {
//		std::cout << "justChanged " << p_functionality << ": " << (__m_justChanged[__m_funcKeyMap[p_functionality]] && __m_oncePerKeyPress[__m_funcKeyMap[p_functionality]]) << std::endl;
    return __m_justChanged[__m_funcKeyMap[p_functionality]] && __m_oncePerKeyPress[__m_funcKeyMap[p_functionality]] && __m_keyPressed[__m_funcKeyMap[p_functionality]];
}
