#include <OutputWindow.h>

#include <IL/il.h>
#include <IL/ilu.h>

#include <sstream>
#include <iostream>

#include <optixu/optixu_math_namespace.h>

#include <FreeImagePlus.h>

using namespace std;

OutputWindow::OutputWindow ( unsigned int p_width, unsigned int p_height, float p_displayFactor) {

    __m_window = tr1::shared_ptr<sf::RenderWindow> ( new sf::RenderWindow ( sf::VideoMode ( p_width, p_height, 32 ), "Spark" , sf::Style::Close) );
	sf::Vector2u newSize = sf::Vector2u(p_width * p_displayFactor, p_height * p_displayFactor);
	__m_window->setSize ( newSize );
    __m_width = p_width;
    __m_height = p_height;
    __m_displayFactor = p_displayFactor;
    __m_texture = tr1::shared_ptr<sf::Texture> ( new sf::Texture ( ) );
	__m_texture->create(p_width, p_height);
    __m_sprite = tr1::shared_ptr<sf::Sprite> ( new sf::Sprite ( *__m_texture ) );
    __m_window->setMouseCursorVisible ( true );
	__m_window->setKeyRepeatEnabled(false);
	
	__m_font.loadFromFile("arial.ttf");
	
    if ( p_displayFactor != 1.f ) {
        __m_texture->setSmooth ( true );

    } else {
        __m_texture->setSmooth ( false );
    }

    __m_showStats = false;
    __m_quit = false;
//	__m_renderManager = p_renderManager;
    __m_kbdController.mapFunctionality ( sf::Keyboard::Key::Escape, ENUM_KBDFUNC_FUNC_QUIT, ENUM_KBDFUNC_MODTYPES_TOGGLE, true, &__m_quit, 0, 1, 1 );
    __m_kbdController.mapFunctionality ( sf::Keyboard::Key::Num0, ENUM_KBDFUNC_FUNC_SHOWSTATS, ENUM_KBDFUNC_MODTYPES_TOGGLE, true, &__m_showStats, 0, 1, 1 );
    __m_focalPointY = .5f;
	

}


OutputWindow::~OutputWindow() {
}


void OutputWindow::eventLoop ( bool &focusChanged ) {
    __m_kbdController.resetInput();
    sf::Event event;
	//const sf::Input &input = __m_window->GetInput();

	sf::Vector2i mousePos = sf::Mouse::getPosition(*__m_window.get());

    __m_sixAxis.resetTriggers();

    while ( __m_window->pollEvent ( event ) ) {
        if ( event.type == sf::Event::MouseButtonPressed ) {
			std::cout << mousePos.x << ", " << mousePos.y << std::endl;
            __m_focalPointX = ( float ) ( mousePos.x ) / ( float ) __m_window->getSize().x;
            __m_focalPointY = 1.f - ( float ) ( mousePos.y ) / ( float ) __m_window->getSize().y; //(__m_height * __m_displayFactor - input.GetMouseY() - 1) / __m_displayFactor;
            focusChanged = true;

        } else if ( event.type == sf::Event::KeyPressed || event.type == sf::Event::KeyReleased ) {
//             __m_kbdController.processEventBasedInput ( input, event ); // process input each subframe, so movement does not get slower
			__m_kbdController.processRealTimeInput ();

        } else {
            __m_sixAxis.handleInteraction ( event );
        }
    }
}


void OutputWindow::drawImage() {
	__m_texture->update ( __m_imageData );
    __m_window->draw ( *__m_sprite );
}


void OutputWindow::drawStats ( BenchmarkValues &bmv, RenderParameters *renderParameters, unsigned int numPreviousFrames, float eye_x, float eye_y, float eye_z, float direction_x, float direction_y, float direction_z ) {
    optix::float3 eye = optix::make_float3 ( eye_x, eye_y, eye_z );
    optix::float3 direction = optix::make_float3 ( direction_x, direction_y, direction_z );

    if ( !__m_showStats ) {
        return;
    }

    sf::RectangleShape statBox = sf::RectangleShape ( sf::Vector2f(200, 200));
	statBox.setPosition(5, 5);
	statBox.setFillColor(sf::Color(0, 0, 0, 120));
	
    sf::Text text;
    std::ostringstream strs;
    strs.setf ( ios::fixed, ios::floatfield );
    strs.precision ( 1 );
    strs << "Res: " << __m_width << "x" << __m_height << "\navgMrays/s: " << bmv.mrays << "\nfps: " << ( 1.f / bmv.time_total ) << "\nspp: " << ( numPreviousFrames ) << "\nmax. recursion depth: " << renderParameters->getParameterInt ( ENUM_RENDERPARAMETER_MAX_DEPTH ) << "\n-----" << "\ntime (pre rendering): " << ( int ) ( bmv.time_prepare * 1000 ) << " (" << bmv.time_prepare / bmv.time_total * 100.f << " %)" << "\ntime (rendering): " << ( int ) ( bmv.time_rendering * 1000 ) << " (" << bmv.time_rendering / bmv.time_total * 100.f << " %)" << "\ntime (postproc): " << ( int ) ( bmv.time_postproc * 1000 ) << " (" << bmv.time_postproc / bmv.time_total * 100.f << " %)" << "\ntime (display): " << ( int ) ( bmv.time_display * 1000 ) << " (" << bmv.time_display / bmv.time_total * 100.f << " %)" << "\ntime (total): " << ( int ) ( bmv.time_total * 1000 );
    std::ostringstream strs2;
    strs2.setf ( ios::fixed, ios::floatfield );
    strs2.precision ( 3 );
    strs2 << "eye: " << eye.x << ", " << eye.y << ", " << eye.z << "\n" << "dir: " << direction.x << ", " << direction.y << ", " << direction.z;
    std::string textContent = strs.str() + "\n-----\n" + strs2.str();
    text.setString ( textContent );
    text.setFont ( __m_font );
    text.setCharacterSize ( 12 );
    text.setPosition ( 8, 8 );
    text.setColor ( sf::Color ( 255, 255, 255, 255 ) );
    __m_window->draw ( statBox );
     __m_window->draw ( text );
}



void OutputWindow::saveImage (TrialConfig p_tc, unsigned int numPreviousFrames, float4* p_buf) {
    FreeImage_Initialise();    

    fipImage image(FIT_RGBAF, __m_width, __m_height, 128);
    memcpy(image.accessPixels(), p_buf, sizeof(float4) * __m_width * __m_height);

    std::stringstream ss;
    ss << numPreviousFrames;
    std::string theString = ss.str();
    std::string theString2 = ss.str();

    while ( theString.length() < 6 ) {
        theString = "0" + theString;
        theString2 = "0" + theString2;
    }

    std::stringstream ss2_exr;
    std::stringstream ss2_png;
    ss2_exr << p_tc.id << "/exr/";
    ss2_png << p_tc.id << "/png/";

    theString = p_tc.outputPath + ss2_exr.str() + theString + ".exr";

    // std::cout << "output path: " << theString << std::endl;
    // exit(0);

    image.save(theString.c_str());

    // theString2 = p_tc.outputPath + ss2_png.str() + theString2 + ".png";

    // bool success = image.convertToRGBF();
    // if (!success) {
    //     std::cout << "could not convert 128bit to 96bit image" << std::endl;
    //     exit(1);
    // }

    // success = image.toneMapping(FITMO_REINHARD05);
    // if (!success) {
    //     std::cout << "could not perform tonemapping" << std::endl;
    //     exit(1);
    // }

    // image.save(theString2.c_str());

}


SixAxis *OutputWindow::getSixAxis() {
    return &__m_sixAxis;
}


void OutputWindow::setImageData ( void *p_imageData ) {
    __m_imageData = ( sf::Uint8 * ) p_imageData;
}


sf::RenderWindow *OutputWindow::getWindow() {
    return __m_window.get();
}

sf::Image *OutputWindow::getImage() {
	static sf::Image image;
	image = __m_texture->copyToImage();
    return &image;
}

sf::Sprite *OutputWindow::getSprite() {
    return __m_sprite.get();
}

sf::Uint8 *OutputWindow::getImageData() {
    return __m_imageData;
}


float OutputWindow::getFocalPointX() {
    return __m_focalPointX;
}


float OutputWindow::getFocalPointY() {
    return __m_focalPointY;
}


KeyboardController *OutputWindow::getKbdController() {
    return &__m_kbdController;
}

bool OutputWindow::isRunning() {
    return !__m_quit;
}


