#include <PostProcessor.h>

// grip
#include <Data/DataUtils.h>

// own includes
#include <GripBuffers.h>
#include <BufferTypes.h>

using namespace grip;





PostProcessor::PostProcessor ( const std::string &p_pathToGraph, std::vector<BufferInfo> *p_bufInfo, std::vector<CUdeviceptr> *p_devicePointers ) {
    __m_pathToGraph = p_pathToGraph;
    __m_grunner.init ( __m_pathToGraph, true, ENUM_GRAPHVISUALIZER_NONE, true );
    __m_bufInfo = p_bufInfo;
    __m_gripBuffers.initialize ( ENUM_NUM_BUFFERTYPES );
    __m_devicePointers = p_devicePointers;
    createGripBuffers();
    setupGripBuffers();
}


PostProcessor::~PostProcessor() {
}



void PostProcessor::createGripBuffers() {
    for ( int i = 0; i < ENUM_NUM_BUFFERTYPES; ++i ) {
        std::cout << "Creating Buffer " << i << std::endl;
        enum_BufferTypes bufferType = ( enum_BufferTypes ) i;
        unsigned int width = ( *__m_bufInfo ) [bufferType].m_width;
        unsigned int height = ( *__m_bufInfo ) [bufferType].m_height;

        switch ( ( *__m_bufInfo ) [bufferType].m_format ) {
            case RT_FORMAT_FLOAT  :
                __m_gripBuffers.m_buffers[bufferType] = dt_UniversalImage_float_1_sptr ( new dt_UniversalImage_float_1_t ( width, height ) );

            case RT_FORMAT_FLOAT4 :
                __m_gripBuffers.m_buffers[bufferType] = dt_UniversalImage_float_4_sptr ( new dt_UniversalImage_float_4_t ( width, height ) );

            default:
                break;
        }
    }
}




void PostProcessor::setupGripBuffers() {
    for ( int i = 0; i < ENUM_NUM_BUFFERTYPES; ++i ) {
        enum_BufferTypes bufferType = ( enum_BufferTypes ) i;
        std::string bufferName = ( *__m_bufInfo ) [bufferType].m_name_gripPrefix + ( *__m_bufInfo ) [bufferType].m_name;

        if ( ( *__m_bufInfo ) [bufferType].m_format == RT_FORMAT_FLOAT ) {
            dt_UniversalImage_float_1_sptr dataSptr = reinterpret_cast_void_sptr<dt_UniversalImage_float_1_sptr> ( __m_gripBuffers.m_buffers[bufferType] );
            __m_grunner.setParameter<dt_UniversalImage_float_1_sptr> ( bufferName, dataSptr, 1 );
            __m_grunner.setDirty ( bufferName, ENUM_COMPENG_CUDA );
            __m_grunner.setDevicePtr ( bufferName, ( *__m_devicePointers ) [bufferType] );

        } else if ( ( *__m_bufInfo ) [bufferType].m_format == RT_FORMAT_FLOAT4 ) {
            dt_UniversalImage_float_4_sptr dataSptr = reinterpret_cast_void_sptr<dt_UniversalImage_float_4_sptr> ( __m_gripBuffers.m_buffers[bufferType] );
            __m_grunner.setParameter<dt_UniversalImage_float_4_sptr> ( bufferName, dataSptr, 1 );
            __m_grunner.setDirty ( bufferName, ENUM_COMPENG_CUDA );
            __m_grunner.setDevicePtr ( bufferName, ( *__m_devicePointers ) [bufferType] );
        }
    }
}


void PostProcessor::setDevicePointers ( std::vector<CUdeviceptr> *p_devicePointers ) {
    __m_devicePointers = p_devicePointers;
}



void PostProcessor::markGripBuffersDirty() {
    for ( int i = 0; i < ENUM_NUM_BUFFERTYPES; ++i ) {
        enum_BufferTypes bufferType = ( enum_BufferTypes ) i;
        std::string bufferName = ( *__m_bufInfo ) [bufferType].m_name_gripPrefix + ( *__m_bufInfo ) [bufferType].m_name;
        __m_grunner.setDirty ( bufferName, ENUM_COMPENG_CUDA );
    }
}


void PostProcessor::run() {
    __m_grunner.run();
}


bool PostProcessor::parameterSetChanged ( const std::string &p_parameterSetName ) {
    return __m_grunner.parameterSetChanged ( p_parameterSetName );
}
