#include <RenderManager.h>


#include <RenderProcess.h>
#include <BufferTypes.h>
#include <BufferData_UserTypes.h>


#include <trdUtils/Lightfield_octree.h>


#include <unistd.h>
#include <getopt.h>

#include <SFML/System/Clock.hpp>
#include <FreeImagePlus.h>

#include <fstream>
#include <string>
#include <iostream>

enum enum_cmdLineParams {

    ENUM_CMDLNPRM_WIDTH = 0,
    ENUM_CMDLNPRM_HEIGHT = 1,
    ENUM_CMDLNPRM_SCALE,
    ENUM_CMDLNPRM_ENVMAP,
    ENUM_CMDLNPRM_MODEL,
    ENUM_CMDLNPRM_UPVEC,
    ENUM_CMDLNPRM_FIRSTGPU,
    ENUM_CMDLNPRM_NUMGPUS,
    ENUM_CMDLNPRM_GAMMA,
    ENUM_CMDLNPRM_CAMFILE,
    ENUM_CMDLNPRM_EVALCONFIG,

    ENUM_NUM_CMDLNPRMS

};


std::ofstream* out;
trd::Lightfield theLightfield;

void RenderManager::parseCommandLineParameters ( int p_argc, char **p_argv ) {
    bool optReq[ENUM_NUM_CMDLNPRMS];
    optReq[ENUM_CMDLNPRM_WIDTH]        = true;
    optReq[ENUM_CMDLNPRM_HEIGHT]       = true;
    optReq[ENUM_CMDLNPRM_SCALE]        = true;
    optReq[ENUM_CMDLNPRM_ENVMAP]       = true;
    optReq[ENUM_CMDLNPRM_MODEL]        = true;
    optReq[ENUM_CMDLNPRM_UPVEC]        = true;
    optReq[ENUM_CMDLNPRM_FIRSTGPU]     = true;
    optReq[ENUM_CMDLNPRM_NUMGPUS]      = true;
    optReq[ENUM_CMDLNPRM_GAMMA]        = true;
    optReq[ENUM_CMDLNPRM_CAMFILE]      = false;
    optReq[ENUM_CMDLNPRM_EVALCONFIG]      = false;
    int numReqOpts = 0;

    for ( int i = 0; i < ENUM_NUM_CMDLNPRMS; ++i ) {
        if ( optReq[i] ) { ++numReqOpts; }
    }

    static struct option long_options[] = {

        {"width", 	 	 required_argument, 0, ENUM_CMDLNPRM_WIDTH},
        {"height", 	 	 required_argument, 0, ENUM_CMDLNPRM_HEIGHT},
        {"scale", 	 	 required_argument, 0, ENUM_CMDLNPRM_SCALE},
        {"envmap", 	 	 required_argument, 0, ENUM_CMDLNPRM_ENVMAP},
        {"model", 	 	 required_argument, 0, ENUM_CMDLNPRM_MODEL},
        {"upvec", 	 	 required_argument, 0, ENUM_CMDLNPRM_UPVEC},
        {"firstgpu", 	 required_argument, 0, ENUM_CMDLNPRM_FIRSTGPU},
        {"numgpus",  	 required_argument, 0, ENUM_CMDLNPRM_NUMGPUS},
        {"gamma", 	 	 required_argument, 0, ENUM_CMDLNPRM_GAMMA},
        {"camfile",      required_argument, 0, ENUM_CMDLNPRM_CAMFILE},
        {"evalconf",     required_argument, 0, ENUM_CMDLNPRM_EVALCONFIG},
        {0, 0, 0, 0}

    };
    
    std::string optAbbrvs;
    int index = 0;

    for ( int i = 0; i < ENUM_NUM_CMDLNPRMS; ++i ) {
        optAbbrvs += i;

        switch ( long_options[i].has_arg ) {
            case required_argument:
                { optAbbrvs += ":"; break; }

            case optional_argument:
                { optAbbrvs += "::"; break; }

            default:
                break;
        }
    }

    uint32_t cmdLineParamMask_full = powf ( 2, numReqOpts ) - 1; // bitmask for command line parameters
    std::cout << cmdLineParamMask_full << std::endl;
    uint32_t cmdLineParamMask = 0;

    while ( true ) {
        int option_index = 0;
        int c = getopt_long ( p_argc, p_argv, optAbbrvs.c_str(), long_options, &option_index );

        if ( c != -1 ) {
            switch ( c ) {
                case ENUM_CMDLNPRM_WIDTH:
                    { __m_cmdLineParams.width = atoi ( optarg ); cmdLineParamMask |= 1 << 0;	break; }

                case ENUM_CMDLNPRM_HEIGHT:
                    { __m_cmdLineParams.height = atoi ( optarg ); cmdLineParamMask |= 1 << 1; break; }

                case ENUM_CMDLNPRM_SCALE:
                    { __m_cmdLineParams.displayFactor = atof ( optarg ); cmdLineParamMask |= 1 << 2; break; }

                case ENUM_CMDLNPRM_ENVMAP:
                    { __m_cmdLineParams.texturePath = optarg; cmdLineParamMask |= 1 << 3; break; }

                case ENUM_CMDLNPRM_MODEL:
                    { __m_cmdLineParams.modelPath = optarg; cmdLineParamMask |= 1 << 4; break; }

                case ENUM_CMDLNPRM_UPVEC:
                    { __m_cmdLineParams.upVector = optarg; cmdLineParamMask |= 1 << 5; break; }

                case ENUM_CMDLNPRM_FIRSTGPU:
                    { __m_cmdLineParams.firstGPU = atoi ( optarg ); cmdLineParamMask |= 1 << 6; break; }

                case ENUM_CMDLNPRM_NUMGPUS:
                    { __m_cmdLineParams.numGPUs = atoi ( optarg ); cmdLineParamMask |= 1 << 7; break; }

                case ENUM_CMDLNPRM_GAMMA:
                    { __m_cmdLineParams.envGamma = atof ( optarg ); cmdLineParamMask |= 1 << 8; break; }

                case ENUM_CMDLNPRM_CAMFILE:
                    { __m_cmdLineParams.camFile = optarg; break; }

                case ENUM_CMDLNPRM_EVALCONFIG:
                    { __m_cmdLineParams.evalConfig = optarg; break; }

                default:
                    { std::cout << "invalid argument" << std::endl; break; }
            }

        } else { break; }
    }

    if ( cmdLineParamMask != cmdLineParamMask_full ) {
        std::cout << "Usage: " << p_argv[0] << " [options], where options are as follows:" << std::endl;
        std::cout << "--width <width>, -h <height>: the width and height of the rendered image, respectively" << std::endl;
        std::cout << "--scale <displayFactor>: the rendered image will be upscaled by this factor before displaying it" << std::endl;
        std::cout << "--envmap <hdrEnvMap>: Path to an .hdr environment map" << std::endl;
        std::cout << "--gamma <gamma>: gamma value for environment map (e.g. 2.2)" << std::endl;
        std::cout << "--model <objModel>: Path to the model to be loaded in obj format" << std::endl;
        std::cout << "--upvec <upVector>: direction which is \"up\" in the model file. Can be [+|-][x|y|z]" << std::endl;
        std::cout << "--firstgpu <firstGPU>: first GPU to be used" << std::endl;
        std::cout << "--numgpus <numGPUs>: number of GPUs to be used" << std::endl;
        exit ( 1 );
    }
}




//TODO Add buffer description!
void RenderManager::createBufferDescriptions() {
    int width = __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_RES_X );
    int height = __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_RES_Y );
    __m_bufInfo[ENUM_BUFFERTYPES_DIRECT_LIGHT] = BufferInfo ( "directLight", "input_color_", width, height, RT_FORMAT_FLOAT4 );
    __m_bufInfo[ENUM_BUFFERTYPES_INDIRECT_LIGHT] = BufferInfo ( "indirectLight", "input_color_", width, height, RT_FORMAT_FLOAT4 );
    __m_bufInfo[ENUM_BUFFERTYPES_INDIRECT_LIGHT_SQUARED] = BufferInfo ( "indirectLight_squared", "input_color_", width, height, RT_FORMAT_FLOAT4 );
    __m_bufInfo[ENUM_BUFFERTYPES_POSITION] = BufferInfo ( "position", "input_", width, height, RT_FORMAT_FLOAT4 );
    __m_bufInfo[ENUM_BUFFERTYPES_RAY_DIRECTION] = BufferInfo ( "rayDirection", "input_", width, height, RT_FORMAT_FLOAT4 );
    __m_bufInfo[ENUM_BUFFERTYPES_RAY_DIRECTION_SQUARED] = BufferInfo ( "rayDirection_squared", "input_", width, height, RT_FORMAT_FLOAT4 );
    __m_bufInfo[ENUM_BUFFERTYPES_TEXTURE] = BufferInfo ( "texture", "input_", width, height, RT_FORMAT_FLOAT4 );
    __m_bufInfo[ENUM_BUFFERTYPES_DEPTH_NORMAL] = BufferInfo ( "depth_normal", "input_", width, height, RT_FORMAT_FLOAT4 );
    __m_bufInfo[ENUM_BUFFERTYPES_CENTER_DEPTH] = BufferInfo ( "centerDepth", "input_", 1, RT_FORMAT_FLOAT );
	__m_bufInfo[ENUM_BUFFERTYPES_OUTPUT_OCTREE] = BufferInfo ( "octreeReconstruction", "input_", width, height, RT_FORMAT_FLOAT4 );
	__m_bufInfo[ENUM_BUFFERTYPES_LIGHT_FIELD] = BufferInfo ( "lightField", "input_", width, height, RT_FORMAT_USER, RT_BUFFER_OUTPUT, sizeof(UserType_lightField));
    // std::cout << "Size of lightfield type: " << sizeof(UserType_lightField) << std::endl;
    // exit(0);
}





RenderManager::RenderManager ( int argc, char **argv, trd::Fifar* p_fifar ) {
    __m_fifar = p_fifar;

    __m_cmdLineParams.camFile = "";

    parseCommandLineParameters ( argc, argv );

    tc.read(getCmdLineParams().evalConfig);
    tc.print();

    __m_renderParameters.setParameterInt ( ENUM_RENDERPARAMETER_RES_X, __m_cmdLineParams.width );
    __m_renderParameters.setParameterInt ( ENUM_RENDERPARAMETER_RES_Y, __m_cmdLineParams.height );
    __m_renderParameters.setParameterInt ( ENUM_RENDERPARAMETER_FIRST_GPU, __m_cmdLineParams.firstGPU );
    __m_renderParameters.setParameterInt ( ENUM_RENDERPARAMETER_NUM_GPUS, __m_cmdLineParams.numGPUs );
    __m_renderParameters.setParameterInt ( ENUM_RENDERPARAMETER_MAX_DEPTH, 3 );
    __m_renderParameters.setParameterInt ( ENUM_RENDERPARAMETER_MAX_SOBOL_DEPTH, 32 );
    __m_renderParameters.setParameterFloat ( ENUM_RENDERPARAMETER_ENV_GAMMA, __m_cmdLineParams.envGamma );
    __m_outputBuffers.initialize ( ENUM_NUM_BUFFERTYPES );
    __m_bufInfo.resize ( ENUM_NUM_BUFFERTYPES );
    createBufferDescriptions();
    initializeRenderProcess();
    loadScene();
			
    setupCamera();
		
    createAccumulatedDataDevicePointers();
    validateAndCompile();
}


RenderManager::~RenderManager() {
    std::cout << "DELETING RENDERMANAGER" << std::endl;
}


/* SINGLE PROCESS IMPLEMENTATION, DEVICE POINTERS ARE SIMPLY COPIED! */
void RenderManager::createAccumulatedDataDevicePointers() {
    __m_accumulatedDevicePointers.resize ( ENUM_NUM_BUFFERTYPES );

    for ( int i = 0; i < ENUM_NUM_BUFFERTYPES; ++i ) {
        enum_BufferTypes bufferType = ( enum_BufferTypes ) i;
        __m_accumulatedDevicePointers[bufferType] = __m_outputBuffers.m_devicePointers[bufferType];
    }
}

void RenderManager::initializeRenderProcess() {
    __m_renderProcess.initialize ( &__m_renderParameters, &__m_outputBuffers, &__m_bufInfo );
}

void RenderManager::loadScene() {
    __m_renderProcess.loadScene ( tc, __m_cmdLineParams.texturePath, __m_cmdLineParams.upVector );
}

void RenderManager::setupCamera() {
optix::float3 size;
    float maxSize;
    size = __m_renderProcess.getScene().getSceneSize();
    maxSize = __m_renderProcess.getScene().getMaxSceneExtent();
    unsigned int flipYZ = __m_renderProcess.getContextWrapper().getContext()["envFlipYZ"]->getUint();

    optix::float3 eye = optix::make_float3 ( size.x / maxSize * 0.5, - size.y / maxSize * 2.f, size.z / maxSize * 0.5 );
    optix::float3 up = optix::make_float3 ( 0.f, 0.f, 1.f );
    optix::float3 look_at = optix::make_float3 (size.x / maxSize * 0.5, 0.0, size.z / maxSize * 0.5);
    if (flipYZ) {
        eye = optix::make_float3 ( size.x / maxSize * 0.5, size.y / maxSize * 0.5, size.z / maxSize * 2.f );
        up = optix::make_float3 ( 0.f, 1.f, 0.f );
        look_at = optix::make_float3 (size.x / maxSize * 0.5, size.y / maxSize * 0.5, 0.0);
    }
    

    optix::float2 res = optix::make_float2 ( __m_cmdLineParams.width, __m_cmdLineParams.height );
    __m_camera.initialize ( &eye.x, &up.x, &look_at.x, 90.f, &res.x );
    __m_renderProcess.setCamera ( __m_camera );
	
	//camera data could be written to lightfield configuration file instead
// 	out->write((char*)&(eye.x), sizeof(float) * 3);
// 	out->write((char*)&(up.x), sizeof(float) * 3);
// 	out->write((char*)&(look_at.x), sizeof(float) * 3);
}

void RenderManager::updateCameraAndParameters() {
    __m_renderProcess.setCamera ( __m_camera );
    __m_renderProcess.updateRenderParameters();
}


void RenderManager::refocusCamera ( float p_depthSampleX, float p_depthSampleY ) {
    optix::Context context = __m_renderProcess.getContextWrapper().getContext();
    context["depthSampleX"]->setFloat ( p_depthSampleX );
    context["depthSampleY"]->setFloat ( p_depthSampleY );
	
	std::cout << p_depthSampleX << ", " << p_depthSampleY << std::endl;
    context->launch ( ENUM_ENTRYPOINT_REFOCUS, 1 );
    optix::Buffer outputBuffer_centerDepth = context[string ( "output_buffer_" ) + __m_bufInfo[ENUM_BUFFERTYPES_CENTER_DEPTH].m_name]->getBuffer();
    float curFocalDist = * ( float * ) ( outputBuffer_centerDepth->map() );
    outputBuffer_centerDepth->unmap();
    context["focalDist"]->setFloat ( curFocalDist );
}


void RenderManager::validateAndCompile() {
    __m_renderProcess.getContextWrapper().getContext()->validate();
    __m_renderProcess.getContextWrapper().getContext()->compile();
}


void RenderManager::saveLightfield() {

	theLightfield.save("/tmp/input.txt");
	theLightfield.clearSamples();
	
}




float4* outbuf_h = 0;
float4* RenderManager::writeOutbufToHost() {
    uint renderWidth = __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_RES_X );
    uint renderHeight = __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_RES_Y );
    if (!outbuf_h) {
        outbuf_h = new float4[renderWidth * renderHeight];
    }
    cudaMemcpy(outbuf_h, (float4*) (__m_outputBuffers.m_devicePointers[ENUM_BUFFERTYPES_DIRECT_LIGHT]), renderWidth * renderHeight * sizeof(float4), cudaMemcpyDeviceToHost);
    // for (int y = 0; y < renderHeight; ++y) {
    //     for (int x = 0; x < renderWidth; ++x) {
    //         std::cout << outbuf_h[y * renderWidth + x].x << "; ";
    //     }
    //     std::cout << std::endl;
    // }
    return outbuf_h;
}





void RenderManager::render ( std::ofstream& logfile, unsigned int p_oldKeyFrameIndex, unsigned int &p_numPreviousFrames, unsigned int &p_frame, unsigned int p_maxSamples, float p_lfWriteProb, unsigned int p_calcReconstruction ) {
	
	

	optix::Context context = __m_renderProcess.getContextWrapper().getContext();
    context->compile();
    context["addRays"]->setUint ( 0 );
    context["numPreviousFrames"]->setUint ( p_numPreviousFrames );
    context["frame"]->setUint ( p_frame );

	sf::Clock clock;
    float time_trace = 0;
    clock.restart();
	uint samplesPerLaunch = __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_SAMPLES_PER_LAUNCH );
    if ( p_numPreviousFrames < p_maxSamples || p_maxSamples == -1 ) {
        clock.restart();
        std::cout << p_numPreviousFrames << std::endl;
        for ( int i = 0; i < samplesPerLaunch; ++i ) {
            std::cout << i << " / " << samplesPerLaunch << std::endl;
            int programId = (p_lfWriteProb > 0.f || p_calcReconstruction) ? ENUM_ENTRYPOINT_PT_LF : ENUM_ENTRYPOINT_PT;
			
			context->launch (
                programId,
                __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_RES_X ),
                __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_RES_Y )
            );
            context["numPreviousFrames"]->setUint ( ++p_numPreviousFrames );
            context["frame"]->setUint ( ++p_frame );
            // context["frame"]->setUint ( ++p_frame );
            // context["addRays"]->setUint ( 1 );
			
		  
          time_trace += (double) clock.getElapsedTime().asMilliseconds() / (double) samplesPerLaunch;	
		  clock.restart();	
		}


        uint width = __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_RES_X );
        uint height = __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_RES_Y );
        fipImage image(FIT_RGBAF, width, height, 128);
        
        cudaMemcpy (image.accessPixels(), (void*)__m_outputBuffers.m_devicePointers[ENUM_BUFFERTYPES_DEPTH_NORMAL], sizeof(float4) * width * height, cudaMemcpyDeviceToHost);

        // memcpy(image.accessPixels(), p_buf, sizeof(float4) * __m_width * __m_height);


        image.save("/tmp/depth_normal.exr");

        
        context["frame"]->setUint ( ++p_frame );
        if (p_maxSamples > 0) {
            logfile << p_oldKeyFrameIndex << "; " << time_trace << "; " << std::endl;
        }
    }
}



void RenderManager::setMinDepthMode() {
    __m_renderProcess.getContextWrapper().getContext() [__m_renderParameters.getParameterNameOptix ( ENUM_RENDERPARAMETER_MAX_DEPTH )]->setUint ( __m_renderParameters.getParameterInt ( ENUM_RENDERPARAMETER_MIN_DEPTH ) );
}


std::vector<CUdeviceptr> *RenderManager::getAccumulatedDevicePointers() {
    return &__m_accumulatedDevicePointers;
}


std::vector<BufferInfo> *RenderManager::getBufferInfo() {
    return &__m_bufInfo;
}



CmdLineParams RenderManager::getCmdLineParams() {
    return __m_cmdLineParams;
}


RenderProcess RenderManager::getRenderProcess() {
    return __m_renderProcess;
}


RenderParameters *RenderManager::getRenderParameters() {
    return &__m_renderParameters;
}



Camera *RenderManager::getCamera() {
    return &__m_camera;
}


