#include <Scene.h>

#include <optixu/optixpp_namespace.h>
#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>
#include <ObjLoader.h>
#include <AssimpLoader.h>
#include <ImageLoader.h>

#include <SphereLight.cu>
#include <PointLight.cu>

#include <Distribution2D.h>

#include <path.h>

#include <iostream>


#include <AssimpLoader.h>


Scene::Scene() {
    __m_isInitialized = false;
}


Scene::Scene ( optix::Context p_context ) {
    initialize ( p_context );
}


void Scene::initialize ( optix::Context p_context ) {
    assert ( !__m_isInitialized );
    __m_context = p_context;
    __m_sceneGroup = __m_context->createGroup();
    __m_sceneGroup->setAcceleration ( __m_context->createAcceleration ( "Bvh", "Bvh" ) );
    __m_lightGroup = __m_context->createGroup();
    __m_lightGroup->setAcceleration ( __m_context->createAcceleration ( "NoAccel", "NoAccel" ) );
    __m_topGroup = __m_context->createGroup();
    __m_topGroup->setAcceleration ( __m_context->createAcceleration ( "Bvh", "Bvh" ) );
    __m_sceneTransform = __m_context->createTransform();
    __m_isInitialized = true;
}

Scene::~Scene() {
}



void Scene::setGroups() {
    /* create top group and set its children (light sources and geometry) and acceleration */
    std::cout << "SETTING CHILDREN" << std::endl;
    __m_topGroup->setChildCount ( 2 );
    std::cout << "LIGHT GROUP" << std::endl;
    __m_topGroup->setChild ( 0, __m_lightGroup );
    std::cout << "TRANSFORMATION NODE" << std::endl;
    __m_topGroup->setChild ( 1, __m_sceneTransform );
    __m_sceneTransform->setChild ( __m_sceneGroup );
    __m_context["top_object"]->set ( __m_sceneTransform );
    __m_context["light_group"]->set ( __m_lightGroup );
    // __m_context["scene_group"]->set ( __m_sceneTransform );
}


void Scene::loadModel ( const std::string &p_modelPath, const std::string &p_upVector ) {
    // ObjLoader objloader0 ( p_modelPath.c_str(), __m_context, __m_sceneGroup, "Sbvh", "Bvh", false );
    AssimpLoader assimpLoader;



    // define rotation to transform the model to its correct orientation (where "up" has to be +z then)
    float radiansToRotate = 0.f;
    optix::float3 rotAxis = optix::make_float3 ( 0.f, 0.f, 1.f );

    //std::cout << "argv[3] is " << (cmdLineParams.upVector ? cmdLineParams.upVector : "NULL") << std::endl;

    // if ( p_upVector == "+x" ) { // rotate around +z by 0.5 * M_PIf radians
    //     std::cout << "rotating around +y by 0.5 * pi radians" << std::endl;
    //     radiansToRotate = 0.5f * M_PIf;
    //     rotAxis = optix::make_float3 ( 0.f, 1.f, 0.f );

    // } else if ( p_upVector == "-x" ) { // rotate around -z by 0.5 * M_PIf radians
    //     std::cout << "rotating around -y by 0.5 * pi radians" << std::endl;
    //     radiansToRotate = 0.5f * M_PIf;
    //     rotAxis = optix::make_float3 ( 0.f, -1.f, 0.f );

    // } else 



    if ( p_upVector == "+y" ) {
        // std::cout << "rotating around +x by 0.5 * pi radians" << std::endl;
        // radiansToRotate = 0.5f * M_PIf;
        // rotAxis = optix::make_float3 ( 1.f, 0.f, 0.f );
        std::cout << "Flipping Y and Z for environment" << std::endl;
        __m_context["envFlipYZ"]->setUint(1);

    // } else if ( p_upVector == "-y" ) {
    //     std::cout << "Rotating around -x by pi radians" << std::endl;
    //     radiansToRotate = M_PIf;
    //     rotAxis = optix::make_float3 ( -1.f, 0.f, 0.f );

    // } 
    } else if ( p_upVector == "+z" ) { // rotate around -x by 0.5 * M_PIf radians
        std::cout << "No rotation necessary, orientation is already correct for up = +z" << std::endl;
        __m_context["envFlipYZ"]->setUint(0);
    } 
    // else if ( p_upVector == "-z" ) {
    //     std::cout << "rotating around +x by pi radians" << std::endl;
    //     radiansToRotate = M_PIf;
    //     rotAxis = optix::make_float3 ( 1.f, 0.f, 0.f );
    // }

    radiansToRotate = 0.f;

    assimpLoader.load_file(__m_context, Path(p_modelPath), __m_sceneGroup, optix::Matrix4x4::rotate ( radiansToRotate, rotAxis ));
    optix::Aabb theAabb = assimpLoader.getSceneBBox();
    // objloader0.load ( optix::Matrix4x4::rotate ( radiansToRotate, rotAxis ) );
    // optix::Aabb theAabb = objloader0.getSceneBBox();

    

    __m_sceneSize.x = theAabb.extent ( 0 );
    __m_sceneSize.y = theAabb.extent ( 1 );
    __m_sceneSize.z = theAabb.extent ( 2 );
    float farPlane = sqrtf ( __m_sceneSize.x * __m_sceneSize.x + __m_sceneSize.y * __m_sceneSize.y + __m_sceneSize.z * __m_sceneSize.z );
    __m_context["scene_extent"]->setFloat ( farPlane );
    __m_maxSize = fmaxf ( __m_sceneSize );
    optix::Matrix4x4 transformScale = optix::Matrix4x4::scale ( optix::make_float3 ( 1.f / __m_maxSize) );
    optix::Matrix4x4 transformTranslate = optix::Matrix4x4::translate ( optix::make_float3 ( -theAabb.m_min.x, -theAabb.m_min.y, -theAabb.m_min.z) );
    optix::Matrix4x4 transform = transformScale * transformTranslate;
    __m_sceneTransform->setMatrix ( false, transform.getData(), transform.inverse().getData() );

    normalizedSceneBB = optix::Aabb(optix::make_float3(transformScale * transformTranslate * optix::make_float4(theAabb.m_min, 1.f)), optix::make_float3(transformScale * transformTranslate * optix::make_float4(theAabb.m_max, 1.f)));
    uniformSceneBB = optix::Aabb(make_float3(0.f), make_float3(1.f));

    std::cout << "normalizedSceneBB min: " << normalizedSceneBB.m_min.x << ", " << normalizedSceneBB.m_min.y << ", " << normalizedSceneBB.m_min.z << std::endl;
    std::cout << "normalizedSceneBB max: " << normalizedSceneBB.m_max.x << ", " << normalizedSceneBB.m_max.y << ", " << normalizedSceneBB.m_max.z << std::endl;
    std::cout << "uniformSceneBB min: " << uniformSceneBB.m_min.x << ", " << uniformSceneBB.m_min.y << ", " << uniformSceneBB.m_min.z << std::endl;
    std::cout << "uniformSceneBB max: " << uniformSceneBB.m_max.x << ", " << uniformSceneBB.m_max.y << ", " << uniformSceneBB.m_max.z << std::endl;
    
    // exit(1);
}



void Scene::createLights(TrialConfig p_tc) {
    /* create light sources */
    std::cout << "Creating light sources" << std::endl;
    /* Create groups for geometrically represented light sources and geometry*/
    std::cout << "Creating light source group" << std::endl;
    __m_lightGroup->setChildCount ( 0 );
    std::cout << "- sphereLights" << std::endl;
    //  float fluxPerUnitArea = 0;
    // define sphere lights and copy buffer
    // LightSourceSphere sphereLights[0];
    
    if (p_tc.pointLightIntensity > 0.f) {
        // sphereLights = new LightSourceSphere(__m_context);
        LightSourceSphere sphereLights[1] = { LightSourceSphere(__m_context ) };
        // if (fluxPerUnitArea > 0.f) {
            // sphereLights[0].setPosition(optix::make_float3(3.f, 4.2f, 0.6f ));
            // sphereLights[0].setSize(farPlane / 200.f);
            // sphereLights[0].setColor(optix::make_float3(1.0f, 1.0f, .95f));
            // //   sphereLights[0].setFluxTotal(farPlane * 200.f);
            // sphereLights[0].setFluxPerUnitArea(atof(argv[2]));
            // stopObjects = 0.f;
        sphereLights[0].setPosition(optix::make_float3(p_tc.plPosX, p_tc.plPosY, p_tc.plPosZ));
        sphereLights[0].setSize(0.01f);
        sphereLights[0].setColor(optix::make_float3(p_tc.plColR, p_tc.plColG, p_tc.plColB));
        sphereLights[0].setFluxPerUnitArea(p_tc.pointLightIntensity);
    
        optix::Buffer sphereLightBuffer = __m_context->createBuffer ( RT_BUFFER_INPUT );
        sphereLightBuffer->setFormat ( RT_FORMAT_USER );
        sphereLightBuffer->setElementSize ( sizeof ( LightSourceSphere ) );
        unsigned int numSphereLights = sizeof ( sphereLights ) / sizeof ( LightSourceSphere );
        sphereLightBuffer->setSize ( numSphereLights );
        std::cout << "SphereLights: " << numSphereLights << std::endl;
        memcpy ( sphereLightBuffer->map(), sphereLights, sizeof ( sphereLights ) );
        sphereLightBuffer->unmap();
        __m_context["sphereLights"]->set ( sphereLightBuffer );
        int curChildCount = __m_lightGroup->getChildCount();
        __m_lightGroup->setChildCount ( curChildCount + numSphereLights );
        optix::Transform lightSourceSphereTransform;

        for ( int i = 0; i < numSphereLights; ++i ) {
            lightSourceSphereTransform = sphereLights[i].createGeometry ( __m_context );
            __m_lightGroup->setChild ( curChildCount + i, lightSourceSphereTransform );
        }
    } else {
         LightSourceSphere sphereLights[0];
         optix::Buffer sphereLightBuffer = __m_context->createBuffer ( RT_BUFFER_INPUT );
        sphereLightBuffer->setFormat ( RT_FORMAT_USER );
        sphereLightBuffer->setElementSize ( sizeof ( LightSourceSphere ) );
        unsigned int numSphereLights = sizeof ( sphereLights ) / sizeof ( LightSourceSphere );
        sphereLightBuffer->setSize ( numSphereLights );
        std::cout << "SphereLights: " << numSphereLights << std::endl;
        memcpy ( sphereLightBuffer->map(), sphereLights, sizeof ( sphereLights ) );
        sphereLightBuffer->unmap();
        __m_context["sphereLights"]->set ( sphereLightBuffer );
        int curChildCount = __m_lightGroup->getChildCount();
        __m_lightGroup->setChildCount ( curChildCount + numSphereLights );
        optix::Transform lightSourceSphereTransform;

        for ( int i = 0; i < numSphereLights; ++i ) {
            lightSourceSphereTransform = sphereLights[i].createGeometry ( __m_context );
            __m_lightGroup->setChild ( curChildCount + i, lightSourceSphereTransform );
        }
    }

    std::cout << __m_lightGroup->getChildCount() << std::endl;
    std::cout << "- pointLights" << std::endl;
    LightSourcePoint pointLights[0];
    // LightSourcePoint pointLights[1] = { LightSourcePoint()};
    // pointLights[0].setPosition(optix::make_float3(p_tc.plPosX, p_tc.plPosY, p_tc.plPosZ));
    // pointLights[0].setColor(optix::make_float3(p_tc.plColR, p_tc.plColG, p_tc.plColB));
    // pointLights[0].setFluxTotal(sqrtf(2.f) * p_tc.pointLightIntensity);
    // pointLights[1].setPosition(optix::make_float3(500.0f, 1000.f, 10.0f));
    // pointLights[1].setColor(optix::make_float3(1.0f, 0.5f, 0.0f));
    // pointLights[1].setFluxTotal(farPlane * 50.f);

    optix::Buffer pointLightBuffer = __m_context->createBuffer ( RT_BUFFER_INPUT );
    pointLightBuffer->setFormat ( RT_FORMAT_USER );
    pointLightBuffer->setElementSize ( sizeof ( LightSourcePoint ) );
    unsigned int numPointLights = sizeof ( pointLights ) / sizeof ( LightSourcePoint );
    pointLightBuffer->setSize ( numPointLights );
    std::cout << "PointLights: " << numPointLights << std::endl;
    memcpy ( pointLightBuffer->map(), pointLights, sizeof ( pointLights ) );
    pointLightBuffer->unmap();
    __m_context["pointLights"]->set ( pointLightBuffer );
    __m_directionalLight_direction = optix::normalize ( SUN_STARTPOSITION );
    __m_directionalLight_fluxRGB = calculateSunRGB ( &__m_directionalLight_direction.x, 1.f );
    __m_context["directionalLight_direction"]->setFloat ( __m_directionalLight_direction );
    __m_context["directionalLight_fluxRGB"]->setFloat ( __m_directionalLight_fluxRGB );
}



void Scene::setSun ( optix::Context &p_context, std::string &p_brightnessName, std::string &p_scaleName, float p_phi, float p_theta, float p_sunBrightness, float p_sunScale ) {
    __m_directionalLight_direction = optix::normalize ( optix::make_float3 ( cosf ( p_theta ) * cosf ( p_phi ), cosf ( p_theta ) * sinf ( p_phi ), sinf ( p_theta ) ) );
    p_context["directionalLight_direction"]->setFloat ( __m_directionalLight_direction );
    __m_directionalLight_fluxRGB = calculateSunRGB ( &__m_directionalLight_direction.x, p_sunBrightness );
    p_context["directionalLight_fluxRGB"]->setFloat ( __m_directionalLight_fluxRGB );
    p_context[p_scaleName]->setFloat ( p_sunScale );
    p_context[p_brightnessName]->setFloat ( p_sunBrightness );
}


void Scene::setEnvironmentMap ( const std::string &p_texturePath, float p_envGamma ) {
    /* Load HDR environment map and prepare it for importance sampling */
    optix::TextureSampler hdrEnvMap = loadEnvironmentTexture ( __m_context, p_texturePath, 1.f, 1.f, 1.f );
    optix::Buffer hdrEnvMapBuffer;
    hdrEnvMapBuffer = hdrEnvMap->getBuffer ( 0, 0 );
    RTsize w = 0;
    RTsize h = 0;
    hdrEnvMapBuffer->getSize ( w, h );
    optix::Buffer envMap_importance = __m_context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT, w, h );
    optix::float4 *h_envMap = ( optix::float4 * ) ( hdrEnvMapBuffer->map() );
    float *h_envMap_importance = ( float * ) ( envMap_importance->map() );

    /* do gamma stuff */
    for ( int v = 0; v < h; ++v ) {
        for ( int u = 0; u < w; ++u ) {
            h_envMap[v * w + u].x = powf ( h_envMap[v * w + u].x, p_envGamma );
            h_envMap[v * w + u].y = powf ( h_envMap[v * w + u].y, p_envGamma );
            h_envMap[v * w + u].z = powf ( h_envMap[v * w + u].z, p_envGamma );
        }
    }

    /* do precomputation stuff (blur the envMap luminance) */
    for ( int v = 0; v < h; ++v ) {
        float sinTheta = sin ( M_PIf * ( ( float ) ( v ) / ( float ) ( h ) ) );

        for ( int u = 0; u < w; ++u ) {
            unsigned int numVals = 0;
            float curVal = 0;

            for ( int kv = fmaxf ( 0, v - 1 ); kv <= fminf ( v + 1, h - 1 ); ++kv ) {
                for ( int ku = fmaxf ( 0, u - 1 ); ku <= fminf ( u + 1, w - 1 ); ++ku ) {
                    numVals++;
                    curVal += optix::luminance ( optix::make_float3 ( h_envMap[kv * w + ku] ) );
                }
            }

            curVal /= numVals;
            curVal *= sinTheta;
            h_envMap_importance[v * w + u] = curVal;
        }
    }

    Distribution2D envMapDistribution ( __m_context, h_envMap_importance, w, h );
    hdrEnvMapBuffer->unmap();
    envMap_importance->unmap();
    __m_context["bg_map"]->setTextureSampler ( hdrEnvMap );
}


optix::Group Scene::getLightGroup() {
    return __m_lightGroup;
}

optix::Group Scene::getSceneGroup() {
    return __m_sceneGroup;
}

optix::Transform Scene::getSceneTransform() {
    return __m_sceneTransform;
}

optix::float3 Scene::getSceneSize() {
    return __m_sceneSize;
}

float Scene::getMaxSceneExtent() {
    return __m_maxSize;
}

