#include <SixAxis.h>

#include <math.h>

#include <iostream>

void SixAxis::handleAnalogSticks ( sf::Event &p_event ) {
    // movement
    int sig = 0;
#define JOYMOVE_EPSILON 4.0

    if ( p_event.joystickMove.position > JOYMOVE_EPSILON ) {
        sig = 1;

    } else if ( p_event.joystickMove.position < -JOYMOVE_EPSILON ) {
        sig = -1;
    }

    float absJoyPos = ( fabsf ( p_event.joystickMove.position ) - JOYMOVE_EPSILON ) * ( 100.0 / ( 100.0 - JOYMOVE_EPSILON ) );
    float valHor_movement = sig * __m_controllerSensitivity_movement * absJoyPos / 100.f;
    float valVer_movement = -sig * __m_controllerSensitivity_movement * absJoyPos / 100.f;
    float valHor_rotation = sig * __m_controllerSensitivity_rotation * absJoyPos / 100.f;
    float valVer_rotation = -sig * __m_controllerSensitivity_rotation * absJoyPos / 100.f;
	
	//std::cout << valHor_movement << ", " << valVer_movement << ", " << valHor_rotation << ", " << valVer_rotation << std::endl;

    switch ( p_event.joystickMove.axis ) {
        case ENUM_SIXAXIS_LEFT_ANALOG_HORIZONTAL: {
                __m_controllerVals[ENUM_CONTROLLER_VAL_STRAFE] = valHor_movement;
                break;
            }

        case ENUM_SIXAXIS_LEFT_ANALOG_VERTICAL: {
                if ( __m_buttonsActive[ENUM_SIXAXIS_L1] ) {
                    __m_controllerVals[ENUM_CONTROLLER_VAL_UP] = valVer_movement;

                } else {
                    __m_controllerVals[ENUM_CONTROLLER_VAL_FORWARD] = valVer_movement;
                }

                break;
            }

        case ENUM_SIXAXIS_RIGHT_ANALOG_HORIZONTAL: {
                if ( __m_buttonsActive[ENUM_SIXAXIS_R1] ) {
                    __m_controllerVals[ENUM_CONTROLLER_VAL_ROLL] = -valHor_rotation;

                } else {
                    __m_controllerVals[ENUM_CONTROLLER_VAL_PHI] = valHor_rotation;
                }

                break;
            }

        case ENUM_SIXAXIS_RIGHT_ANALOG_VERTICAL: {
                __m_controllerVals[ENUM_CONTROLLER_VAL_THETA] = -valVer_rotation;
                break;
            }

        default:
            break;
    }
}


void SixAxis::handleButtonPress ( sf::Event &p_event ) {
    // button pressed
    __m_buttonsActive[p_event.joystickButton.button] = true;

	if ( p_event.joystickButton.button == ENUM_SIXAXIS_R1 ) { // change phi to roll
        __m_controllerVals[ENUM_CONTROLLER_VAL_ROLL] = -__m_controllerVals[ENUM_CONTROLLER_VAL_PHI];
        __m_controllerVals[ENUM_CONTROLLER_VAL_PHI] = 0.f;
    }

    if ( p_event.joystickButton.button == ENUM_SIXAXIS_L1 ) { // change forward to up
        __m_controllerVals[ENUM_CONTROLLER_VAL_UP] = __m_controllerVals[ENUM_CONTROLLER_VAL_FORWARD];
        __m_controllerVals[ENUM_CONTROLLER_VAL_FORWARD] = 0.f;
    }

    if ( p_event.joystickButton.button == ENUM_SIXAXIS_X ) { 
        __m_storeKeyFrame = true;
    }

    if ( p_event.joystickButton.button == ENUM_SIXAXIS_START) {
        __m_writeKeyFramesToDisk = true;
    }
}


void SixAxis::handleButtonRelease ( sf::Event &p_event ) {
    // button released
    //std::cout << "Joystick " << event.JoyButton.JoystickId << ", Button " << event.JoyButton.Button << " released" << std::endl;
	__m_buttonsActive[p_event.joystickButton.button] = false;

	if ( p_event.joystickButton.button == ENUM_SIXAXIS_R1 ) { // change Azimuth to Roll
        __m_controllerVals[ENUM_CONTROLLER_VAL_PHI] = -__m_controllerVals[ENUM_CONTROLLER_VAL_ROLL];
        __m_controllerVals[ENUM_CONTROLLER_VAL_ROLL] = 0.f;
    }

    if ( p_event.joystickButton.button == ENUM_SIXAXIS_L1 ) { // change forward to up
        __m_controllerVals[ENUM_CONTROLLER_VAL_FORWARD] = __m_controllerVals[ENUM_CONTROLLER_VAL_UP];
        __m_controllerVals[ENUM_CONTROLLER_VAL_UP] = 0.f;
    }

}




SixAxis::SixAxis() {    
    __m_storeKeyFrame = false;

    __m_buttonsActive.resize ( ENUM_SIXAXIS_NUM_BUTTONS );

    for ( int i = 0; i < ENUM_SIXAXIS_NUM_BUTTONS; ++i ) {
        __m_buttonsActive[i] = false;
    }

    for ( int i = 0; i < ENUM_NUM_CONTROLLER_VALS; ++i ) {
        __m_controllerVals[i] = 0;
    }
}


SixAxis::~SixAxis() {
}


bool SixAxis::isButtonPressed ( sfml_sixaxis_buttons p_button ) {
    return __m_buttonsActive[p_button];
}


void SixAxis::setButtonActive ( sfml_sixaxis_buttons p_button ) {
    __m_buttonsActive[p_button] = true;
}


bool SixAxis::setButtonInactive ( sfml_sixaxis_buttons p_button ) {
    __m_buttonsActive[p_button] = false;
}


void SixAxis::resetTriggers() {
    __m_storeKeyFrame = false;
    __m_writeKeyFramesToDisk = false;
}

void SixAxis::handleInteraction ( sf::Event &p_event ) {

    if ( p_event.type == sf::Event::JoystickMoved ) {
        handleAnalogSticks ( p_event );
		
    } else if ( p_event.type == sf::Event::JoystickButtonPressed ) {
        handleButtonPress ( p_event );

    } else if ( p_event.type == sf::Event::JoystickButtonReleased ) {
        handleButtonRelease ( p_event );
    }
}



void SixAxis::setControllerSensitivity_movement ( float p_value ) {
    __m_controllerSensitivity_movement = p_value;
}

void SixAxis::setControllerSensitivity_rotation ( float p_value ) {
    __m_controllerSensitivity_rotation = p_value;
}


float SixAxis::getValue ( enum_controller_vals p_controllerVal ) {
    return __m_controllerVals[p_controllerVal];
}



bool SixAxis::getStoreKeyFrame() {
    return __m_storeKeyFrame;

}

bool SixAxis::getWriteKeyFramesToDisk() {
    return __m_writeKeyFramesToDisk;
}


bool SixAxis::movement() {
    for ( int i = 0; i < ENUM_NUM_CONTROLLER_VALS; ++i ) {
        if ( __m_controllerVals[i] != 0.f ) {
            return true;
        }
    }

    return false;
}