#include "TrialConfig.h"
#include <json/json.h>
#include <iostream>
#include <string>
#include <fstream>

void TrialConfig::print() {
	std::cout << "Trial config: " << std::endl;

	std::cout << "Scene: " << sceneFile << std::endl;
	std::cout << "Celldat: " << celldat << std::endl;
	std::cout << "Fifar: " << fifar << std::endl;
	std::cout << "Output: " << outputPath << std::endl;
	std::cout << "Pointlight pos: " << plPosX << ", " << plPosY << ", " << plPosZ << std::endl;
	std::cout << "Pointlight col: " << plColR << ", " << plColG << ", " << plColB << std::endl;
	std::cout << "Pointlight intensity: " << pointLightIntensity << std::endl;
	
	std::cout << "id: " << id << std::endl;
	std::cout << "reference: " << reference << std::endl;
	std::cout << "frames: " << frames << std::endl;
	std::cout << "render_each: " << render_each << std::endl;
	std::cout << "blur: " << blur << std::endl;
	std::cout << "samples: " << samples << std::endl;
	std::cout << "caching_probability: " << caching_probability << std::endl;
	std::cout << "envRot_phi: " << envRot_phi << std::endl;
}


void TrialConfig::read(std::string p_file) {
	Json::Value root;

	std::cout << "Reading file: " << p_file << std::endl;
	std::ifstream file(p_file.c_str());
	file >> root;



	sceneFile = root.get("scene", "").asString();
	celldat = root.get("celldat", "").asString();
	fifar = root.get("fifar", "").asString();
	outputPath = root.get("output_path", "").asString();

	const Json::Value pointlightPos = root["pointlight"]["pos"];
	plPosX = pointlightPos[0].asFloat();
	plPosY = pointlightPos[1].asFloat();
	plPosZ = pointlightPos[2].asFloat();

	const Json::Value pointlightCol = root["pointlight"]["col"];
	plColR = pointlightCol[0].asFloat();	
	plColG = pointlightCol[1].asFloat();
	plColB = pointlightCol[2].asFloat();

	pointLightIntensity = root["pointlight"]["intensity"].asFloat();
	
	
	id = root["id"].asUInt();
	reference = root["reference"].asBool();
	frames = root["frames"].asUInt();
	render_each = root["render_each"].asUInt();
	blur = root["blur"].asFloat();
	samples = root["samples"].asUInt();
	caching_probability = root["caching_probability"].asFloat();
	envRot_phi = root["envRot_phi"].asFloat();
}





