/*
 * helpers.h
 *
 *  Created on: May 13, 2011
 *      Author: troth22s
 */


#ifndef HELPERS_H_
#define HELPERS_H_

#include <constants.h>
#include <optixu/optixu_math_namespace.h>
#include <math.h>
#include <iostream>





#ifndef __CUDACC__

#include <optixu/optixpp_namespace.h>
#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>
#include <sys/time.h>

std::string generatePtxPath ( const std::string &base ) {
    std::cout << "cuda_compile_ptx_1_generated_" + base + ".ptx" << std::endl;
    return "cuda_compile_ptx_1_generated_" + base + ".ptx";
}


/** @brief create an instance of p_geometry with the desired material, color and transformation */
optix::Transform createObject ( optix::Context p_context, optix::Geometry p_geometry, optix::Material p_material, float p_exponent, optix::float3 p_color, optix::float3 p_translate, optix::float3 p_scale, float p_rotate ) {
    optix::Transform transform = p_context->createTransform();
    optix::Matrix4x4 scaleMatrix = optix::Matrix4x4::scale ( p_scale );
    optix::Matrix4x4 translateMatrix = optix::Matrix4x4::translate ( p_translate );
    optix::Matrix4x4 rotateMatrix = optix::Matrix4x4::rotate ( p_rotate, optix::make_float3 ( 0.0f, 1.0f, 0.0f ) );
    transform->setMatrix ( 0, ( translateMatrix * rotateMatrix * scaleMatrix ).getData(), ( translateMatrix * rotateMatrix * scaleMatrix ).inverse().getData() );
    optix::GeometryInstance gi = p_context->createGeometryInstance();
    gi->setMaterialCount ( 1 );
    gi->setGeometry ( p_geometry );
    gi->setMaterial ( 0, p_material );
    gi["reflectivity"]->setFloat ( 0.f );
    gi["phong_exp"]->setFloat ( 0.f );
    //    gi[ "illum"        ]->setInt( mp.illum );
    gi["alpha"]->setFloat ( 0.f );
    gi["kd"]->setFloat ( 0.f, 0.f, 0.f );
    gi["ks"]->setFloat ( 0.f, 0.f, 0.f );
    gi["tf"]->setFloat ( 0.f, 0.f, 0.05f );
    gi["ni"]->setFloat ( 1.5f );
    //    gi[ "ambient_map"  ]->setTextureSampler( mp.ambient_map );
    optix::TextureSampler sampler = p_context->createTextureSampler();
    sampler->setWrapMode ( 0, RT_WRAP_REPEAT );
    sampler->setWrapMode ( 1, RT_WRAP_REPEAT );
    sampler->setWrapMode ( 2, RT_WRAP_REPEAT );
    sampler->setIndexingMode ( RT_TEXTURE_INDEX_NORMALIZED_COORDINATES );
    sampler->setReadMode ( RT_TEXTURE_READ_NORMALIZED_FLOAT );
    sampler->setMaxAnisotropy ( 1.0f );
    sampler->setMipLevelCount ( 1u );
    sampler->setArraySize ( 1u );
    // Create buffer with single texel set to default_color
    optix::Buffer buffer = p_context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_UNSIGNED_BYTE4, 1u, 1u );
    unsigned char *buffer_data = static_cast<unsigned char *> ( buffer->map() );
    buffer_data[0] = ( unsigned char ) optix::clamp ( ( int ) ( 255.0f ), 0, 255 );
    buffer_data[1] = ( unsigned char ) optix::clamp ( ( int ) ( 255.0f ), 0, 255 );
    buffer_data[2] = ( unsigned char ) optix::clamp ( ( int ) ( 255.0f ), 0, 255 );
    buffer_data[3] = 255;
    buffer->unmap();
    sampler->setBuffer ( 0u, 0u, buffer );
    // Although it would be possible to use nearest filtering here, we chose linear
    // to be consistent with the textures that have been loaded from a file. This
    // allows OptiX to perform some optimizations.
    sampler->setFilteringModes ( RT_FILTER_LINEAR, RT_FILTER_LINEAR, RT_FILTER_NONE );
    gi["diffuse_map"]->setTextureSampler ( sampler );
    gi["specular_map"]->setTextureSampler ( sampler );
    gi["dissolve_map"]->setTextureSampler ( sampler );
    gi["bump_map"]->setTextureSampler ( sampler );
    //	gi["color"]->setFloat(p_color);
    //	gi["exponent"]->setFloat(p_exponent);
    optix::GeometryGroup giGroup = p_context->createGeometryGroup();
    giGroup->setChildCount ( 1 );
    giGroup->setChild ( 0, gi );
    giGroup->setAcceleration ( p_context->createAcceleration ( "NoAccel", "NoAccel" ) );
    transform->setChild ( giGroup );
    //	gi["sphere"]->setFloat(optix::make_float4(0.f, 0.f, 0.f, 1.f));
    return transform;
}


optix::Transform createObject ( optix::Context p_context, optix::Geometry p_geometry, optix::Material p_material, optix::float3 p_color, optix::float3 p_translate, optix::float3 p_scale, float p_rotate ) {
    return createObject ( p_context, p_geometry, p_material, 1.f, p_color, p_translate, p_scale, p_rotate );
}





/** @brief add the elapsed timespan between p_start and p_end to p_elapsedSeconds */
void updateElapsedSeconds ( timespec &p_start, timespec &p_end, double &p_elapsedSeconds ) {
    if ( p_end.tv_nsec < p_start.tv_nsec ) {
        p_elapsedSeconds += ( 999999999 - p_start.tv_nsec + p_end.tv_nsec ) / 1000000000.0;

    } else {
        p_elapsedSeconds += ( p_end.tv_nsec - p_start.tv_nsec ) / 1000000000.0;
    }
}

#endif

#endif /* HELPERS_H_ */
