/*
 * NormalShader.cu
 *
 *  Created on: May 11, 2011
 *      Author: troth22s
 */

#include <optix.h>
#include <optixu/optixu_math_namespace.h>

#include <common_structs.h>
#include <random.h>
#include <tr1/float.h>

//#include <lightsources/CubeLight.cu>
#include <SphereLight.cu>
#include <LightSampler.cu>

#include <sobol_cuda.cu>
//#include <intersectionVars.cu>

#include <constants.h>




#include <optix_declaration_macros.h>

#ifndef __OPTIX_VARS_SHADING_INFO__
#define __OPTIX_VARS_SHADING_INFO__
__DECLARE_OPTIX_VARS_SHADING_INFO__
#endif

#ifndef __OPTIX_VARS_TEXTURES_DIFFUSE__
#define __OPTIX_VARS_TEXTURES_DIFFUSE__
__DECLARE_OPTIX_VARS_TEXTURES_DIFFUSE__
#endif

#ifndef __OPTIX_VARS_SAMPLER_SETTINGS__
#define __OPTIX_VARS_SAMPLER_SETTINGS__
__DECLARE_OPTIX_VARS_SAMPLER_SETTINGS__
#endif





class DiffuseBRDF {

  public:

    __device__ float pdf ( const optix::float3 &p_wo, const optix::float3 &p_wi, const optix::float3 &p_normal ) {
        return optix::dot ( p_wi, p_normal ) * INV_PI;
    }

    __device__ float sample_f ( const float3 &p_wo, float3 *p_wi, float3 &p_normal, float p_u1, float p_u2, float *p_pdf ) {
        // orthonormal basis for choosing the new sampling direction
        optix::Onb onb ( p_normal );
        optix::cosine_sample_hemisphere ( p_u1, p_u2, *p_wi );
        onb.inverse_transform ( *p_wi );
        *p_wi = optix::normalize ( *p_wi );
        return f ( p_wo, *p_wi, p_normal, p_pdf );
    }

    __device__ float f ( const float3 &p_wo, const float3 &p_wi, const float3 &p_normal, float *p_pdf ) {
        *p_pdf = pdf ( p_wo, p_wi, p_normal );
        return INV_PI;
    }

};



/* @brief closest hit program for path tracing of purely diffuse materials */
RT_PROGRAM void diffuse_shader() {
    float3 hit_point = ray.origin + t_hit * ray.direction;
    float3 geometricNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, geometric_normal ) );
    float3 shadingNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, shading_normal ) );
    /* front facing normals are necessary if we look at the primitive from behind */
    float3 shadingNormal_world_ff = optix::faceforward ( shadingNormal_world, -ray.direction, geometricNormal_world );
    float3 geometricNormal_world_ff = optix::faceforward ( geometricNormal_world, -ray.direction, geometricNormal_world );
    float3 diffTex = make_float3 ( tex2D ( diffuse_map, texcoord.x, texcoord.y ) ) * kd;
    float u1 = 0.f;
    float u2 = 0.f;
	prd_pt.kd = 1.f;

	prd_pt.shininess = 0.f;
	
    if ( useLdSampling_integrator ) {
        unsigned int dimension_u = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_BSDF_U;
        unsigned int dimension_v = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_BSDF_V;
        u1 = Sobol ( dimension_u, frame, prd_pt.seed );
        u2 = Sobol ( dimension_v, frame, prd_pt.seed );

    } else {
        u1 = rnd ( prd_pt.seed );
        u2 = rnd ( prd_pt.seed );
    }

    DiffuseBRDF dbrdf;
    prd_pt.texCol = diffTex;
    prd_pt.f = make_float3 ( dbrdf.sample_f ( -ray.direction, & ( prd_pt.direction ), shadingNormal_world_ff, u1, u2, & ( prd_pt.pdf ) ) );
    prd_pt.dotProduct = fmaxf ( 0.f, optix::dot ( prd_pt.direction, shadingNormal_world_ff ) );

    if ( optix::dot ( prd_pt.direction, geometricNormal_world_ff ) < 0.f ) {
        prd_pt.done = true;
    }

    if ( prd_pt.depth < 2 ) {
        prd_pt.radiance = sampleAllLightSources<DiffuseBRDF> ( hit_point, shadingNormal_world_ff, geometricNormal_world_ff, lightProb_diffuse, dbrdf );

    } else {
        prd_pt.radiance = sampleOneLightSource<DiffuseBRDF> ( hit_point, shadingNormal_world_ff, geometricNormal_world_ff, lightProb_diffuse, dbrdf );
    }

    prd_pt.directIllumination = false;
    prd_pt.normal = shadingNormal_world_ff;
    prd_pt.normal_geometric = geometricNormal_world_ff;
    prd_pt.origin = hit_point;
    prd_pt.hitDist = t_hit;
}

