/*
 * NormalShader.cu
 *
 *  Created on: May 11, 2011
 *      Author: troth22s
 */

#include <optix.h>
#include <optixu/optixu_math_namespace.h>

#include <common_structs.h>
#include <random.h>
#include <tr1/float.h>

//#include <lightsources/CubeLight.cu>
#include <SphereLight.cu>
#include <LightSampler.cu>

#include <sobol_cuda.cu>
//#include <intersectionVars.cu>

#include <constants.h>




#include <optix_declaration_macros.h>

#ifndef __OPTIX_VARS_SHADING_INFO__
#define __OPTIX_VARS_SHADING_INFO__
__DECLARE_OPTIX_VARS_SHADING_INFO__
#endif

#ifndef __OPTIX_VARS_TEXTURES_DIFFUSE__
#define __OPTIX_VARS_TEXTURES_DIFFUSE__
__DECLARE_OPTIX_VARS_TEXTURES_DIFFUSE__
#endif

#ifndef __OPTIX_VARS_TEXTURES_DISSOLVE__
#define __OPTIX_VARS_TEXTURES_DISSOLVE__
__DECLARE_OPTIX_VARS_TEXTURES_DISSOLVE__
#endif

#ifndef __OPTIX_VARS_TEXTURES_SPECULAR__
#define __OPTIX_VARS_TEXTURES_SPECULAR__
__DECLARE_OPTIX_VARS_TEXTURES_SPECULAR__
#endif

#ifndef __OPTIX_VARS_SAMPLER_SETTINGS__
#define __OPTIX_VARS_SAMPLER_SETTINGS__
__DECLARE_OPTIX_VARS_SAMPLER_SETTINGS__
#endif




/* setting this will result in normal orientation
 * being accounted for when calculation refraction */
#define NORMAL_REFRACTION

/* @brief closest hit program for path tracing of glass materials (transparency, fresnel refraction/reflection ) */
RT_PROGRAM void glass_shader() {
 prd_pt.shininess = 1.f;

    if (prd_pt.kd == 1.f) {
        prd_pt.done = true;
    }
    float3 hit_point = ray.origin + t_hit * ray.direction;
    float3 geometricNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, geometric_normal ) );
    float3 geometricNormal_world_ff = optix::faceforward ( geometricNormal_world, -ray.direction, geometricNormal_world );
    float3 shadingNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, shading_normal ) );
    /* front facing normals are necessary if we look at the primitive from behind */
    float3 shadingNormal_world_ff = optix::faceforward ( shadingNormal_world, -ray.direction, geometricNormal_world );

    if (prd_pt.done) {
        prd_pt.texCol = tf;
        prd_pt.origin = hit_point;
        prd_pt.hitDist = t_hit;
        prd_pt.normal = shadingNormal_world_ff;
        prd_pt.normal_geometric = geometricNormal_world_ff;
        return;
    }
    /* check for alpha stuff */
    optix::float3 dissolveCol = make_float3 ( tex2D ( dissolve_map, texcoord.x, texcoord.y ) );
    float kTransparent = alpha * dissolveCol.x;

    /* if transparency is 100% without anything filtered out, we have already handeled it in the alpha_tester,
     * so it does not have to be handeled here again. kTransparent = 1.f does just this
     */
    if ( ni == 1.f && tf.x == 1.f && tf.y == 1.f && tf.z == 1.f ) {
        kTransparent = 1.f;
    }

    unsigned int dimension = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_BSDF_TRANSPARENCY;

    if ( Sobol ( dimension, frame, prd_pt.seed ) > kTransparent ) {
        /* if no refraction is present, only filter the color and pass through */
        if ( ni == 1.f ) {
            prd_pt.f = tf;
            prd_pt.direction = ray.direction;
            prd_pt.dotProduct = optix::dot ( prd_pt.direction, -shadingNormal_world_ff );
            //prd_pt.depth = fmaxf(prd_pt.depth - 1, 1); // just transmit this ray with some attenuation
            prd_pt.texCol = tf; // is this correct? ;)
            prd_pt.normal = -shadingNormal_world_ff;
            prd_pt.normal_geometric = -geometricNormal_world_ff;

        } else {
            // first handle transparency
            float theDot = optix::dot ( -ray.direction, shadingNormal_world );
            float ni1 = 1.f;
            float ni2 = ni;

            if ( theDot < 0.f ) {
                ni1 = ni;
                ni2 = 1.f;
                theDot = optix::dot ( -ray.direction, -shadingNormal_world );
            }

            /* check if ray is refracted or reflected */
            bool rayRefracted = optix::refract ( prd_pt.direction, ray.direction, shadingNormal_world, ni );

            if ( rayRefracted ) {
                float theDot2 = optix::dot ( prd_pt.direction, -shadingNormal_world );

                if ( theDot2 < 0.f ) {
                    theDot2 = optix::dot ( prd_pt.direction, shadingNormal_world );
                }

                float theReflectivity = fresnelDiel ( theDot, theDot2, ni1, ni2 );
                dimension = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_BSDF_FRESNEL;

                if ( Sobol ( dimension, frame, prd_pt.seed ) < theReflectivity ) { // reflect!
                    prd_pt.f = make_float3 ( 1.f );
                    prd_pt.direction = optix::reflect ( ray.direction, shadingNormal_world_ff );
                    prd_pt.dotProduct = optix::dot ( prd_pt.direction, shadingNormal_world_ff );
                    prd_pt.normal = shadingNormal_world_ff;
                    prd_pt.normal_geometric = geometricNormal_world_ff;

                } else {
                    prd_pt.f = tf;
                    prd_pt.dotProduct = theDot2;
                    prd_pt.normal = -shadingNormal_world_ff;
                    prd_pt.normal_geometric = -geometricNormal_world_ff;
                }

            } else {
                prd_pt.f = make_float3 ( 1.f ); //, 0.f, 0.f);
                prd_pt.direction = optix::reflect ( ray.direction, shadingNormal_world_ff );
                prd_pt.dotProduct = optix::dot ( prd_pt.direction, shadingNormal_world_ff );
                prd_pt.normal = shadingNormal_world_ff;
                prd_pt.normal_geometric = geometricNormal_world_ff;
            }
        }

        prd_pt.pdf = 1.f;
        prd_pt.texCol = tf;
        prd_pt.directIllumination = true;
    }

    prd_pt.origin = hit_point;
    prd_pt.hitDist = t_hit;
}

