/*
 * NormalShader.cu
 *
 *  Created on: May 11, 2011
 *      Author: troth22s
 */

#include <optix.h>
#include <optixu/optixu_math_namespace.h>

#include <common_structs.h>
#include <random.h>
#include <tr1/float.h>

#include <ConfigurationMacros.h>

#include <LightSampler.cu>

#include <sobol_cuda.cu>
//#include <intersectionVars.cu>

#include <constants.h>


#include <optix_declaration_macros.h>

#ifndef __OPTIX_VARS_SHADING_INFO__
#define __OPTIX_VARS_SHADING_INFO__
__DECLARE_OPTIX_VARS_SHADING_INFO__
#endif

#ifndef __OPTIX_VARS_SAMPLER_SETTINGS__
#define __OPTIX_VARS_SAMPLER_SETTINGS__
__DECLARE_OPTIX_VARS_SAMPLER_SETTINGS__
#endif



#ifndef __OPTIX_VARS_TEXTURES_SPECULAR__
#define __OPTIX_VARS_TEXTURES_SPECULAR__
__DECLARE_OPTIX_VARS_TEXTURES_SPECULAR__
#endif



class MirrorBRDF {

  public:

    __device__ float pdf ( const optix::float3 &p_wo, const optix::float3 &p_wi, const optix::float3 &p_normal ) {
        return 1.f;//optix::dot ( p_wi, p_normal ) * INV_PI;
    }

    __device__ float sample_f ( const float3 &p_wo, float3 *p_wi, float3 &p_normal, float p_u1, float p_u2, float *p_pdf ) {
        prd_pt.f = make_float3 ( 1.f );
        *p_wi = optix::reflect ( ray.direction, p_normal );
        return f(p_wo, *p_wi, p_normal, p_pdf);
    }

    __device__ float f ( const float3 &p_wo, const float3 &p_wi, const float3 &p_normal, float *p_pdf ) {
        *p_pdf = pdf ( p_wo, p_wi, p_normal );
        return 1.f;
    }

};


/* @brief closest hit program for path tracing of mirror materials */
RT_PROGRAM void mirror_shader() {
    // rtPrintf("MIRROR SHADER!\n");
    if (prd_pt.kd == 1.f) {
        prd_pt.done = true;
    }

    float3 hit_point = ray.origin + t_hit * ray.direction;
    float3 specTex = make_float3 ( tex2D ( specular_map, texcoord.x, texcoord.y ) ) * ks;
    float3 geometricNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, geometric_normal ) );
    float3 geometricNormal_world_ff = optix::faceforward ( geometricNormal_world, -ray.direction, geometricNormal_world );

    float3 shadingNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, shading_normal ) );
    /* front facing normals are necessary if we look at the primitive from behind */
    float3 shadingNormal_world_ff = optix::faceforward ( shadingNormal_world, -ray.direction, geometricNormal_world );

    if (prd_pt.done) {
        prd_pt.origin = hit_point;
        prd_pt.texCol = specTex;
        prd_pt.hitDist = t_hit;
        prd_pt.normal_geometric = geometricNormal_world_ff;
        prd_pt.normal = shadingNormal_world_ff;
        return;
    }

    prd_pt.shininess = 1.f;
    prd_pt.kd = 0.f;

 
   


    prd_pt.f = specTex;
    prd_pt.direction = optix::reflect ( ray.direction, shadingNormal_world_ff );
    prd_pt.dotProduct = optix::dot ( prd_pt.direction, shadingNormal_world_ff );
    prd_pt.normal = shadingNormal_world_ff;
    prd_pt.normal_geometric = geometricNormal_world_ff;
    prd_pt.pdf = 1.f;
    prd_pt.texCol = specTex;
    
    // if (!prd_pt.cachingSearch) {
    //  if (RENDER_CAUSTICS || prd_pt.depth < 1) {
    //      MirrorBRDF mbrdf;
    //      if ( prd_pt.depth < 1 ) {
    //          prd_pt.radiance = sampleAllLightSources<MirrorBRDF> ( hit_point, shadingNormal_world_ff, geometricNormal_world_ff, 0.f, mbrdf );

    //      } else {
    //          prd_pt.radiance = sampleOneLightSource<MirrorBRDF> ( hit_point, shadingNormal_world_ff, geometricNormal_world_ff, 0.f, mbrdf );
    //      }
    //  }
    // }
    
    prd_pt.directIllumination = true;
    
//  if (RENDER_CAUSTICS || prd_pt.depth < 1) {
//      prd_pt.directIllumination = true;
//  } else {
//      prd_pt.directIllumination = false;
//  }
//  
        
    prd_pt.origin = hit_point;
    prd_pt.hitDist = t_hit;
}

