/*
 * box.cu
 *
 *  Created on: May 17, 2011
 *      Author: troth22s
 */

#include <optix_world.h>

rtDeclareVariable ( float3, boxmin, , );
rtDeclareVariable ( float3, boxmax, , );

rtDeclareVariable ( float3, geometric_normal, attribute geometric_normal, );
rtDeclareVariable ( float3, shading_normal, attribute shading_normal, );
rtDeclareVariable ( optix::Ray, ray, rtCurrentRay, );
rtDeclareVariable ( float, intersectedPrimitiveIndex, attribute intersectedPrimitiveIdx, );

rtDeclareVariable ( float3, texU, attribute texU, );
rtDeclareVariable ( float3, texV, attribute texV, );

rtDeclareVariable ( float3, texcoord, attribute texcoord, );

__device__ float3 box_normal ( float t ) {
    float3 t0 = ( boxmin - ray.origin ) / ray.direction;
    float3 t1 = ( boxmax - ray.origin ) / ray.direction;
    float3 neg = make_float3 ( t == t0.x ? 1 : 0, t == t0.y ? 1 : 0, t == t0.z ? 1 : 0 );
    float3 pos = make_float3 ( t == t1.x ? 1 : 0, t == t1.y ? 1 : 0, t == t1.z ? 1 : 0 );
    return pos - neg;
}

RT_PROGRAM void box_intersect ( int ) {
    texcoord = make_float3 ( 0.f );
    float3 t0 = ( boxmin - ray.origin ) / ray.direction;
    float3 t1 = ( boxmax - ray.origin ) / ray.direction;
    float3 near = fminf ( t0, t1 );
    float3 far = fmaxf ( t0, t1 );
    float tmin = fmaxf ( near );
    float tmax = fminf ( far );

    if ( tmin <= tmax ) {
        bool check_second = true;

        if ( rtPotentialIntersection ( tmin ) ) {
            texU = make_float3 ( 0.f );
            texV = make_float3 ( 0.f );
            intersectedPrimitiveIndex = 0;
            //texcoord = make_float3(0.0f);
            shading_normal = geometric_normal = box_normal ( tmin );

            if ( rtReportIntersection ( 0 ) )
            { check_second = false; }
        }

        if ( check_second ) {
            if ( rtPotentialIntersection ( tmax ) ) {
                texU = make_float3 ( 0.f );
                texV = make_float3 ( 0.f );
                intersectedPrimitiveIndex = 0;
                //texcoord = make_float3(0.0f);
                shading_normal = geometric_normal = box_normal ( tmax );
                rtReportIntersection ( 0 );
            }
        }
    }
}

RT_PROGRAM void box_bounds ( int, float result[6] ) {
    optix::Aabb *aabb = ( optix::Aabb * ) result;
    aabb->set ( boxmin, boxmax );
}
