/*
 * sphere.cu
 *
 *  Created on: May 11, 2011
 *      Author: troth22s
 */

#include <optix_world.h>

rtDeclareVariable ( float4, sphere, , );
rtDeclareVariable ( float3, geometric_normal, attribute geometric_normal, );
rtDeclareVariable ( float3, shading_normal, attribute shading_normal, );
rtDeclareVariable ( float3, texcoord, attribute texcoord, );
rtDeclareVariable ( float, intersectedPrimitiveIndex, attribute intersectedPrimitiveIdx, );

rtDeclareVariable ( optix::Ray, ray, rtCurrentRay, );
rtDeclareVariable ( float3, texU, attribute texU, );
rtDeclareVariable ( float3, texV, attribute texV, );




/** @brief bounding box program for spheres */
RT_PROGRAM void sphere_bounds ( int, float result[6] ) {
    optix::float3 cen = optix::make_float3 ( sphere );
    optix::float3 rad = optix::make_float3 ( sphere.w );
    optix::Aabb *aabb = ( optix::Aabb * ) result;

    if ( rad.x > 0.0f && !isinf ( rad.x ) ) {
        aabb->m_min = cen - rad;
        aabb->m_max = cen + rad;

    } else {
        aabb->invalidate();
    }
}

/** @brief intersection program for spheres, including refinement (whatever that is ;)) */
RT_PROGRAM void sphere_intersect ( int primitiveIndex ) {
    float3 center = make_float3 ( sphere );
    float3 O = ray.origin - center;
    float3 D = ray.direction;
    float radius = sphere.w;
    float b = optix::dot ( O, D );
    float c = optix::dot ( O, O ) - radius * radius;
    float disc = b * b - c;

    if ( disc > 0.0f ) {
        float sdisc = sqrtf ( disc );
        float root1 = ( -b - sdisc );
        bool do_refine = false;
        float root11 = 0.0f;

        if ( do_refine ) {
            // refine root1
            float3 O1 = O + root1 * ray.direction;
            b = optix::dot ( O1, D );
            c = optix::dot ( O1, O1 ) - radius * radius;
            disc = b * b - c;

            if ( disc > 0.0f ) {
                sdisc = sqrtf ( disc );
                root11 = ( -b - sdisc );
            }
        }

        bool check_second = true;

        if ( rtPotentialIntersection ( root1 + root11 ) ) {
            intersectedPrimitiveIndex = 0;
            texU = make_float3 ( 0.f );
            texV = make_float3 ( 0.f );
            texcoord = make_float3 ( 0.f );
            shading_normal = geometric_normal = ( O + ( root1 + root11 ) * D ) / radius;

            if ( rtReportIntersection ( 0 ) )
            { check_second = false; }
        }

        if ( check_second ) {
            float root2 = ( -b + sdisc ) + ( do_refine ? root1 : 0 );

            if ( rtPotentialIntersection ( root2 ) ) {
                intersectedPrimitiveIndex = 0;
                texU = make_float3 ( 0.f );
                texV = make_float3 ( 0.f );
                texcoord = make_float3 ( 0.f );
                shading_normal = geometric_normal = ( O + root2 * D ) / radius;
                rtReportIntersection ( 0 );
            }
        }
    }
}
