/*
 * PinholeCamera.cu
 *
 *  Created on: May 10, 2011
 *      Author: troth22s
 */




#include <random.h>

#include <helpers.h>

#include <common_structs.h>

#include <constants.h>

#include <optix_declaration_macros.h>

#ifndef __OPTIX_VARS_CAMERA__
#define __OPTIX_VARS_CAMERA__
__DECLARE_OPTIX_VARS_CAMERA__
#endif

#ifndef __OPTIX_VARS_FRAMEINFO__
#define __OPTIX_VARS_FRAMEINFO__
__DECLARE_OPTIX_VARS_FRAMEINFO__
#endif

#ifndef __OPTIX_VARS_SAMPLER_SETTINGS__
#define __OPTIX_VARS_SAMPLER_SETTINGS__
__DECLARE_OPTIX_VARS_SAMPLER_SETTINGS__
#endif

#ifndef __OPTIX_VARS_OUTPUT_BUFFERS__
#define __OPTIX_VARS_OUTPUT_BUFFERS__
__DECLARE_OPTIX_VARS_OUTPUT_BUFFERS__
#endif

#ifndef __OPTIX_VARS_RAY_TYPES__
#define __OPTIX_VARS_RAY_TYPES__
__DECLARE_OPTIX_VARS_RAY_TYPES__
#endif


#ifndef __OPTIX_VARS_GEOMETRY_INFO__
#define __OPTIX_VARS_GEOMETRY_INFO__
__DECLARE_OPTIX_VARS_GEOMETRY_INFO__
#endif

#include <sobol_cuda.cu>

enum enum_projection_types {

    ENUM_PROJECTION_TYPE_PERSPECTIVE = 0,
    ENUM_PROJECTION_TYPE_ORTHOGRAPHIC = 1

};


namespace rti_internal_register {
    __device__ unsigned reg_device_id;
}


rtDeclareVariable ( unsigned int, addRays, , ); // necessary to correctly count rays when performing separate rendering passes consecutively


RT_PROGRAM void exception_program ( void ) {
    //  float widthInterval = (float)displayWidth / (float)launch_dim.x;
    //  float heightInterval = (float)displayHeight / (float)launch_dim.y;
    //  launch_index = make_uint2((unsigned int)((float)launch_index.x * (widthInterval)),
    //                        (unsigned int)((float)launch_index.y * (heightInterval)));
    //  if (rtGetExceptionCode() == RT_EXCEPTION_STACK_OVERFLOW) {
    optix::float4 theColor = make_float4 ( 100000.f, 0.f, 100000.f, 0.f );
    output_buffer_directLight[launch_index] = theColor;
    //      fillBuffer<optix::float4>(launch_index.x, launch_index.y, (unsigned int) widthInterval, (unsigned int) heightInterval, theColor, output_buffer_directLight);
    //  } else {
    rtPrintExceptionDetails();
    //  }
}

__device__ void reInitializePrd ( PerRayData_pt &p_prd, const float3 &p_origin, const float3 &p_direction, const unsigned int &p_seed, const unsigned int &p_uniformSeed ) {
    p_prd.directIllumination = true;
    p_prd.depth = 0;
    p_prd.seed = p_seed;
    p_prd.uniformSeed = p_uniformSeed;
    p_prd.done = false;
    p_prd.origin = p_origin;
    p_prd.direction = p_direction;
    p_prd.background = false;
    p_prd.dotProduct = 0.f;
    p_prd.pdf = 1.f;
    p_prd.f = make_float3 ( 0.f );
    p_prd.texCol = make_float3 ( 0.f );
    p_prd.radiance = make_float3 ( 0.f );
    p_prd.kd = 0.f;
    p_prd.frame = numPreviousFrames;
}

RT_PROGRAM void pt_camera() {
    optix::float3 directLightingResult = make_float3 ( 0.f );
    optix::float3 indirectLightingResult = make_float3 ( 0.f );
    optix::float3 textureResult = make_float3 ( 0.f );
    optix::float3 rayDirectionResult = make_float3 ( 0.f );
    float depthResult = 0.f;
    optix::float3 normalResult = make_float3 ( 0.f );
    float backgroundPercentage = 0.f;
    unsigned int seed = 0;
    unsigned int uniformSeed = 0;
    unsigned int pixelIndex = launch_index.y * launch_dim.x + launch_index.x;
    PerRayData_pt prd;
    prd.numCastRays = 0;
    unsigned int rrSeed = hash ( frame + hash ( pixelIndex ) );
    optix::float3 throughput = make_float3 ( 1.0f );
    seed = tea<16> ( pixelIndex, numPreviousFrames );
    //      seed = r;
    uniformSeed = hash ( numPreviousFrames + 1 );

    float u1 = 0.f;
    float u2 = 0.f;

    if ( useLdSampling_image ) {
        u1 = Sobol ( ENUM_DIMENSION_IMAGE_U, numPreviousFrames, seed, 1 );
        u2 = Sobol ( ENUM_DIMENSION_IMAGE_V, numPreviousFrames, seed, 1 );

    } else {
        u1 = rnd ( seed );
        u2 = rnd ( seed );
    }

    float2 pixelSample = make_float2 ( launch_index ) + make_float2 ( u1, u2 );
    float2 d = pixelSample / make_float2 ( launch_dim ) * 2.f - 1.f;

    if ( useLdSampling_lens ) {
        u1 = Sobol ( ENUM_DIMENSION_LENS_U, numPreviousFrames, seed, 1 );
        u2 = Sobol ( ENUM_DIMENSION_LENS_V, numPreviousFrames, seed, 1 );

    } else {
        u1 = rnd ( seed );
        u2 = rnd ( seed );
    }

    if ( projectionType == ENUM_PROJECTION_TYPE_PERSPECTIVE ) {
        float3 theDir = d.x * U + d.y * V + W;
        float3 rayTarget = eye + theDir * ( focalDist );
        float2 sample = optix::square_to_disk ( make_float2 ( u1, u2 ) );
        float3 rayOrigin = eye + ( apertureRadius ) * ( sample.x * optix::normalize ( U ) + sample.y * optix::normalize ( V ) );
        theDir = optix::normalize ( rayTarget - rayOrigin );

        if ( numPreviousFrames == 0 ) { // for the first frame, shoot ray through pixel center
            pixelSample = make_float2 ( launch_index ) + make_float2 ( 0.5f, 0.5f );
            d = pixelSample / make_float2 ( launch_dim ) * 2.f - 1.f;
            theDir = optix::normalize ( d.x * U + d.y * V + W );
            rayOrigin = eye;
        }

        reInitializePrd ( prd, rayOrigin, theDir, seed, uniformSeed );

    } else if ( projectionType == ENUM_PROJECTION_TYPE_ORTHOGRAPHIC ) {
        float3 rayOrigin = eye + d.x * U + d.y * V;
        float3 theDir = optix::normalize ( W );
        reInitializePrd ( prd, rayOrigin, theDir, seed, uniformSeed );
    }
    
    float diffContrib = 0.f;


    do {

        if ( !prd.done ) {
            if ( prd.depth > 1 ) {
                prd.origin = getAdjustedRayStart ( prd.origin, prd.normal_geometric, scene_epsilon );
            }

            if ( prd.depth == 2 ) {
                rayDirectionResult = prd.direction;
            }

            optix::Ray ray = optix::make_Ray ( prd.origin, prd.direction, pt_ray_type, scene_epsilon, RT_DEFAULT_MAX );
            prd.dotProduct = 0.f;
            prd.pdf = 0.f;
            prd.f = make_float3 ( 0.f );
            prd.texCol = make_float3 ( 0.f );
            prd.radiance = make_float3 ( 0.f );
            rtTrace ( top_object, ray, prd );
            ++prd.numCastRays;

            
                
            prd.texCol *= prd.texCol; // more intense textures!
            directLightingResult += fmaxf ( make_float3 ( 0.f ), prd.radiance * prd.texCol * throughput );
            throughput *= prd.texCol * prd.f * prd.dotProduct / prd.pdf; // TODO throughput should come later
            
            if ( prd.depth == 0 ) {
                diffContrib = prd.kd;
                textureResult = prd.texCol;
                normalResult = ( optix::normalize ( prd.normal ) + make_float3 ( 1.f ) ) / 2.f;

                if ( prd.background ) {
                    backgroundPercentage = 1.f;
                    depthResult = -1.f;

                } else {
                    backgroundPercentage = 0.f;
                    depthResult = prd.hitDist;
                }

               

            } 
            // else {
            //     directLightingResult += fmaxf ( make_float3 ( 0.f ), prd.radiance * throughput );
            //     throughput *= prd.texCol * prd.f * prd.dotProduct / prd.pdf;
            // }

            
            
            float r = rnd ( rrSeed );
            float contProb = prd.depth > 1 ? fminf ( 1.f, optix::luminance ( throughput ) ) : 1;

            if ( r > contProb || prd.pdf <= 0.f || fmaxf ( throughput ) <= 0.f || fmaxf ( prd.f ) <= 0.f ) { // no contribution
                prd.done = true;

            } else {
                throughput /= contProb; // higher contribution
            }
        }

        ++prd.depth;
        // prd.done = true;
    } while ( prd.depth <= maxDepth && !prd.done );

    float oldWeight = ( float ) numPreviousFrames / ( float ) ( numPreviousFrames + 1 );
    float newWeight = 1.0f - oldWeight;

    if ( out_direct ) {
        //  correct errors if they happen, which they should not ;)
        directLightingResult = optix::fmaxf ( make_float3 ( 0.f ), directLightingResult );
        // indirectLightingResult = optix::fmaxf ( make_float3 ( 0.f ), indirectLightingResult );
        float4 theResult_direct = make_float4 ( 0.f );
        theResult_direct = oldWeight * output_buffer_directLight[launch_index] + newWeight * optix::make_float4 ( directLightingResult);
        theResult_direct.w = prd.numCastRays;

        if ( addRays ) {
            theResult_direct.w += output_buffer_directLight[launch_index].w;
        }

        output_buffer_directLight[launch_index] = theResult_direct;
    }

  //   if ( out_indirect ) {
  //       float4 theResult_indirect = make_float4 ( 0.f );
  //       theResult_indirect = oldWeight * output_buffer_indirectLight[launch_index] + newWeight * optix::make_float4 ( indirectLightingResult.x, indirectLightingResult.y, indirectLightingResult.z, diffContrib );
        // float4 theResult_indirect_squared = oldWeight * output_buffer_indirectLight_squared[launch_index] + newWeight * optix::make_float4 ( indirectLightingResult.x * indirectLightingResult.x, indirectLightingResult.y * indirectLightingResult.y, indirectLightingResult.z * indirectLightingResult.z, 1.f );
  //       output_buffer_indirectLight[launch_index] = theResult_indirect;
  //       output_buffer_indirectLight_squared[launch_index] = theResult_indirect_squared;
  //   }

    if ( out_raydir ) {
        float4 theResult_rayDirection = oldWeight * output_buffer_rayDirection[launch_index] + newWeight * optix::make_float4 ( rayDirectionResult.x, rayDirectionResult.y, rayDirectionResult.z, 0.f );
        float4 theResult_rayDirection_squared = oldWeight * output_buffer_rayDirection_squared[launch_index] + newWeight * optix::make_float4 ( rayDirectionResult.x * rayDirectionResult.x, rayDirectionResult.y * rayDirectionResult.y, rayDirectionResult.z * rayDirectionResult.z, 0.f );
        output_buffer_rayDirection[launch_index] = theResult_rayDirection;
        output_buffer_rayDirection_squared[launch_index] = theResult_rayDirection_squared;
    }

    if ( out_texture ) {
          output_buffer_texture[launch_index] = oldWeight * output_buffer_texture[launch_index] + newWeight * make_float4 ( 1.f, 1.f, 1.f, backgroundPercentage );
        //output_buffer_texture[launch_index] = oldWeight * output_buffer_texture[launch_index] + newWeight * make_float4 ( textureResult );
    }

    if ( out_gbuffer ) {
        output_buffer_depth_normal[launch_index] = oldWeight * output_buffer_depth_normal[launch_index] + newWeight * make_float4 ( depthResult, fmaxf ( make_float3 ( 0.f ), fminf ( make_float3 ( 1.f ), normalResult ) ) );
    }
}
