#ifndef _CAMERA_HPP_
#define _CAMERA_HPP_

#include <common.h>

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/vec2.hpp> // glm::vec3
#include <glm/vec3.hpp> // glm::vec3
#include <glm/vec4.hpp> // glm::vec4
#include <glm/mat4x4.hpp> // glm::mat4
#include <glm/gtc/matrix_transform.hpp> // glm::translate, glm::rotate, glm::scale,
#include <glm/gtx/rotate_vector.hpp>


class Camera
{

protected:
glm::vec3 __m_eye;
glm::vec3 __m_up;
glm::vec3 __m_look_at;
glm::vec3 __m_direction;

float __m_hfov;
float __m_aspect_ratio;   // 2.49 / 1.;

glm::vec3 __m_local_x;

bool __m_isInitialized;

glm::ivec2 __m_res;

bool __has_changed;

glm::mat4 __m_proj_matrix;
glm::mat4 __m_proj_matrix_inv;

glm::mat4 __model_view_matrix;
glm::mat4 __model_view_matrix_inv;


float far_;
float near_;

struct ScaleAndOffset2D {
        glm::vec2 Scale;
        glm::vec2 Offset;
};

public:

bool fwPressed;

Camera(int width, int height, float p_fov) {

        __m_isInitialized=false;



        //Sibenik image benchmarks
      /*
        glm::vec3 mPos = glm::vec3(0.2405, 0.01071, 0.15936);
        glm::vec3 mUp = glm::vec3(0.329571, 0.94337, 0.037885);
        glm::vec3 mDir = glm::vec3(0.941089, -0.331466, 0.0669513);
      */

        glm::vec3 mPos = glm::vec3(0.0, 0.f, 0.f);
        glm::vec3 mUp = glm::vec3(0.f, 1.f, 0.f);
        glm::vec3 mDir = glm::vec3(0.941089, -0.331466, 0.0669513);

        //Rungholt image benchmark
        initialize(mPos, mUp, mDir, p_fov, glm::ivec2(width, height), width / (float)height);

}


Camera(glm::vec3 p_eye, glm::vec3 p_up, glm::vec3 p_look_at, float p_hfov, glm::ivec2 p_res) {
        initialize(p_eye, p_up, p_look_at, p_hfov, p_res, p_res.x / (float) p_res.y);
}

void initialize(glm::vec3 p_eye, glm::vec3 p_up, glm::vec3 p_dir, float p_hfov, glm::ivec2 p_res, float screenRatio) {

        assert(!__m_isInitialized);

        __m_res = p_res;

        __m_eye = p_eye;
        __m_up = p_up;
        __m_look_at = p_eye + p_dir;
        __m_direction = p_dir;
        glm::normalize(__m_direction);

        __m_hfov = p_hfov;
        __m_aspect_ratio = screenRatio; //p_res.x / p_res.y; // 2.49 / 1.;

        __m_local_x = glm::cross(__m_direction, __m_up);

        near_ = 0.0001f;
        far_ = 100.f;

        // Generates a really hard-to-read matrix, but a normal, standard 4x4 matrix nonetheless
        __m_proj_matrix = glm::perspective<float>(__m_hfov,         // The horizontal Field of View, in degrees : the amount of "zoom". Think "camera lens". Usually between 90° (extra wide) and 30° (quite zoomed in)
                        screenRatio, // Aspect Ratio. Depends on the size of your window. Notice that 4/3 == 800/600 == 1280/960, sounds familiar ?
                        near_,        // Near clipping plane. Keep as big as possible, or you'll get precision issues.
                        far_       // Far clipping plane. Keep as little as possible.
                        );


        //__m_proj_matrix = glm::ortho(0.0f, 24.f, 0.f, 24.0f, near_, far_);

        //__m_proj_matrix = glm::transpose(__m_proj_matrix);

        __m_proj_matrix_inv = glm::inverse(__m_proj_matrix);

        __m_isInitialized = true;

        __has_changed = true;
        update();
}

void setEye(glm::vec3 p_eye) {

        __m_eye = p_eye;
        __m_direction = __m_look_at - p_eye;
        __m_direction = glm::normalize(__m_direction);
        __has_changed = true;
}

void setUp(glm::vec3 p_up) {

        __m_up = p_up;

        __m_local_x = glm::cross(__m_direction, __m_up);
        __has_changed = true;
}

void setDir(glm::vec3 p_dir) {

        __m_direction = p_dir;

        __m_local_x = glm::cross(__m_direction, __m_up);
        setLookAt(__m_eye+__m_direction);
        __has_changed = true;
}

void setLookAt(glm::vec3 p_look_at) {

        __m_look_at = p_look_at;
        __m_direction = p_look_at - __m_eye;
        __m_direction = glm::normalize(__m_direction);

        __m_local_x = glm::cross(__m_direction, __m_up);
        __has_changed = true;
}

void setHfov(float p_hfov) {
        __m_hfov = p_hfov;
        __has_changed = true;
}

glm::vec3 getEye()      {
        return __m_eye;
}
glm::vec3 getUp()       {
        return __m_up;
}
glm::vec3 getLookAt()     {
        return __m_look_at;
}
glm::vec3 getDirection()  {
        return __m_direction;
}
float getHfov()       {
        return __m_hfov;
}
float getAspectRatio()    {
        return __m_aspect_ratio;
}
float getNear()       {
        return near_;
}
float getFar()        {
        return far_;
}


void setCameraWithDir(glm::vec3 p_eye, glm::vec3 p_up, glm::vec3 p_direction) {
        setEye(p_eye);
        setUp(p_up);
        __m_up = glm::normalize(__m_up);
        __m_direction = p_direction;
        __m_direction = glm::normalize(__m_direction);
        __m_local_x = glm::cross(__m_direction, __m_up);
        __m_look_at = p_eye+p_direction;
        __has_changed = true;
}


void transform(float p_phi, float p_theta, float p_roll, glm::vec3 p_move, float p_time=1) {

        if (p_theta != 0.f || p_phi != 0.f || p_roll != 0.f) {


                p_roll *= p_time;
                p_phi *= p_time;
                p_theta *= p_time;

                glm::vec3 oldDir = __m_direction;

                glm::vec3 oldUp = __m_up;

                __m_direction = glm::rotate(__m_direction, p_roll,  oldDir);
                __m_direction = glm::rotate(__m_direction, p_theta, __m_local_x);
                __m_direction = glm::rotate(__m_direction, p_phi, oldUp);

                __m_up = glm::rotate(__m_up, p_roll,  oldDir);
                __m_up = glm::rotate(__m_up, p_theta, __m_local_x);
                __m_up = glm::rotate(__m_up, p_phi, oldUp);


                __m_direction = glm::normalize(__m_direction);
                __m_up = glm::normalize(__m_up);

                __m_local_x = glm::cross(__m_direction, __m_up);
        }

        // std::cout << p_time << std::endl;
        __m_eye = __m_eye + __m_direction * p_move.z * p_time + __m_local_x * p_move.x * p_time + __m_up * p_move.y * p_time;
        __m_look_at = __m_eye + __m_direction * p_time;
        __has_changed = true;
}

void update() {
        __model_view_matrix = glm::lookAt(__m_eye, __m_look_at, __m_up);

        __model_view_matrix_inv = glm::inverse(__model_view_matrix);
}


const glm::mat4& getModelViewMatrix() {
        return __model_view_matrix;
}


const glm::mat4& getModelViewMatrixInv() {
        return __model_view_matrix_inv;
}

const glm::mat4& getProjectionMatrix() {
        return __m_proj_matrix;
}

const glm::mat4& getProjectionMatrixInv() {
        return __m_proj_matrix_inv;
}

void moveFwBw(float amount) {
        transform(0, 0, 0, glm::vec3(0.f, 0.f, amount));
        __has_changed = true;
}

//positive left negative right
void moveLR(float amount) {
        //fwPressed=true;
        transform(0.f, 0.f, 0.f, glm::vec3(amount, 0.f, 0.f));
        __has_changed = true;
}

void moveUpDown(float amount) {
        transform(0.f, 0.f, 0.f, glm::vec3(0.f, amount, 0.f));
        __has_changed = true;
}

void yaw(float angle) {
        transform(0.f,-DEG2RAD(angle), 0.f, glm::vec3(0.f,0.f,0.f));
        __has_changed = true;
}

void pitch(float angle) {
        transform(-DEG2RAD(angle), 0.f, 0.f, glm::vec3(0.f,0.f,0.f));
        __has_changed = true;
}

void roll(float angle) {
        transform(0.f, 0.f, DEG2RAD(angle), glm::vec3(0.f,0.f,0.f));
        __has_changed = true;
}

void endFrame() {
        __has_changed = false;
}

const bool hasChanged() const {
        return __has_changed;
}

};


#endif
