#ifndef _GLFW_EVENT_HANDLER_H_
#define _GLFW_EVENT_HANDLER_H_

#include <utils/Fifar.h>

#include <common.h>
#include <camera.hpp>

#include <event/Eventdata.h>
#include <event/Tweakbar.h>

trd::Fifar fifar;

enum glfw_sixaxis_axes {
    ENUM_SIXAXIS_LEFT_ANALOG_HORIZONTAL = 0,
    ENUM_SIXAXIS_LEFT_ANALOG_VERTICAL = 1,

    ENUM_SIXAXIS_RIGHT_ANALOG_HORIZONTAL = 2,
    ENUM_SIXAXIS_RIGHT_ANALOG_VERTICAL = 3
};

enum glfw_sixaxis_axis_muliplicator {

    ENUM_SIXAXIS_LEFT_ANALOG_LEFT = -1,
    ENUM_SIXAXIS_LEFT_ANALOG_RIGHT = 1,

    ENUM_SIXAXIS_LEFT_ANALOG_UP = -1,
    ENUM_SIXAXIS_LEFT_ANALOG_DOWN = 1,

    ENUM_SIXAXIS_RIGHT_ANALOG_LEFT = -1,
    ENUM_SIXAXIS_RIGHT_ANALOG_RIGHT = 1,

    ENUM_SIXAXIS_RIGHT_ANALOG_UP = -1,
    ENUM_SIXAXIS_RIGHT_ANALOG_DOWN = 1
};

enum glfw_sixaxis_buttons {

    ENUM_SIXAXIS_SELECT = 0,
    ENUM_SIXAXIS_L3 = 1,
    ENUM_SIXAXIS_R3 = 2,
    ENUM_SIXAXIS_START = 3,
    ENUM_SIXAXIS_DPAD_UP = 4,
    ENUM_SIXAXIS_DPAD_RIGHT = 5,
    ENUM_SIXAXIS_DPAD_DOWN = 6,
    ENUM_SIXAXIS_DPAD_LEFT = 7,
    ENUM_SIXAXIS_L2 = 8,
    ENUM_SIXAXIS_R2 = 9,
    ENUM_SIXAXIS_L1 = 10,
    ENUM_SIXAXIS_R1 = 11,
    ENUM_SIXAXIS_TRIANGLE = 12,
    ENUM_SIXAXIS_CIRCLE = 13,
    ENUM_SIXAXIS_X = 14,
    ENUM_SIXAXIS_SQUARE = 15,
    ENUM_SIXAXIS_MENU = 16,

    ENUM_SIXAXIS_NUM_BUTTONS
};



class JoystickData
{
public:

    float mAnalogFB;
    float mAnalogLR;
    float mAnalogUD;
    float mAnalogRoll;
    float mAnalogYaw;
    float mAnalogPitch;

    bool buttonsActive[ENUM_SIXAXIS_NUM_BUTTONS];

    float sensitivity;

    JoystickData() : sensitivity(JOYSTICK_SENSITIVITY_MOVEMENT_INITIAL)
    {
        reset();
    }

    void reset()
    {
        mAnalogFB = 0.f;
        mAnalogLR = 0.f;
        mAnalogUD = 0.f;
        mAnalogRoll=0.f;
        mAnalogYaw = 0.f;
        mAnalogPitch=0.f;

        for(int i=0; i<ENUM_SIXAXIS_NUM_BUTTONS; i++) {
            buttonsActive[i] = false;
        }
    }

};


namespace eventhandler
{

//Mouse
float mCamMoveFrac;
float mouseSpeed = 10.f;

float mMouseXOrigin = -1;
float mMouseDeltaAngleX;
float mAngleX;

float mMouseYOrigin = -1;
float mMouseDeltaAngleY;
float mAngleY;

int mouseButtonPressed;

//Keyboard
float keyboard_sensitivity = KEYBOARD_SENSITIVITY_INITIAL;

//Joystick
JoystickData joy_data;
bool joystick_present=false;


//Global Event Components
Camera *event_cam=NULL;
glm::vec3 camTranslation_ = glm::vec3(0);
glm::vec3 camOrientation_ = glm::vec3(0);

EventData event_data;

namespace callbacks
{

void key(GLFWwindow* window, int key, int scancode, int action, int mods)
{

    if(!TwEventKeyGLFW(key, action)) {

        if (action == GLFW_PRESS || action==GLFW_REPEAT) {
            switch(key) {

            case GLFW_KEY_ESCAPE:
                //glfwDestroyWindow(window);
                glfwSetWindowShouldClose(window, true);
                break;

            case GLFW_KEY_KP_ADD: {
                keyboard_sensitivity+=KEYBOARD_SENSITIVITY_STEP_WIDTH;
                std::cout << "Keybord Sensitivity: " << keyboard_sensitivity << std::endl;
            }
            break;

            case GLFW_KEY_KP_SUBTRACT: {
                if(keyboard_sensitivity-KEYBOARD_SENSITIVITY_STEP_WIDTH<0)
                    keyboard_sensitivity=KEYBOARD_SENSITIVITY_STEP_WIDTH;
                else
                    keyboard_sensitivity-=KEYBOARD_SENSITIVITY_STEP_WIDTH;
                std::cout << "Keybord Sensitivity: " << keyboard_sensitivity << std::endl;
            }
            break;

            case GLFW_KEY_R: {
                camTranslation_ += glm::vec3(0.f, keyboard_sensitivity, 0.f);
            }
            break;
            case GLFW_KEY_F: {
                camTranslation_ -= glm::vec3(0.f, keyboard_sensitivity, 0.f);
            }
            break;

            case GLFW_KEY_E: {
                camOrientation_ += glm::vec3(0, 0, 0.5);
            }
            break;

            case GLFW_KEY_Q: {
                camOrientation_ -= glm::vec3(0, 0, 0.5);
            }
            break;

            case GLFW_KEY_Y: {
                camOrientation_ -= glm::vec3(0.5, 0, 0);
            }
            break;

            case GLFW_KEY_X: {
                camOrientation_ += glm::vec3(0.5, 0, 0);
            }
            break;

            case GLFW_KEY_W: {
                camTranslation_ += glm::vec3(0.f, 0.f, keyboard_sensitivity);
            }
            break;

            case GLFW_KEY_S: {
                camTranslation_ -= glm::vec3(0.f, 0.f, keyboard_sensitivity);
            }
            break;

            case GLFW_KEY_A: {
                camTranslation_ -= glm::vec3(keyboard_sensitivity, 0.f, 0.f);
            }
            break;

            case GLFW_KEY_D: {
                camTranslation_ += glm::vec3(keyboard_sensitivity, 0.f, 0.f);
            }
            break;

            case GLFW_KEY_B: {
                break;
            }

            case GLFW_KEY_C: {
                event_data.play = !event_data.play;
                break;
            }

            case GLFW_KEY_P: {
                std::cout << "Printing Camera Parameters" << std::endl;
                std::cout << "Eye: " << event_cam->getEye().x << ", " << event_cam->getEye().y << ", " << event_cam->getEye().z << std::endl;
                std::cout << "Up: " << event_cam->getUp().x << ", " << event_cam->getUp().y << ", " << event_cam->getUp().z << std::endl;
                std::cout << "Dir: " << event_cam->getDirection().x << ", " << event_cam->getDirection().y << ", " << event_cam->getDirection().z << std::endl;
            }
            break;

            case GLFW_KEY_F1: {
                /* if(!event_data.show_tweakbar) {
                     std::cout << "Show Tweak Bar on" << std::endl;
                	event_data.show_tweakbar = true;
                 } else {
                     std::cout << "Show Tweak Bar off" << std::endl;
                	event_data.show_tweakbar = false;
                 }
                 */
                Tweakbar::Instance().toggle();

            }
            break;

            case GLFW_KEY_F2: {
              // record camera pos in Fifar
              std::cout << "Adding keyframe..." << std::endl;
              std::cout << "Eye: " << event_cam->getEye().x << ", " << event_cam->getEye().y << ", " << event_cam->getEye().z << std::endl;
              std::cout << "Up: " << event_cam->getUp().x << ", " << event_cam->getUp().y << ", " << event_cam->getUp().z << std::endl;
              std::cout << "Dir: " << event_cam->getDirection().x << ", " << event_cam->getDirection().y << ", " << event_cam->getDirection().z << std::endl;

              trd::FrameDescriptor fd;
              fd.pos[0] = event_cam->getEye().x;
              fd.pos[1] = event_cam->getEye().y;
              fd.pos[2] = event_cam->getEye().z;

              fd.dir[0] = event_cam->getDirection().x;
              fd.dir[1] = event_cam->getDirection().y;
              fd.dir[2] = event_cam->getDirection().z;

              fd.up[0] = event_cam->getUp().x;
              fd.up[1] = event_cam->getUp().y;
              fd.up[2] = event_cam->getUp().z;

              fifar.addKeyFrame(fd);
            }
            break;

            case GLFW_KEY_F3: {
              std::cout << "Writing keyframes to file..." << std::endl;
              fifar.writeKeyFrameDescriptorsToFile("fifar_keyframes.txt");
            }
            break;

            case GLFW_KEY_F8: {
                //std::cout << "Saving image" << std::endl;
                //event_data.save_image = true;
            }


            case GLFW_KEY_F12: {
                event_data.windowModeChange=true;
                /*
                static int width, height;
                if(!event_data.fullscreen) {
                    std::cout << "Switching to fullscreen" << std::endl;

                	glfwGetWindowSize(window, &width, &height);

                    GLFWmonitor* monitor = glfwGetWindowMonitor(window);
                    const GLFWvidmode* mode = glfwGetVideoMode(monitor);
                	glfwSetWindowMonitor(window, monitor, 0, 0, mode->width, mode->height, mode->refreshRate);
                    event_data.fullscreen=true;
                } else {
                    glfwSetWindowMonitor(window, NULL, 0, 0, width, height, 0);
                    event_data.fullscreen=false;
                }*/

            }
            break;

            };

        }

    }

}

void cb_char(GLFWwindow* window, unsigned int key)
{
    // - Directly redirect GLFW char events to AntTweakBar
    if(!TwEventCharGLFW(key,1)) {

    }

}


void mouse_button(GLFWwindow* window, int button, int action, int mods)
{

    double xpos, ypos;
    glfwGetCursorPos(window, &xpos, &ypos);

    int width, height;
    glfwGetWindowSize(window, &width, &height);

    if(!TwEventMouseButtonGLFW(button, action)) {

        if(action == GLFW_PRESS) {
            switch(button) {
            case GLFW_MOUSE_BUTTON_LEFT:
                mouseButtonPressed=1;
                //LOG_INFOL("Left Button Pressed");
                mMouseXOrigin = (float) xpos;
                mMouseYOrigin = (float) ypos;
                break;

            case GLFW_MOUSE_BUTTON_MIDDLE:
                mMouseXOrigin = (float) xpos;
                mMouseYOrigin = (float) ypos;

                mouseButtonPressed=2;
                break;
            case GLFW_MOUSE_BUTTON_RIGHT:
                mMouseXOrigin = (float) xpos;
                mMouseYOrigin = (float) ypos;

                event_data.r_mouse_clicked = glm::vec2(mMouseXOrigin / (float) width, mMouseYOrigin / (float) height);

                mouseButtonPressed=3;
                break;

            }
        }

        if(action == GLFW_RELEASE) {
            switch (button) {
            case GLFW_MOUSE_BUTTON_LEFT:
                mouseButtonPressed=1;
                mAngleX += mMouseDeltaAngleX;
                mMouseXOrigin = -1;

                mAngleY -= mMouseDeltaAngleY;
                mMouseYOrigin = -1;

                //LOG_INFOL("Left Button Released");
                break;
            case GLFW_MOUSE_BUTTON_MIDDLE:
                mouseButtonPressed=2;

                mAngleX += mMouseDeltaAngleX;
                mMouseXOrigin = -1;

                mAngleY -= mMouseDeltaAngleY;
                mMouseYOrigin = -1;

                break;


            case GLFW_MOUSE_BUTTON_RIGHT:
                mouseButtonPressed=3;

                mAngleX += mMouseDeltaAngleX;
                mMouseXOrigin = -1;

                mAngleY -= mMouseDeltaAngleY;
                mMouseYOrigin = -1;

                break;
            }

        }

    }
}

void mouse_position(GLFWwindow* window, double xpos, double ypos)
{

    int width, height;
    glfwGetWindowSize(window, &width, &height);

    if(!TwEventMousePosGLFW(xpos, ypos)) {

        if (mMouseXOrigin >= 0 || mMouseYOrigin >= 0) {
            // update deltaAngle
            mMouseDeltaAngleX = (((xpos - mMouseXOrigin)/width)*mouseSpeed);
            mMouseDeltaAngleY = (((ypos - mMouseYOrigin)/height)*mouseSpeed);

            if(mouseButtonPressed==1) {	//Left mouse button
                camOrientation_.x += mMouseDeltaAngleX * MOUSE_SPEED;
                camOrientation_.y += mMouseDeltaAngleY * MOUSE_SPEED;
            }

            if(mouseButtonPressed==2) {

            }

            if(mouseButtonPressed==3) {	//Right mouse button

            }

        }//mouse origin moved

    }
}


void mouse_wheel(GLFWwindow* window, double xoffset, double yoffset)
{
    TwEventMouseWheelGLFW(xoffset);
}

void resize(GLFWwindow* window, int width, int height)
{
    // Send the new window size to AntTweakBar
    TwWindowSize(width, height);
}

void joystick(int joy, int event)
{
    if (event == GLFW_CONNECTED) {
        std::string joystick_name(glfwGetJoystickName(joy));
        std::cout << "Joystick <"<< joy << "> connected name: "<< joystick_name << std::endl;
    } else if (event == GLFW_DISCONNECTED) {
        // The joystick was disconnected
        std::cout << "Joystick <" << joy << "> disconnected" << std::endl;
    }
}


}


void setupEventHandler(GLFWwindow* window, Camera &cam, int width, int height)
{
    glfwSetKeyCallback(window, eventhandler::callbacks::key);
    glfwSetCharCallback(window, eventhandler::callbacks::cb_char);
    glfwSetMouseButtonCallback(window, eventhandler::callbacks::mouse_button);
    glfwSetScrollCallback(window, eventhandler::callbacks::mouse_wheel);
    glfwSetCursorPosCallback(window, eventhandler::callbacks::mouse_position);
    glfwSetWindowSizeCallback(window, eventhandler::callbacks::resize);

    glfwSetInputMode(window, GLFW_STICKY_KEYS, true);

    event_cam = &cam;
    Tweakbar::Instance().setup(width, height, &event_data);

}


void processEvents(double time_elapsed, float scene_size)
{

    if(joystick_present) {
        camTranslation_ -= glm::vec3(0.f, 0.f, joy_data.mAnalogFB);
        camTranslation_ -= glm::vec3(joy_data.mAnalogLR, 0.f, 0.f);
        camTranslation_ -= glm::vec3(0.f, joy_data.mAnalogUD, 0.f);

        if(fabs(joy_data.mAnalogYaw)>JOY_MOVE_EPS)
            camOrientation_.x += joy_data.mAnalogYaw;

        if(fabs(joy_data.mAnalogPitch)>JOY_MOVE_EPS)
            camOrientation_.y -= joy_data.mAnalogPitch;

        if(fabs(joy_data.mAnalogRoll)>JOY_MOVE_EPS)
            camOrientation_.z += joy_data.mAnalogRoll;
    }


    if(camTranslation_.x != 0 || camTranslation_.y != 0 || camTranslation_.z !=0 ||
       camOrientation_.x != 0 || camOrientation_.y != 0 || camOrientation_.z != 0) {
      event_cam->transform(-DEG2RADF(camOrientation_.x), -DEG2RADF(camOrientation_.y), DEG2RADF(camOrientation_.z), camTranslation_ * scene_size * 1.f/(float) SCENE_CAM_SPEED_SCALE, time_elapsed);
    }
//    event_cam->transform(-DEG2RADF(camOrientation_.x), -DEG2RADF(camOrientation_.y), DEG2RADF(camOrientation_.z), camTranslation_ * scene_size * 1.f/(float) SCENE_CAM_SPEED_SCALE, 1);
//    event_cam->transform(-DEG2RADF(camOrientation_.x), -DEG2RADF(camOrientation_.y), DEG2RADF(camOrientation_.z), glm::vec3(0), 1);

    camTranslation_ = glm::vec3(0);
    camOrientation_ = glm::vec3(0);
}


void handleJoystick()
{
    int present = glfwJoystickPresent(GLFW_JOYSTICK_1);
    if(present==0) return;


    joystick_present = true;

    joy_data.reset();

    int count_axes;
    const float* axes = glfwGetJoystickAxes(GLFW_JOYSTICK_1, &count_axes);

    int count_buttons;
    const unsigned char* buttons = glfwGetJoystickButtons(GLFW_JOYSTICK_1, &count_buttons);


    //Handle Joystick movement
    float absJoyPosS1X = (axes[ENUM_SIXAXIS_LEFT_ANALOG_HORIZONTAL]); // transform back to interval [0,100]
    float absJoyPosS1Y = (axes[ENUM_SIXAXIS_LEFT_ANALOG_VERTICAL])  ;   // transform back to interval [0,100]

    float absJoyPosS2X = (axes[ENUM_SIXAXIS_RIGHT_ANALOG_HORIZONTAL]); // transform back to interval [0,100]
    float absJoyPosS2Y = (axes[ENUM_SIXAXIS_RIGHT_ANALOG_VERTICAL])  ;   // transform back to interval [0,100]


    if(buttons[ENUM_SIXAXIS_L1] == GLFW_PRESS) {
        joy_data.mAnalogYaw = joy_data.sensitivity * axes[ENUM_SIXAXIS_RIGHT_ANALOG_HORIZONTAL];
        joy_data.mAnalogPitch = joy_data.sensitivity * axes[ENUM_SIXAXIS_RIGHT_ANALOG_VERTICAL];

        joy_data.mAnalogLR = joy_data.sensitivity * axes[ENUM_SIXAXIS_LEFT_ANALOG_HORIZONTAL];
        joy_data.mAnalogUD = -joy_data.sensitivity * axes[ENUM_SIXAXIS_LEFT_ANALOG_VERTICAL];
    } else if(buttons[ENUM_SIXAXIS_R1] == GLFW_PRESS) {
        joy_data.mAnalogRoll = -joy_data.sensitivity * axes[ENUM_SIXAXIS_RIGHT_ANALOG_HORIZONTAL];
        joy_data.mAnalogPitch = joy_data.sensitivity * axes[ENUM_SIXAXIS_RIGHT_ANALOG_VERTICAL];

        joy_data.mAnalogLR = -joy_data.sensitivity * axes[ENUM_SIXAXIS_LEFT_ANALOG_HORIZONTAL];
        joy_data.mAnalogFB = joy_data.sensitivity * axes[ENUM_SIXAXIS_LEFT_ANALOG_VERTICAL];
    }


    joy_data.mAnalogYaw = joy_data.sensitivity * axes[ENUM_SIXAXIS_RIGHT_ANALOG_HORIZONTAL];
    joy_data.mAnalogPitch = joy_data.sensitivity * axes[ENUM_SIXAXIS_RIGHT_ANALOG_VERTICAL];

    joy_data.mAnalogLR = -joy_data.sensitivity * axes[ENUM_SIXAXIS_LEFT_ANALOG_HORIZONTAL];
    joy_data.mAnalogFB = joy_data.sensitivity * axes[ENUM_SIXAXIS_LEFT_ANALOG_VERTICAL];

    //Handle button pressed
    for(int i=0; i<ENUM_SIXAXIS_NUM_BUTTONS; i++) {
        if(buttons[i] == GLFW_PRESS) {
            joy_data.buttonsActive[i] = true;
        }
    }

    if (buttons[ENUM_SIXAXIS_L2] == GLFW_PRESS) {
        joy_data.sensitivity-=0.05f;
        joy_data.sensitivity = joy_data.sensitivity <= 0.f ? 0.00001f : joy_data.sensitivity;
    }


    if (buttons[ENUM_SIXAXIS_R2] == GLFW_PRESS) {
        joy_data.sensitivity+=0.05f;
    }

    if (buttons[ENUM_SIXAXIS_DPAD_UP] == GLFW_PRESS) {

    }


    if (buttons[ENUM_SIXAXIS_DPAD_DOWN] == GLFW_PRESS) {

    }
    /*
        if (buttons[ENUM_SIXAXIS_X] == GLFW_PRESS) {

            // record camera pos in Fifar
            std::cout << "Adding keyframe..." << std::endl;
            std::cout << "Eye: " << event_cam->getEye().x << ", " << event_cam->getEye().y << ", " << event_cam->getEye().z << std::endl;
            std::cout << "Up: " << event_cam->getUp().x << ", " << event_cam->getUp().y << ", " << event_cam->getUp().z << std::endl;
            std::cout << "Dir: " << event_cam->getDirection().x << ", " << event_cam->getDirection().y << ", " << event_cam->getDirection().z << std::endl;

            trd::FrameDescriptor fd;
            fd.pos[0] = event_cam->getEye().x;
            fd.pos[1] = event_cam->getEye().y;
            fd.pos[2] = event_cam->getEye().z;

            fd.dir[0] = event_cam->getDirection().x;
            fd.dir[1] = event_cam->getDirection().y;
            fd.dir[2] = event_cam->getDirection().z;

            fd.up[0] = event_cam->getUp().x;
            fd.up[1] = event_cam->getUp().y;
            fd.up[2] = event_cam->getUp().z;

            __m_fifar.addKeyFrame(fd);

        }

        if (buttons[ENUM_SIXAXIS_START] == GLFW_PRESS) {

            std::cout << "Writing keyframes to file..." << std::endl;
            __m_fifar.writeKeyFrameDescriptorsToFile("fifar_keyframes.txt");

        }
    */


}


}


#endif
