#ifndef _TWEAK_BAR_H_
#define _TWEAK_BAR_H_


#include <AntTweakBar.h>
#include <utils/Singleton.h>

#include <event/Eventdata.h>

class Tweakbar : public Singleton<Tweakbar>
{


protected:

    TwBar *mainbar_;         // Pointer to a tweak bar
  bool visible_;

    EventData *eventdata_ = NULL;


public:

    Tweakbar() : mainbar_(0), visible_(false)
    {

        // Initialize AntTweakBar
        TwInit(TW_OPENGL, NULL);


    }

    virtual ~Tweakbar() {
      //if(mainbar_) free(mainbar_);
      //if(playbackbar_) free(playbackbar_);

    }


    void setWindowSize(int width, int height)
    {
        // Send the new window size to AntTweakBar
        TwWindowSize(width, height);

    }


    void setup(int width, int height, EventData *eventdata)
    {



        mainbar_ = TwNewBar("Main");

        eventdata_ = eventdata;

        setWindowSize(width, height);

        TwDefine(" GLOBAL help='Ray Mol VR allows to visualize molecules' "); // Message added to the help bar.

        TwAddVarRW(mainbar_, "dummy", TW_TYPE_BOOLCPP, &eventdata->dummy,
                   " label='Dummy' key=n help='Toggle dummy.' group=Dummies");
/*
        //Main Bar

        //Light source
        TwAddVarRW(mainbar_, "Shadows", TW_TYPE_BOOLCPP, &eventdata->with_shadows,
                   " label='Shadow mode' key=n help='Toggle shadows.' group=Light");
        TwAddVarRW(mainbar_, "Rotation", TW_TYPE_QUAT4F, eventdata->light_rotation,
                  " label='Rotation'  help='Light rotation.' group=Light");
        TwAddVarRW(mainbar_, "Color", TW_TYPE_COLOR3F, &eventdata->light_color,
                  " label='Color'  help='Light color.' group=Light");
        TwAddVarRW(mainbar_, "Intensity", TW_TYPE_FLOAT, &eventdata->light_intensity,
                  " label='Color'  help='Light intentsity.' group=Light");

        //AO
        TwAddVarRW(mainbar_, "AOToggle", TW_TYPE_BOOLCPP, &eventdata->with_ao,
                   " label='AO Enabled' key=m help='Toggle ambient occlusion.'  group=AO ");


        TwAddVarRW(mainbar_, "hbao_radius", TW_TYPE_FLOAT, &eventdata->hbaoctrl.r, " label='Radius' min=0 max=50 step=0.1 keyIncr=u keyDecr=U help='Radius of HBAO' group=AO");
        TwAddVarRW(mainbar_, "hbao_numdirs", TW_TYPE_FLOAT, &eventdata->hbaoctrl.numdirections, " label='Directions' min=0 max=16 step=1.0 keyIncr=i keyDecr=I help='Number of directions of HBAO' group=AO");
        TwAddVarRW(mainbar_, "hbao_numsteps", TW_TYPE_FLOAT, &eventdata->hbaoctrl.numsteps, " label='Steps' min=0 max=32 step=1.0 keyIncr=o keyDecr=O help='Number of steps per direction of HBAO' group=AO");
        TwAddVarRW(mainbar_, "hbao_bias", TW_TYPE_FLOAT, &eventdata->hbaoctrl.bias, " label='Bias' min=1 max=1 step=0.01 keyIncr=k keyDecr=K help='Number of steps per direction of HBAO' group=AO");
        TwAddVarRW(mainbar_, "hbao_intensity", TW_TYPE_FLOAT, &eventdata->hbaoctrl.intensity, " label='Intensity' min=1 max=10 step=1.00 keyIncr=l keyDecr=L help='Radius of AO screenspace' group=AO");
        //AO Blur
        TwAddVarRW(mainbar_, "AOBlurToggle", TW_TYPE_BOOLCPP, &eventdata->with_ao_blur,
                   " label='Blur Enabled' key=m help='Toggle Bilateral blur for AO.'  group=AO ");
        TwAddVarRW(mainbar_, "ao_blur_intensity", TW_TYPE_FLOAT, &eventdata->hbaoctrl.blur_sharpness, " label='Sharpness' min=0.01 max=1.0 step=0.01 keyIncr=j keyDecr=J help='Sharpness for AO Blur' group=AO");
        TwAddVarRW(mainbar_, "ao_blur_kernel_radius", TW_TYPE_FLOAT, &eventdata->hbaoctrl.blur_kernel_radius, " label='Radius' min=0.0 max=16.0 step=1.0 keyIncr=h keyDecr=H help='Kernel radius for the bilateral blur on the ao channel AO Blur' group=AO");

        // Add 'bgColor' to 'bar': it is a modifable variable of type TW_TYPE_COLOR3F (3 floats color)
        TwAddVarRW(mainbar_, "bgColor", TW_TYPE_COLOR3F, &eventdata_->bg_color, " label='Background color' ");
        TwAddVarRW(mainbar_, "phong_ka", TW_TYPE_FLOAT, &eventdata->phong.x, " label='kA' min=0 max=1 step=0.02 help='Ambient Phong' group=Phong");
        TwAddVarRW(mainbar_, "phong_kd", TW_TYPE_FLOAT, &eventdata->phong.y, " label='kD' min=0 max=1 step=0.02 help='Diffuse Phong' group=Phong");
        TwAddVarRW(mainbar_, "phong_ks", TW_TYPE_FLOAT, &eventdata->phong.z, " label='kS' min=0 max=1 step=0.02 help='Specular Phong' group=Phong");
        TwAddVarRW(mainbar_, "phong_n", TW_TYPE_FLOAT, &eventdata->phong.w, " label='n'  min=0 max=10000 step=1.0 help='N exp Phong' group=Phong");


        //DoF
        TwAddVarRW(mainbar_, "dof", TW_TYPE_BOOLCPP, &eventdata->with_dof,
                   " label='Enable' key=n help='Toggle DoF' group=DoF");
        TwAddVarRW(mainbar_, "dof_physical", TW_TYPE_UINT32, &eventdata->dof_config.mode,
                  "label='Physical Mode' key=n help='Toggle DoF Physical Mode' group=DoF min=0 max=2 step=1");

        TwAddVarRW(mainbar_, "dof_focus_plane", TW_TYPE_FLOAT, &eventdata->dof_config.focusPlane, " label='Focus Plane Z' min=0 max=1000 step=1 help='DoF Focus Plane' group=DoF");
        TwAddVarRW(mainbar_, "dof_scale", TW_TYPE_FLOAT, &eventdata->dof_config.scale, " label='Scale' min=0 max=1 step=0.0001 help='DoF Focus Scale' group=DoF");

        //Non-physical settings
        TwAddVarRW(mainbar_, "dof_art_near_sharp_plane", TW_TYPE_FLOAT, &eventdata->dof_config.nearSharpPlaneZ, " label='Near Focus Plane Z' min=0 max=1000 step=1 help='DoF near focus plane' group=DoF");
        TwAddVarRW(mainbar_, "dof_art_near_blurry_plane", TW_TYPE_FLOAT, &eventdata->dof_config.nearBlurryPlaneZ, " label='Near Blur Plane Z' min=0 max=1000 step=1 help='DoF near blurry plane' group=DoF");
        TwAddVarRW(mainbar_, "dof_art_near_scale", TW_TYPE_FLOAT, &eventdata->dof_config.nearScale, " label='Near Scale Z' min=0 max=1 step=0.0001 help='DoF near focus scale' group=DoF");

        TwAddVarRW(mainbar_, "dof_art_far_sharp_plane", TW_TYPE_FLOAT, &eventdata->dof_config.farSharpPlaneZ, " label='Far Focus Plane Z' min=0 max=1000 step=1 help='DoF far focus plane' group=DoF");
        TwAddVarRW(mainbar_, "dof_art_far_blurry_plane", TW_TYPE_FLOAT, &eventdata->dof_config.farBlurryPlaneZ, " label='Far Blur Plane Z' min=0 max=1000 step=1 help='DoF far focus plane' group=DoF");
        TwAddVarRW(mainbar_, "dof_art_far_scale", TW_TYPE_FLOAT, &eventdata->dof_config.farScale, " label='Far Scale Z' min=0 max=1 step=0.0001 help='DoF far focus scale' group=DoF");
        // int2 trimBandThickness;
        // float2 writeScaleBias;
        // float2 readScaleBias;
        TwAddVarRW(mainbar_, "dof_coverage_boost", TW_TYPE_FLOAT, &eventdata->dof_config.coverageBoost, " label='Near Coverage Boost' min=0 max=10 step=0.1 help='DoF near coverage boost' group=DoF");
        TwAddVarRW(mainbar_, "dof_far_radius_rescale", TW_TYPE_FLOAT, &eventdata->dof_config.farRadiusRescale, " label='Far Radius Rescale' min=0 max=10 step=0.1 help='DoF far radius rescale' group=DoF");

        TwAddVarRW(mainbar_, "dof_debug", TW_TYPE_UINT32,  &eventdata->dof_config.DoFDebugOption, " min=0 max=7 step=1  group=DoF");

        //Foveated AA
        TwAddVarRW(mainbar_, "foveated_aa", TW_TYPE_BOOLCPP, &eventdata->with_foveated_aa,
                   " label='Foveated AA' key=n help='Toggle Foveated AA' group=Foveated");
        TwAddVarRW(mainbar_, "nbr_of_aa_iterations", TW_TYPE_INT32, &eventdata->nbr_of_aa_iterations, " label='# Iterations' min=0 max=16 help='Number of iterations to perform AA' group=Foveated");


        TwAddVarRW(mainbar_, "gaze_point_x", TW_TYPE_FLOAT, &eventdata->gaze.gp.x, " label='Gaze X' min=0 max=1 step=0.01 help='Position of the gaze point X' group=Foveated");
        TwAddVarRW(mainbar_, "gaze_point_y", TW_TYPE_FLOAT, &eventdata->gaze.gp.y, " label='Gaze Y' min=0 max=1 step=0.01 help='Position of the gaze point Y' group=Foveated");

        //Playback Bar
        if(playback_enabled_) {
          TwAddButton(playbackbar_, "Play", play_cb, (void*) &eventdata_->play, NULL);
          TwAddButton(playbackbar_, "Stop", stop_cb, (void*) &eventdata_->play, NULL);

          TwAddVarRW(playbackbar_, "pb_speed", TW_TYPE_FLOAT, &eventdata->pb_speed, " label='Playback Speed' min=0 max=10 step=0.001 keyIncr=x keyDecr=X help='Playback Speed'");
          std::stringstream ss;
          ss << "label='Current frame' min=0 max=" << (eventdata->max_no_of_frames-1) << " step=1.0 help='Current Frame'";
          TwAddVarRW(playbackbar_, "pb_frame", TW_TYPE_FLOAT, &eventdata->current_frame_no, ss.str().c_str());
        }

        */
//TwAddVarRW(bar, "speed", TW_TYPE_DOUBLE, &speed,
//		           " label='Rot speed' min=0 max=2 step=0.01 keyIncr=s keyDecr=S help='Rotation speed (turns/second)' ");

        toggle();
    }

    void toggle()
    {

        if(visible_) {
            TwDefine(" Main visible=true");
            TwDefine(" TW_HELP visible=true");  // help bar is hidden
        } else {
            TwDefine(" Main visible=false");
            TwDefine(" TW_HELP visible=false ");  // help bar is hidden
        }

        visible_ = !visible_;
    }



    void draw()
    {
/*
        int ao_enabled = eventdata_->with_ao ? 1 : 0;
        TwSetParam(mainbar_, "hbao_radius", "visible", TW_PARAM_INT32, 1, &ao_enabled);
        TwSetParam(mainbar_, "hbao_numdirs", "visible", TW_PARAM_INT32, 1, &ao_enabled);
        TwSetParam(mainbar_, "hbao_numsteps", "visible", TW_PARAM_INT32, 1, &ao_enabled);
        TwSetParam(mainbar_, "hbao_bias", "visible", TW_PARAM_INT32, 1, &ao_enabled);
        TwSetParam(mainbar_, "hbao_intensity", "visible", TW_PARAM_INT32, 1, &ao_enabled);
        TwSetParam(mainbar_, "hbao_intensity", "visible", TW_PARAM_INT32, 1, &ao_enabled);
        TwSetParam(mainbar_, "AOBlurToggle", "visible", TW_PARAM_INT32, 1, &ao_enabled);
*/      // Draw tweak bars
        TwDraw();

    }


    void refresh() {
      TwRefreshBar(mainbar_);

    }


};

#endif
