#ifndef _GRID_BUILDER_H_
#define _GRID_BUILDER_H_

#include "utils/vec.h"
#include "utils/bbox.h"
#include "grid/grid.h"
#include "utils/prims.h"

void build_grid(std::vector<Cell> &cells, uint* sub_table, std::vector<gutil::Tri> tris, uint resolution, gutil::BBox uniform_bb) {

  uint globalCounter = 0;

  std::cout << "numTris: " << tris.size() << std::endl;

  std::cout << "Resolution: " << resolution << std::endl;
  std::cout << "uniform_bb.extents(): " << uniform_bb.extents().x << ", " << uniform_bb.extents().y << ", " << uniform_bb.extents().z << std::endl;

  std::cout << "####BUILDING GRID####"<< std::endl;

  // bool trisSorted = new bool[tris.size()]

  for(auto i=0; i < tris.size(); ++i) {
      if (i % 1000 == 0) {
        std::cout << i << " / " << tris.size() << std::endl;
      }

      const gutil::Tri* tri = &tris[i];
      auto v0 = tri->v0;
      // std::cout << "tri v0: " << v0.x << ", " << v0.y << ", " << v0.z << std::endl;

      auto v1 = v0 - tri->e1;
      auto v2 = v0 + tri->e2;

      // std::cout << "tri v1: " << v1.x << ", " << v1.y << ", " << v1.z << std::endl;
      // std::cout << "tri v2: " << v2.x << ", " << v2.y << ", " << v2.z << std::endl;


      uint sub = sub_table[i];

      if (MIN_BASE_LEVEL == MAX_BASE_LEVEL) {
        assert( sub == MAX_BASE_LEVEL);
      }

      uint cell_res = pow(2, sub);
      float cell_size = uniform_bb.extents().x / cell_res;

      // std::cout << "cell_size: " << cell_size << std::endl;

      gutil::BBox prims_bbox = tri->bbox();

      prims_bbox.min -= uniform_bb.min;
      prims_bbox.max -= uniform_bb.min;

      gutil::vec3 cell_idx_min = prims_bbox.min / gutil::vec3(cell_size);
      gutil::vec3 cell_idx_max = prims_bbox.max / gutil::vec3(cell_size);

      //std::cout << "cell_idx_min: " << cell_idx_min.x << ", " << cell_idx_min.y << "," << cell_idx_min.z <<std::endl;
      //std::cout << "cell_idx_max: " << cell_idx_max.x << ", " << cell_idx_max.y << "," << cell_idx_max.z <<std::endl;

      gutil::uivec3 start_cell((uint)floorf(cell_idx_min.x),
                               (uint)floorf(cell_idx_min.y),
                               (uint)floorf(cell_idx_min.z));

      gutil::uivec3 end_cell((uint)ceilf(cell_idx_max.x),
                                (uint)ceilf(cell_idx_max.y),
                                (uint)ceilf(cell_idx_max.z));

      end_cell.x = std::min(end_cell.x, cell_res - 1);
      end_cell.y = std::min(end_cell.y, cell_res - 1);
      end_cell.z = std::min(end_cell.z, cell_res - 1);
      // std::cout << "start_cell: " << start_cell.x << ", " << start_cell.y << ", " << start_cell.z << std::endl;
      // std::cout << "end_cell: " << end_cell.x << ", " << end_cell.y << ", " << end_cell.z << std::endl;
      // std::cout << "prims_bbox.min: " << prims_bbox.min.x << ", " << prims_bbox.min.y << ", " << prims_bbox.min.z << std::endl;

      for(auto x = start_cell.x; x <= end_cell.x; ++x) {
        for(auto y = start_cell.y; y <= end_cell.y; ++y) {
          for(auto z = start_cell.z; z <= end_cell.z; ++z) {
#define eps 0.0001f

            ++globalCounter;
            gutil::vec3 minBox(uniform_bb.min.x + x * cell_size - eps,
                               uniform_bb.min.y + y * cell_size - eps,
                               uniform_bb.min.z + z * cell_size - eps
                              );


            gutil::vec3 maxBox(uniform_bb.min.x + (x+1) * cell_size + eps,
                               uniform_bb.min.y + (y+1) * cell_size + eps,
                               uniform_bb.min.z + (z+1) * cell_size + eps
                              );



            gutil::BBox cell_box(minBox, maxBox);

       

            if(intersect_prim_cell(*tri, cell_box)) {
              cells.push_back(Cell(x,y,z,sub));

            }
      }

    }
  }
      // std::cout << std::endl;
      // std::cout << std::endl;

      /*
      if( cell_idx_max.x > 1000 || cell_idx_max.y > 1000 || cell_idx_max.z > 1000) {
        exit(1);
      }*/

  }


  std::cout << "globalCounter: " << globalCounter << std::endl;

}

#endif
