#ifndef _ASSIMP_VBO_SCENE_H_
#define _ASSIMP_VBO_SCENE_H_

#include <iostream>
#include <fstream>

// --------------------------CGAL----------------------------------------------
//#include <locale>
//#include <algorithm>
//#include <string>
//#include <CGAL/Simple_cartesian.h>
//#include <CGAL/Surface_mesh.h>

#include <sys/stat.h>
#include <cstring>
#include <cerrno>
#include <map>


#include <assimp/Importer.hpp>  //OO version Header!
#include <assimp/types.h>
#include <assimp/cimport.h>
#include <assimp/postprocess.h>
#include <assimp/scene.h>
#include <assimp/DefaultLogger.hpp>
#include <assimp/LogStream.hpp>
#include <assimp/material.h>
#include <assimp/color4.h>

#include <SFML/System/Err.hpp>
//#include <SFML/Graphics/stb_image/stb_image.h>

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtx/transform.hpp>

#include <utils/vec.h>
#include <utils/bbox.h>

#include <utils/prims.h>

/*
typedef CGAL::Simple_cartesian<double> K;
typedef CGAL::Surface_mesh<K::Point_3> tFaceNeighbours;
typedef tFaceNeighbours::Vertex_index vertex_descriptor;
typedef tFaceNeighbours::Face_index face_descriptor;
tFaceNeighbours::Property_map<face_descriptor, unsigned int> face_ids;
std::vector<vertex_descriptor> vhandle;
*/


/*
// -------------------- OpenMesh
#include <OpenMesh/Core/IO/MeshIO.hh>
#include <OpenMesh/Core/Mesh/PolyMesh_ArrayKernelT.hh>
#include <OpenMesh/Core/Mesh/TriMesh_ArrayKernelT.hh>

typedef OpenMesh::TriMesh_ArrayKernelT<>  tFaceNeighbours;
std::vector<tFaceNeighbours::VertexHandle> vhandle;
OpenMesh::FPropHandleT<unsigned int> face_id_prop_handle;
*/

//#include <Common.h>
//to map image filenames to textureIds
//#include <gutil/datatypes/Vec4f.h>
//#include <gutil/datatypes/AABBox.h>
//#include <gutil/geometry/TriangleMesh.h>
//#include <gutil/geometry/GLTriangle.h>

// #include "CImg.h"
// using namespace cimg_library;

#include <set>

using namespace Assimp;


// Information to render each assimp node
struct AssimpMesh {
        GLuint vao;
        GLuint texIndex;
        GLuint uniformBlockIndex;
        int numFaces;
};


// This is for a shader uniform block
struct AssimpMaterial {
        float diffuse[4];
        float ambient[4];
        float specular[4];
        float emissive[4];
        float shininess;
        int texCount;
};



class AssimpVBOScene
{

public:
// Vertex Attribute Locations
GLuint vertexLoc;
GLuint normalLoc;
GLuint texCoordLoc;
GLuint primIDLoc;

// Uniform Bindings Points
GLuint matricesUniLoc;
GLuint materialUniLoc;
std::vector<struct AssimpMesh> myMeshes;

uint64_t mNbrOfFaces;
uint64_t mNbrOfVertices;

typedef std::vector<std::set<size_t>> tFaceNeighbours;
typedef std::vector<std::vector<size_t>> tVertexPrimMap;
tVertexPrimMap vertex_prim_rel_;
tFaceNeighbours prim_neigh_rel_;

typedef std::vector<gutil::Tri> tRawTriangles;
tRawTriangles triangles_;

protected:


// For push and pop matrix
std::vector<float *> matrixStack;

const aiScene* mScene;

// images / texture
std::map<std::string, GLuint*> mTextureIdMap;     // map image filenames to textureIds
GLuint*   mTextureIds;                  // pointer to texture Array


GLuint mSceneList;
GLuint mWhiteTexID;

// Create an instance of the Importer class
Assimp::Importer mImporter;

gutil::BBox mSceneBBox;
gutil::BBox mUniformBBox;
float mLargestExtend;
float mRecipLargestExtend;





// -----
void computeBoundingBoxRecursive (const aiNode* nd,
                            gutil::BBox &aBB, aiMatrix4x4* trafo) {

        aiMatrix4x4 prev;
        unsigned int n = 0, t;
        prev = *trafo;
        aiMatrix4x4 trp = nd->mTransformation;

        //trp.Scaling(aiVector3D(1, 1, 1), trp);
        //trp.Transpose();

        aiMultiplyMatrix4(trafo,&trp);

        unsigned int mesh_vertex_total = 0;
        for (; n < nd->mNumMeshes; ++n) {
                const aiMesh* mesh = mScene->mMeshes[nd->mMeshes[n]];

                for (t = 0; t < mesh->mNumVertices; ++t) {
                        aiVector3D tmp = mesh->mVertices[t];
                        aiTransformVecByMatrix4(&tmp,trafo);

                      //  vhandle.push_back(omesh.add_vertex(tFaceNeighbours::Point(tmp.x, tmp.y, tmp.z)));
                      //  vhandle.push_back(omesh.add_vertex(K::Point_3(tmp.x, tmp.y, tmp.z)));
                        aBB.extend(gutil::vec3(tmp.x, tmp.y, tmp.z));

                        mesh_vertex_total++;
                }
        }

        //vertex_total+=mesh_vertex_total;
        for (n = 0; n < nd->mNumChildren; ++n) {
                computeBoundingBoxRecursive(nd->mChildren[n],aBB,trafo);
        }
        *trafo = prev;
}



// -----
void buildNeighRelRecursive (const aiNode* nd, aiMatrix4x4* trafo, uint64_t &prim_id_total) {
        aiMatrix4x4 prev;
        unsigned int n = 0, t;
        prev = *trafo;
        aiMatrix4x4 trp = nd->mTransformation;

        //trp.Scaling(aiVector3D(1, 1, 1), trp);
        //trp.Transpose();

        aiMultiplyMatrix4(trafo,&trp);

        unsigned int mesh_vertex_total = 0;
        for (; n < nd->mNumMeshes; ++n) {
                const aiMesh* mesh = mScene->mMeshes[nd->mMeshes[n]];

                for(t = 0; t < mesh->mNumFaces; ++t) {
                        const struct aiFace* face = &mesh->mFaces[t];
                        if(face->mNumIndices != 3) continue;

                        prim_neigh_rel_[prim_id_total].insert(std::begin(vertex_prim_rel_[face->mIndices[0]]), std::end(vertex_prim_rel_[face->mIndices[0]]));
                        prim_neigh_rel_[prim_id_total].insert(std::begin(vertex_prim_rel_[face->mIndices[1]]), std::end(vertex_prim_rel_[face->mIndices[1]]));
                        prim_neigh_rel_[prim_id_total].insert(std::begin(vertex_prim_rel_[face->mIndices[2]]), std::end(vertex_prim_rel_[face->mIndices[2]]));
                        prim_id_total++;
                }
        }

        //vertex_total+=mesh_vertex_total;
        for (n = 0; n < nd->mNumChildren; ++n) {
                buildNeighRelRecursive(nd->mChildren[n],trafo,prim_id_total);
        }
        *trafo = prev;
}

// -----
void parseSceneDataRecursive (const aiNode* nd, aiMatrix4x4* trafo,
                            uint64_t &vertex_total, uint64_t &prim_id_total) {
        aiMatrix4x4 prev;
        unsigned int n = 0, t;
        prev = *trafo;
        aiMatrix4x4 trp = nd->mTransformation;

        //  trp.Scaling(aiVector3D(1, 1, 1), trp);

        //      trp.Transpose();
        aiMultiplyMatrix4(trafo,&trp);

        unsigned int mesh_vertex_total = 0;
        for (; n < nd->mNumMeshes; ++n) {
                const aiMesh* mesh = mScene->mMeshes[nd->mMeshes[n]];

                // for (t = 0; t < mesh->mNumVertices; ++t) {
                //         aiVector3D tmp = mesh->mVertices[t];
                //         aiTransformVecByMatrix4(&tmp,trafo);

                //       //  vhandle.push_back(omesh.add_vertex(tFaceNeighbours::Point(tmp.x, tmp.y, tmp.z)));
                //       //  vhandle.push_back(omesh.add_vertex(K::Point_3(tmp.x, tmp.y, tmp.z)));
                //         aBB.extend(gutil::vec3(tmp.x, tmp.y, tmp.z));

                //         mesh_vertex_total++;
                // }

                for(t = 0; t < mesh->mNumFaces; ++t) {
                        const struct aiFace* face = &mesh->mFaces[t];
                        if(face->mNumIndices != 3) continue;

                        const aiVector3D* av0 = &mesh->mVertices[face->mIndices[0]];
                        const aiVector3D* av1 = &mesh->mVertices[face->mIndices[1]];
                        const aiVector3D* av2 = &mesh->mVertices[face->mIndices[2]];


                        gutil::vec3 extents = mSceneBBox.extents();
                        float maxSize = fmaxf(extents.x, fmaxf(extents.y, extents.z));
                        float scaleFactor = 1.f / maxSize * 0.9999f;
                        gutil::vec3 translation = gutil::vec3(- mSceneBBox.min.x, - mSceneBBox.min.y, - mSceneBBox.min.z);

                        gutil::vec3 av0_tfd(
                            (av0->x + translation.x) * scaleFactor,
                            (av0->y + translation.y) * scaleFactor,
                            (av0->z + translation.z) * scaleFactor
                        );
                        gutil::vec3 av1_tfd(
                            (av1->x + translation.x) * scaleFactor,
                            (av1->y + translation.y) * scaleFactor,
                            (av1->z + translation.z) * scaleFactor
                        );
                        gutil::vec3 av2_tfd(
                            (av2->x + translation.x) * scaleFactor,
                            (av2->y + translation.y) * scaleFactor,
                            (av2->z + translation.z) * scaleFactor
                        );





                        vertex_prim_rel_[face->mIndices[0]].push_back(prim_id_total);
                        vertex_prim_rel_[face->mIndices[1]].push_back(prim_id_total);
                        vertex_prim_rel_[face->mIndices[2]].push_back(prim_id_total);
                        //
                        //tFaceNeighbours::Face_index fidx = omesh.add_face(vhandle[vertex_total + face->mIndices[0]], vhandle[vertex_total + face->mIndices[2]], vhandle[vertex_total + face->mIndices[1]]);
                        //if(fidx == tFaceNeighbours::null_face()) {
                        //  std::cout << "Null Face returned: " << std::endl;
                        //  continue;
                        //}
                        //face_ids[fidx] = prim_id_total++;

                        gutil::vec3 v0(av0_tfd.x,av0_tfd.y,av0_tfd.z);
                        gutil::vec3 v1(av1_tfd.x,av1_tfd.y,av1_tfd.z);
                        gutil::vec3 v2(av2_tfd.x,av2_tfd.y,av2_tfd.z);

                        gutil::vec3 e1 = v0 - v1;
                        gutil::vec3 e2 = v2 - v0;

                        gutil::vec3 normal = gutil::cross(e1, e2);
                        normal = gutil::normalize(normal);

                        gutil::Tri tri(v0, normal.x,
                                       e1, normal.y,
                                       e2, normal.z);

                        triangles_.push_back(tri);
                        prim_id_total++;

                        //std::cout << "fidx: " << fidx << std::endl;
/*
                        std::vector<tFaceNeighbours::VertexHandle>  face_vhandles;
                        //face_vhandles.clear();
                        //face_vhandles.push_back(vhandle[vertex_total + face->mIndices[0]]);
                        //face_vhandles.push_back(vhandle[vertex_total + face->mIndices[1]]);
                        //face_vhandles.push_back(vhandle[vertex_total + face->mIndices[2]]);
                        //tFaceNeighbours::FaceHandle fhdl = omesh.add_face(face_vhandles);
                        tFaceNeighbours::FaceHandle fhdl = omesh.add_face(vhandle[vertex_total + face->mIndices[0]], vhandle[vertex_total + face->mIndices[2]], vhandle[vertex_total + face->mIndices[1]]);

                        if(!fhdl.is_valid()) {
                          std::cout << "This is what I thought" << std::endl;
                          exit(1);
                        } else {
                          std::cout << "I am good" << std::endl;
                          omesh.property(face_id_prop_handle, fhdl) = prim_id_total++;
                        }
*/
                        //
                }
        }

        vertex_total+=mesh_vertex_total;
        for (n = 0; n < nd->mNumChildren; ++n) {
                parseSceneDataRecursive(nd->mChildren[n],trafo, vertex_total, prim_id_total);
        }
        *trafo = prev;
}

// ----------------------------------------------------------------------------

void parseSceneData() {



        uint64_t vertex_total=0;
        uint64_t prim_id_total=0;

        //Register Custom property - OpenMesh
        //omesh.add_property(face_id_prop_handle);

        //bool created;
        //boost::tie(face_ids, created) = omesh.add_property_map<face_descriptor,unsigned int>("v:name",0);
        //assert(created);
        prim_neigh_rel_.resize(mNbrOfFaces);
        vertex_prim_rel_.resize(mNbrOfVertices);
        aiMatrix4x4 trafo;
        aiIdentityMatrix4(&trafo);
        parseSceneDataRecursive(mScene->mRootNode,&trafo, vertex_total, prim_id_total);

        aiIdentityMatrix4(&trafo);
        uint64_t prim_id_total_n = 0;
        buildNeighRelRecursive(mScene->mRootNode,&trafo, prim_id_total_n);

        assert(prim_id_total_n == prim_id_total);
        //std::cout << "SCENES BBox: MIN " << mSceneBBox.min() << std::endl;
        //std::cout << "SCENES BBox: MAX " << mSceneBBox.getMax() << std::endl;
        //std::cout << "SCENES BBox: extend " << mSceneBBox.getExtend() << std::endl;

        mLargestExtend = gutil::max(mSceneBBox.extents().x, gutil::max(mSceneBBox.extents().y, mSceneBBox.extents().z));
        mUniformBBox = gutil::BBox(gutil::vec3(0.f), gutil::vec3(mLargestExtend));
        mRecipLargestExtend = 1.f/mLargestExtend;



        //mUniformBBoxOrg = gutil::AABBox(gutil::Vec4f(0), mSceneBBox.getExtend()*mRecipLargestExtend);

}


GLint createWhiteTexture(uint pWidth, uint pHeigth) {
        unsigned char* data;

        // Allocate the needed space.
        data = new unsigned char[pWidth * pHeigth];
        memset(data,1, pWidth * pHeigth * sizeof(unsigned char));

        // Generate white OpenGL texture.
        GLuint whiteTextureID;
        glGenTextures(1, &whiteTextureID);
        glBindTexture(GL_TEXTURE_2D, whiteTextureID);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, pWidth, pHeigth, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

        return whiteTextureID;
}


public:


/*
AssimpVBOScene(std::string aPath, std::vector<gutil::Tri> &prims) : mScene(NULL), mSceneList(0), mNbrOfFaces(0), mWhiteTexID(0), vertexLoc(0), normalLoc(1), texCoordLoc(2), primIDLoc(3), matricesUniLoc(1), materialUniLoc(2) {
        mWhiteTexID = createWhiteTexture(4,4);
        load(aPath, prims);
}
*/

AssimpVBOScene() : mScene(NULL), mSceneList(0), mNbrOfFaces(0), mNbrOfVertices(0), mWhiteTexID(0), vertexLoc(0), normalLoc(1), texCoordLoc(2), primIDLoc(3), matricesUniLoc(1), materialUniLoc(2) {
        mWhiteTexID = createWhiteTexture(4,4);
}


virtual ~AssimpVBOScene() {
        // cleaning up
        mTextureIdMap.clear();

        // clear myMeshes stuff
        for (unsigned int i = 0; i < myMeshes.size(); ++i) {

                glDeleteVertexArrays(1,&(myMeshes[i].vao));
                glDeleteTextures(1,&(myMeshes[i].texIndex));
                glDeleteBuffers(1,&(myMeshes[i].uniformBlockIndex));
        }
}

void load(const std::string aPath) {

        std::cout <<  "Loading Scene" << std::endl;
        mScene = mImporter.ReadFile( aPath, aiProcessPreset_TargetRealtime_Quality | aiProcess_PreTransformVertices | aiProcess_Triangulate | aiProcess_JoinIdenticalVertices | aiProcess_SortByPType | aiProcess_SplitLargeMeshes);

        // If the import failed, report it
        if( !mScene) {
                std::cerr << mImporter.GetErrorString() << std::endl;
                exit(1);
        }

        //Extract basepath
        size_t lastPos = aPath.rfind("/");
        // LOG_OUTL("LastPos: " << lastPos);


        std::string basepath = aPath.substr(0,lastPos) + "/";
        // LOG_OUTL("BasePath: " << basepath);

        // Now we can access the file's contents.
        std::cout << "Import of scene " + aPath + " succeeded." << std::endl;
        // std::cout << "Loading Textures" << std::endl;
        // LoadGLTextures(basepath);


        mSceneBBox = gutil::BBox::empty();
        aiMatrix4x4 trafo;
        aiIdentityMatrix4(&trafo);
        computeBoundingBoxRecursive(mScene->mRootNode,mSceneBBox,&trafo);

        readAiScene();
        parseSceneData();


        mRecipLargestExtend = 1.f/mLargestExtend;
        mLargestExtend = gutil::max(mSceneBBox.extents().x, gutil::max(mSceneBBox.extents().y, mSceneBBox.extents().z));
        mUniformBBox = gutil::BBox(mSceneBBox.center() - gutil::vec3(mLargestExtend) * 0.5f,  mSceneBBox.center() + gutil::vec3(mLargestExtend) * 0.5f);



}


gutil::BBox getUnitUniformBBox() {
    return gutil::BBox(gutil::vec3(0, 0, 0), gutil::vec3(1, 1, 1));
}

tRawTriangles& getRawTriangles() {
  return triangles_;
}

tVertexPrimMap& getVertexPrimRel() {
  return vertex_prim_rel_;
}

tFaceNeighbours& getPrimNeighRel() {
  return prim_neigh_rel_;
}

const aiScene* getScene() {
        return mScene;
}

protected:


int LoadGLTextures(const std::string &aBasePath) {

        if (mScene->HasTextures()) std::cerr << "Support for meshes with embedded textures is not implemented" << std::endl;

        /* getTexture Filenames and Numb of Textures */
        for (unsigned int m=0; m<mScene->mNumMaterials; m++) {
                int texIndex = 0;
                aiReturn texFound = AI_SUCCESS;

                aiString path; // filename



                while (texFound == AI_SUCCESS) {
                        int numTexPerIndex=0;

                        if(mScene->mMaterials[m]->GetTextureCount(aiTextureType_DIFFUSE)>0) {
                                std::cout << "Diffuse Texture found" << std::endl;
                                texFound = mScene->mMaterials[m]->GetTexture(aiTextureType_DIFFUSE, texIndex, &path);
                                std::cout << "Path: " << path.data << std::endl;
                                mTextureIdMap[path.data] = NULL; //fill map with textures, pointers still NULL yet
                                numTexPerIndex++;
                        }
                        if(mScene->mMaterials[m]->GetTextureCount(aiTextureType_OPACITY )>0) {
                                aiString opath; // filename
                                std::cout << "Opacity Texture found" << std::endl;
                                texFound = mScene->mMaterials[m]->GetTexture(aiTextureType_OPACITY, texIndex, &opath);
                                std::cout << "OPath: " << opath.data << std::endl;
                                mTextureIdMap[opath.data] = NULL; //fill map with textures, pointers still NULL yet
                                numTexPerIndex++;
                        }

                        if(numTexPerIndex==0) {
                                texFound = AI_FAILURE;
                        }
                        texIndex++;
                }
        }

        int numTextures = mTextureIdMap.size();


        /* create and fill array with GL texture ids */
        mTextureIds = new GLuint[numTextures];
        glGenTextures(numTextures, mTextureIds); /* Texture name generation */

        /* define texture path */
        //std::string texturepath = "../../../test/models/Obj/";

        /* get iterator */
        std::map<std::string, GLuint*>::iterator itr = mTextureIdMap.begin();

        for (int i=0; i<numTextures; i++) {

                //save image ID
                std::string filename = (*itr).first; // get filename
                (*itr).second =  &mTextureIds[i]; // save texture id for filename in map
                itr++;              // next texture

                std::cout << "Loading Texture: " << filename << std::endl;


                std::string fileloc = aBasePath + filename; //
                //std::string fileloc = "C:\\Users\\mweier2s\\Documents\\Visual Studio 2010\\Projects\\TreeOpterix\\trunk\\Ressources\\trees2\\maps\\EU13brk.png";
                //std::string fileloc = "C:\\out\\img91.jpg";

                bool success = false;


                // if(filename.length()>0) {

                //         //image.display();
                //         /*
                //            FREE_IMAGE_FORMAT formato = FreeImage_GetFileType(fileloc.c_str(),0);//Automatocally detects the format(from over 20 formats!)
                //            FIBITMAP* imagen = FreeImage_Load(formato, fileloc.c_str());

                //            FIBITMAP* temp = imagen;
                //            imagen = FreeImage_ConvertTo32Bits(imagen);
                //            FreeImage_Unload(temp);

                //            int w = FreeImage_GetWidth(imagen);
                //            int h = FreeImage_GetHeight(imagen);
                //            //cout<<"The size of the image is: "<<textureFile<<" es "<<w<<"*"<<h<<endl; //Some debugging code



                //            char* pixels = (char*)FreeImage_GetBits(imagen);
                //          */

                //         CImg<unsigned char> image;
                //         image.load(fileloc.c_str());

                //         uint w =  image.width();
                //         uint h =  image.height();
                //         GLubyte* texture = new GLubyte[4*w*h];

                //         cimg_forXY(image,x,y) {
                //                 texture[(y*w * 4) + x * 4 + 0]= *image.data(x,y, 0, 0);
                //                 texture[(y*w * 4) + x * 4 + 1]= *image.data(x,y, 0, 1);
                //                 texture[(y*w * 4) + x * 4 + 2]= *image.data(x,y, 0, 2);
                //                 texture[(y*w * 4) + x * 4 + 3]= *image.data(x,y, 0, 3);

                //                 //texture[(x*h * 4) + y * 4 + 0]= *image.data(x,y, 0, 0);
                //                 //texture[(x*h * 4) + y * 4 + 1]= *image.data(x,y, 0, 1);
                //                 //texture[(x*h * 4) + y * 4 + 2]= *image.data(x,y, 0, 2);
                //                 //texture[(x*h * 4) + y * 4 + 3]= *image.data(x,y, 0, 3);
                //         }

                //         glBindTexture(GL_TEXTURE_2D,  mTextureIds[i]);
                //         glTexImage2D(GL_TEXTURE_2D,  0,  GL_RGBA, w, h,  0,  GL_RGBA,  GL_UNSIGNED_BYTE, (GLvoid*) texture);
                //         //gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, image.width(), image.height(), GL_RGBA, GL_UNSIGNED_BYTE, image.data());

                //         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
                //         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

                //         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
                //                         GL_LINEAR);
                //         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                //                         GL_LINEAR_MIPMAP_LINEAR);
                //         glGenerateMipmap(GL_TEXTURE_2D);

                //         free(texture);


                // } else {
                //         // Error occured
                //         std::cerr << "Couldn't load image: " << fileloc << std::endl;
                //         //return false;
                // }


        }

        //return success;
        return true;
}



void readAiScene() {
        struct AssimpMesh aMesh;
        struct AssimpMaterial aMat;
        GLuint buffer;

        static unsigned int primID=0;


        // For each mesh
        for (unsigned int n = 0; n < mScene->mNumMeshes; ++n) {
                const aiMesh* mesh = mScene->mMeshes[n];

                // create array with faces
                // have to convert from Assimp format to array
                unsigned int *faceArray;
                faceArray = (unsigned int *)malloc(sizeof(unsigned int) * mesh->mNumFaces * 3);

                bool validMesh = true;
                for (unsigned int t = 0; t < mesh->mNumFaces; ++t) {
                    const aiFace *face = &mesh->mFaces[t];
                    if (face->mNumIndices != 3) {
                        validMesh = false;
                        break;
                    }
                }
                if (!validMesh) continue;

                unsigned int faceIndex = 0;

               for (unsigned int t = 0; t < mesh->mNumFaces; ++t) {
                        const aiFace* face = &mesh->mFaces[t];

                        if(face->mNumIndices!=3) {
                                std::cout << "Wrong face mode: " << face->mNumIndices << std::endl;
                                std::cout << "Only triangle meshes are supported" << std::endl;
                                exit(1);
                        }

                        memcpy(&faceArray[faceIndex], face->mIndices,3 * sizeof(unsigned int));
                        faceIndex += 3;
                        mNbrOfFaces++;
                }
                // mNbrOfFaces += faceIndex;



                aMesh.numFaces = mScene->mMeshes[n]->mNumFaces;

                // generate Vertex Array for mesh
                glGenVertexArrays(1,&(aMesh.vao));
                glBindVertexArray(aMesh.vao);
                GL_CHECK_ERROR();
                // buffer for faces
                glGenBuffers(1, &buffer);
                glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer);
                glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned int) * mesh->mNumFaces * 3, faceArray, GL_STATIC_DRAW);
                GL_CHECK_ERROR();

                gutil::vec3 extents = mSceneBBox.extents();
                float maxSize = fmaxf(extents.x, fmaxf(extents.y, extents.z));
                float scaleFactor = 1.f / maxSize;
                gutil::vec3 translation = gutil::vec3(- mSceneBBox.min.x, - mSceneBBox.min.y, - mSceneBBox.min.z);

                float *transformedVertices = new float[3 * 3 * mesh->mNumVertices];


                for (int i = 0; i < mesh->mNumVertices; ++i) {
                    transformedVertices[i * 3 + 0] = (mesh->mVertices[i].x + translation.x) * scaleFactor;
                    transformedVertices[i * 3 + 1] = (mesh->mVertices[i].y + translation.y) * scaleFactor;
                    transformedVertices[i * 3 + 2] = (mesh->mVertices[i].z + translation.z) * scaleFactor;
                    //std::cout << "sceneBBox: " << mSceneBBox.min.x << ", " << mSceneBBox.min.y << ", " << mSceneBBox.min.z << std::endl;
                    //std::cout << "ScaleFactor: " << scaleFactor << std::endl;
                    //std::cout << "translation: " << translation.x << ", " << translation.y << ", " << translation.z << std::endl;
                    //std::cout << "Old vertex: " << mesh->mVertices[i].x << ", " << mesh->mVertices[i].y << ", " << mesh->mVertices[i].z << std::endl;
                    //std::cout << "Vertex: " << transformedVertices[i * 3 + 0] << ", " << transformedVertices[i * 3 + 1] << ", " << transformedVertices[i * 3 + 2] << std::endl;
                    //exit(1);
                }
                // buffer for vertex positions
                if (mesh->HasPositions()) {
                        glGenBuffers(1, &buffer);
                        glBindBuffer(GL_ARRAY_BUFFER, buffer);

                        // glBufferData(GL_ARRAY_BUFFER, sizeof(float)*3*mesh->mNumVertices, mesh->mVertices, GL_STATIC_DRAW);
                        glBufferData(GL_ARRAY_BUFFER, sizeof(float)*3*mesh->mNumVertices, transformedVertices, GL_STATIC_DRAW);
                        mNbrOfVertices += mesh->mNumVertices;
                        glEnableVertexAttribArray(vertexLoc);
                        glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, 0, 0, 0);
                }
                GL_CHECK_ERROR();
                // buffer for vertex normals
                if (mesh->HasNormals()) {
                        glGenBuffers(1, &buffer);
                        glBindBuffer(GL_ARRAY_BUFFER, buffer);
                        glBufferData(GL_ARRAY_BUFFER, sizeof(float)*3*mesh->mNumVertices, mesh->mNormals, GL_STATIC_DRAW);
                        glEnableVertexAttribArray(normalLoc);
                        glVertexAttribPointer(normalLoc, 3, GL_FLOAT, 0, 0, 0);
                }
                GL_CHECK_ERROR();
                // buffer for vertex texture coordinates
                if (mesh->HasTextureCoords(0)) {
                        float *texCoords = (float *)malloc(sizeof(float)*2*mesh->mNumVertices);
                        for (unsigned int k = 0; k < mesh->mNumVertices; ++k) {

                                texCoords[k*2]   = mesh->mTextureCoords[0][k].x;
                                texCoords[k*2+1] = mesh->mTextureCoords[0][k].y;

                        }
                        glGenBuffers(1, &buffer);
                        glBindBuffer(GL_ARRAY_BUFFER, buffer);
                        glBufferData(GL_ARRAY_BUFFER, sizeof(float)*2*mesh->mNumVertices, texCoords, GL_STATIC_DRAW);
                        glEnableVertexAttribArray(texCoordLoc);
                        glVertexAttribPointer(texCoordLoc, 2, GL_FLOAT, 0, 0, 0);
                }
                GL_CHECK_ERROR();

                GL_CHECK_ERROR();
                // unbind buffers
                glBindVertexArray(0);
                glBindBuffer(GL_ARRAY_BUFFER,0);
                glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
                GL_CHECK_ERROR();
                // create material uniform buffer
                aiMaterial *mtl = mScene->mMaterials[mesh->mMaterialIndex];

                // aiString texPath; //contains filename of texture
                // if(AI_SUCCESS == mtl->GetTexture(aiTextureType_DIFFUSE, 0, &texPath)) {
                //     //bind texture
                //     unsigned int texId = *mTextureIdMap[texPath.data];
                //     aMesh.texIndex = texId;
                //     aMat.texCount = 1;
                // } else { // we do not need textures, so leave that out!
                    aMat.texCount = 0;
                // }




                float c[4];
                set_float4(c, 0.8f, 0.8f, 0.8f, 1.0f);
                aiColor4D diffuse;
                if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_DIFFUSE, &diffuse))
                        color4_to_float4(&diffuse, c);
                memcpy(aMat.diffuse, c, sizeof(c));

                set_float4(c, 0.2f, 0.2f, 0.2f, 1.0f);
                aiColor4D ambient;
                if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_AMBIENT, &ambient))
                        color4_to_float4(&ambient, c);
                memcpy(aMat.ambient, c, sizeof(c));

                set_float4(c, 0.0f, 0.0f, 0.0f, 1.0f);
                aiColor4D specular;
                if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_SPECULAR, &specular))
                        color4_to_float4(&specular, c);
                memcpy(aMat.specular, c, sizeof(c));

                set_float4(c, 0.0f, 0.0f, 0.0f, 1.0f);
                aiColor4D emission;
                if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_EMISSIVE, &emission))
                        color4_to_float4(&emission, c);
                memcpy(aMat.emissive, c, sizeof(c));

                float shininess = 0.0;
                unsigned int max;
                aiGetMaterialFloatArray(mtl, AI_MATKEY_SHININESS, &shininess, &max);
                aMat.shininess = shininess;

                glGenBuffers(1,&(aMesh.uniformBlockIndex));
                glBindBuffer(GL_UNIFORM_BUFFER,aMesh.uniformBlockIndex);
                glBufferData(GL_UNIFORM_BUFFER, sizeof(aMat), (void *)(&aMat), GL_STATIC_DRAW);

                myMeshes.push_back(aMesh);
        }
}


// Can't send color down as a pointer to aiColor4D because AI colors are ABGR.
void Color4f(const aiColor4D *color) {
        glColor4f(color->r, color->g, color->b, color->a);
}

void set_float4(float f[4], float a, float b, float c, float d) {
        f[0] = a;
        f[1] = b;
        f[2] = c;
        f[3] = d;
}

void color4_to_float4(const aiColor4D *c, float f[4]) {
        f[0] = c->r;
        f[1] = c->g;
        f[2] = c->b;
        f[3] = c->a;
}

public:

uint64_t getNbrOfFaces() {
        return mNbrOfFaces;
}

gutil::BBox getScenesBBox() {
        return mSceneBBox;
}

gutil::BBox getUniformBox() {
    return mUniformBBox;
}

float getLargestExtend() {
        return mLargestExtend;
}

float getLargestExtendRecip() {
        return mRecipLargestExtend;
}


};



#endif
