#ifndef _SUB_DATA_IO_H_
#define _SUB_DATA_IO_H_

#include <vector>
#include <io/details/file_sub_data.h>
#include <bitset>

namespace io {


  void writeSubData(std::string path, uint32_t* sub_data, unsigned int primitive_count, const std::vector<Cell> &cellData,
                  uint width, uint height, uint max_sub, const gutil::BBox &scene_bbox, const gutil::BBox &uniform_bbox) {

        FILE * fdata=NULL;
        fdata = fopen (path.c_str(), "wb");
        if (!fdata) {
                std::cerr << "ERROR: Couldn't open file '" << path << "' for writing" << std::endl;
                exit(1);
        }


        fwrite((const char*) &io::data::magic, sizeof(uint32_t), 1, fdata);

        io::data::HEADER hdr;
        hdr.width = width;
        hdr.height = height;
        hdr.max_sub = max_sub;
        hdr.sub_table_size=primitive_count;
        hdr.grid_data_size=cellData.size();

        hdr.scene_bb_min_x = scene_bbox.min.x;
        hdr.scene_bb_min_y = scene_bbox.min.y;
        hdr.scene_bb_min_z = scene_bbox.min.z;

        hdr.scene_bb_max_x = scene_bbox.max.x;
        hdr.scene_bb_max_y = scene_bbox.max.y;
        hdr.scene_bb_max_z = scene_bbox.max.z;


        hdr.uniform_bb_min_x = uniform_bbox.min.x;
        hdr.uniform_bb_min_y = uniform_bbox.min.y;
        hdr.uniform_bb_min_z = uniform_bbox.min.z;

        hdr.uniform_bb_max_x = uniform_bbox.max.x;
        hdr.uniform_bb_max_y = uniform_bbox.max.y;
        hdr.uniform_bb_max_z = uniform_bbox.max.z;


        fwrite((const char*) &hdr, sizeof(io::data::HEADER), 1, fdata);


        //Write sub_data
        fwrite((const char*) sub_data, sizeof(unsigned int) * primitive_count, 1, fdata);

        //Write cell data
        fwrite((const char*) cellData.data(), sizeof(io::data::CellData) * cellData.size(), 1, fdata);

        fclose(fdata);
}



  void writeHashDataGPU(std::string path, uint32_t* sub_data, unsigned int primitive_count, unsigned int *cells, size_t no_cells,
                    uint width, uint height, uint min_sub, uint max_sub, const gutil::BBox &scene_bbox, const gutil::BBox &uniform_bbox) {
        FILE * fdata=NULL;
        fdata = fopen (path.c_str(), "wb");
        if (!fdata) {
                std::cerr << "ERROR: Couldn't open file '" << path << "' for writing" << std::endl;
                exit(1);
        }


        uint buildRes = powf(2, max_sub);

        fwrite((const char*) &io::data::magic, sizeof(uint32_t), 1, fdata);

        long int f_header_pos = ftell (fdata);
        io::data::HEADER hdr;
        fwrite((const char*) &hdr, sizeof(io::data::HEADER), 1, fdata);

        //Write sub_data
        fwrite((const char*) sub_data, sizeof(unsigned int) * primitive_count, 1, fdata);

        size_t level_cell_blocks = 0;

        size_t active_cells=0;

        for(auto i=MIN_BASE_LEVEL; i <= MAX_BASE_LEVEL; ++i) {
          auto per_level = pow(2,i);

          std::cout << "Processing level: " << i << std::endl;

          //32bit = 4*4*2
          unsigned int blocks_in_x = per_level / 4;
          unsigned int blocks_in_y = per_level / 4;
          unsigned int blocks_in_z = per_level / 2;

          for(auto j=0; j < blocks_in_x * blocks_in_y * blocks_in_z; ++j) {
            unsigned int current_cell = cells[level_cell_blocks + j];
            if (current_cell == 0)  continue;
            //std::cout << "Processing block: " << j << std::endl;

            //active_cells = active_cells + __builtin_popcount(current_cell);

            unsigned int x,y,z;
            unsigned int idx = j;

            // std::cout << "idx, x, y, z: " << idx;
            z = idx / (blocks_in_x * blocks_in_y);
            idx -= (z * (blocks_in_x * blocks_in_y));
            y = idx / blocks_in_x;
            x = idx % blocks_in_x;

            x *= 4;
            y *= 4;
            z *= 2;

            // std::cout << " " << x << ", " << y << ", " << z << std::endl;
            //std::cout << "Processing" << std::endl;
            //process bit offsets
            for(int zb = 0; zb < 2; ++zb) {
              for(int yb = 0; yb < 4; ++yb) {
                for(int xb = 0; xb < 4; ++xb) {

                  //std::bitset<32> db_bit(current_cell);

                  if(current_cell & 0x1) {
                    //std::cout << "D: " << db_bit << std::endl;
                    Cell c(x+xb, y+yb, z+zb, i);
                    // std::cout << "Writing cell " << (x + xb) << ", " << (y + yb) << ", " << (z + zb) << std::endl;

                    assert(x+xb < buildRes && y+yb < buildRes && z + zb < buildRes);

                    // std::cout << "(" << (x + xb) << ", " << (y + yb) << ", " << (z + zb) << ")" << std::endl;

                    fwrite((const char*) &c, sizeof(Cell), 1, fdata);
                    active_cells++;
                  }
                  current_cell = current_cell>>1;
                }
              }
            }

          }

          level_cell_blocks += blocks_in_x * blocks_in_y * blocks_in_z;
        }

        hdr.width = width;
        hdr.height = height;
        hdr.max_sub = max_sub;
        hdr.sub_table_size=primitive_count;
        hdr.grid_data_size=active_cells;
        std::cout << "IO Active Cells: " << active_cells << std::endl;

        hdr.scene_bb_min_x = scene_bbox.min.x;
        hdr.scene_bb_min_y = scene_bbox.min.y;
        hdr.scene_bb_min_z = scene_bbox.min.z;

        hdr.scene_bb_max_x = scene_bbox.max.x;
        hdr.scene_bb_max_y = scene_bbox.max.y;
        hdr.scene_bb_max_z = scene_bbox.max.z;


        hdr.uniform_bb_min_x = uniform_bbox.min.x;
        hdr.uniform_bb_min_y = uniform_bbox.min.y;
        hdr.uniform_bb_min_z = uniform_bbox.min.z;

        hdr.uniform_bb_max_x = uniform_bbox.max.x;
        hdr.uniform_bb_max_y = uniform_bbox.max.y;
        hdr.uniform_bb_max_z = uniform_bbox.max.z;


        std::cout << hdr.width << ", " << hdr.height << ", " << hdr.max_sub << ", " << hdr.sub_table_size << ", " << hdr.grid_data_size << std::endl;

        fseek (fdata , f_header_pos , SEEK_SET );
        fwrite((const char*) &hdr, sizeof(io::data::HEADER), 1, fdata);
        fseek (fdata , 0 , SEEK_END );
  }

void writeHashData(std::string path, uint32_t* sub_data, unsigned int primitive_count, const std::vector<Cell> &cellData,
                  uint width, uint height, uint max_sub, const gutil::BBox &scene_bbox, const gutil::BBox &uniform_bbox) {

        FILE * fdata=NULL;
        fdata = fopen (path.c_str(), "wb");
        if (!fdata) {
                std::cerr << "ERROR: Couldn't open file '" << path << "' for writing" << std::endl;
                exit(1);
        }

        fwrite((const char*) &io::data::magic, sizeof(uint32_t), 1, fdata);

        io::data::HEADER hdr;
        hdr.width = width;
        hdr.height = height;
        hdr.max_sub = max_sub;
        hdr.sub_table_size=primitive_count;
        hdr.grid_data_size=cellData.size();

        hdr.scene_bb_min_x = scene_bbox.min.x;
        hdr.scene_bb_min_y = scene_bbox.min.y;
        hdr.scene_bb_min_z = scene_bbox.min.z;

        hdr.scene_bb_max_x = scene_bbox.max.x;
        hdr.scene_bb_max_y = scene_bbox.max.y;
        hdr.scene_bb_max_z = scene_bbox.max.z;


        hdr.uniform_bb_min_x = uniform_bbox.min.x;
        hdr.uniform_bb_min_y = uniform_bbox.min.y;
        hdr.uniform_bb_min_z = uniform_bbox.min.z;

        hdr.uniform_bb_max_x = uniform_bbox.max.x;
        hdr.uniform_bb_max_y = uniform_bbox.max.y;
        hdr.uniform_bb_max_z = uniform_bbox.max.z;


        fwrite((const char*) &hdr, sizeof(io::data::HEADER), 1, fdata);

        //Write sub_data
        fwrite((const char*) sub_data, sizeof(unsigned int) * primitive_count, 1, fdata);

        //Write cell data
        fwrite((const char*) cellData.data(), sizeof(io::data::CellData) * cellData.size(), 1, fdata);

        fclose(fdata);
}

template <class tBBox>
void readSubData(std::string path, std::vector<uint32_t> &sub_data, std::vector<io::data::CellData> &grid_data,
                 uint &width, uint &height, uint &max_sub, tBBox &scene_bbox, tBBox &uniform_bbox) {

        FILE * fdata=NULL;
        fdata = fopen (path.c_str(), "rb");
        if (!fdata) {
                std::cerr << "ERROR: Couldn't open BVH file '" << path << "' for writing" << std::endl;
                exit(1);
        }

        //Read Magic Number
         uint magic=0;
         fread(&magic, sizeof(uint), 1, fdata);

         if(magic != io::data::magic) {
             fclose(fdata);
             std::cerr << "Given file not a valid sub file" << std::endl;
         }

         io::data::HEADER hdr;
         fread(&hdr, sizeof(io::data::HEADER), 1, fdata);

         width = hdr.width;
         height = hdr.height;
         max_sub = hdr.max_sub;

         scene_bbox=tBBox({hdr.scene_bb_min_x, hdr.scene_bb_min_y, hdr.scene_bb_min_z},
                          {hdr.scene_bb_max_x, hdr.scene_bb_max_y, hdr.scene_bb_max_z});

         uniform_bbox=tBBox({hdr.uniform_bb_min_x, hdr.uniform_bb_min_y, hdr.uniform_bb_min_z},
                            {hdr.uniform_bb_max_x, hdr.uniform_bb_max_y, hdr.uniform_bb_max_z});


         sub_data.resize(hdr.sub_table_size);
         fread(sub_data.data(), sizeof(uint32_t) * hdr.sub_table_size, 1, fdata);


         grid_data.resize(hdr.grid_data_size);
         fread(grid_data.data(), sizeof(io::data::CellData) * hdr.grid_data_size, 1, fdata);


        fclose(fdata);

}

}

#endif
