#ifndef _ASSIMP_DATA_VBO_RENDERER_HPP_
#define _ASSIMP_DATA_VBO_RENDERER_HPP_

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtx/transform.hpp>
#include <glm/gtc/type_ptr.hpp>

#include <io/AssimpVBOScene.hpp>
#include <utils/glShaderManager.h>
//#include <gutil/datatypes/Matrix4f.h>

class AssimpDataVBORenderer
{
protected:

AssimpVBOScene &mAssimpScene;

// For push and pop matrix
std::vector<glm::mat4*> matrixStack;

// Model Matrix (part of the OpenGL Model View Matrix)
//float modelMatrix[16];
glm::mat4 modelMatrix;

// Uniform Bindings Points
GLuint matricesUniLoc;
// GLuint materialUniLoc;

GLuint sub_table;
uint* sub_table_data;

// Uniform Buffer for Matrices
// this buffer will contain 3 matrices: projection, view and model
// each matrix is a float array with 16 components
GLuint matricesUniBuffer;
    #define MatricesUniBufferSize sizeof(float) * 16 * 3
    #define ProjMatrixOffset 0
    #define ViewMatrixOffset sizeof(float) * 16
    #define ModelMatrixOffset sizeof(float) * 16 * 2
    #define MatrixSize sizeof(float) * 16


//The sampler uniform for textured models
//we are assuming a single texture so this will
//always be texture unit 0
GLuint texUnit;

//Shader Types
//Build Shader
//Inspect Shader
glShaderManager *mShader[2];
uint active_shader;


// Push and Pop for modelMatrix

void pushMatrix() {
        glm::mat4 *aux = new glm::mat4(modelMatrix);
        matrixStack.push_back(aux);
}

void popMatrix() {

        glm::mat4 *m = matrixStack[matrixStack.size()-1];
        modelMatrix = *m;
        matrixStack.pop_back();
        free(m);
}

void recursiveRender(const aiNode* nd, uint pFrom, uint pTo, size_t &pFacesProcessed) {
        // Get node transformation matrix
        aiMatrix4x4 m = nd->mTransformation;

        // OpenGL matrices are column major
        m.Transpose();


        // save model matrix and apply node transformation
        pushMatrix();

        //float aux[16];
        //memcpy(aux, &m,sizeof(float) * 16);
        glm::mat4 aux = glm::make_mat4((float*) &m);

        modelMatrix = modelMatrix * aux;

        setModelMatrix();


        // draw all meshes assigned to this node
        for (unsigned int n=0; n < nd->mNumMeshes; ++n) {
                size_t pFacesProcessedOld=pFacesProcessed;
                pFacesProcessed+=mAssimpScene.myMeshes[nd->mMeshes[n]].numFaces;
                if(pFacesProcessed>=pFrom) {
                        size_t toOffset;
                        if(pFacesProcessed<=pTo) {
                                toOffset = pTo-pFacesProcessedOld;
                                toOffset = std::min(toOffset,(size_t) mAssimpScene.myMeshes[nd->mMeshes[n]].numFaces);

                        } else {
                                continue;
                        }

                        GL_CHECK_ERROR();

                        glUniform1i(mShader[active_shader]->getUniformLocation("primitiveIDStartIdx"), pFacesProcessedOld);
                        GL_CHECK_ERROR();

                        long int fromOffset=pFrom-pFacesProcessedOld;
                        fromOffset = std::max((long int) 0,  fromOffset);

                        // bind material uniform
                        glBindBufferRange(GL_UNIFORM_BUFFER, mAssimpScene.materialUniLoc, mAssimpScene.myMeshes[nd->mMeshes[n]].uniformBlockIndex, 0, sizeof(struct AssimpMaterial));

                        // bind texture
                        glBindTexture(GL_TEXTURE_2D, mAssimpScene.myMeshes[nd->mMeshes[n]].texIndex);



                        // bind VAO
                        glBindVertexArray(mAssimpScene.myMeshes[nd->mMeshes[n]].vao);

                        glDrawRangeElements(GL_TRIANGLES,  fromOffset,  toOffset,  (toOffset-fromOffset)*3,  GL_UNSIGNED_INT,0);


                        glBindTexture(GL_TEXTURE_2D,0);
                        glBindVertexArray(0);
                        glBindBuffer(GL_UNIFORM_BUFFER,0);
                }

        }

        // draw all children
        for (unsigned int n=0; n < nd->mNumChildren; ++n) {
                recursiveRender(nd->mChildren[n], pFrom, pTo, pFacesProcessed);
        }
        popMatrix();
}

void loadDefaultShader() {
        //if(mShader!=NULL)
        //        delete mShader;

        mShader[0] = new glShaderManager();

        mShader[0]->loadShaderFromFile("../shader/subcalc/subcalc_vert.glsl", glShaderManager::VERTEX_SHADER);
        mShader[0]->loadShaderFromFile("../shader/subcalc/subcalc_geom.glsl", glShaderManager::GEOMETRY_SHADER);
        mShader[0]->loadShaderFromFile("../shader/subcalc/subcalc_frag.glsl", glShaderManager::FRAGMENT_SHADER);


        GL_CHECK_ERROR();
        if(!mShader[0]->compile(true)) {
                std::cerr << "ERROR: Couldn't initialize shader" << std::endl;
                exit(1);
        }

        glBindFragDataLocation(mShader[0]->getShaderProgram(), 0, "resultCol");
        glBindAttribLocation(mShader[0]->getShaderProgram(),mAssimpScene.vertexLoc,"position");
        glBindAttribLocation(mShader[0]->getShaderProgram(),mAssimpScene.normalLoc,"normal");
        glBindAttribLocation(mShader[0]->getShaderProgram(),mAssimpScene.texCoordLoc,"texCoord");

        mShader[0]->link();
        mShader[0]->validate();
//
//
        mShader[1] = new glShaderManager();
        mShader[1]->loadShaderFromFile("../shader/inspect/inspect_vert.glsl", glShaderManager::VERTEX_SHADER);
        mShader[1]->loadShaderFromFile("../shader/inspect/inspect_geom.glsl", glShaderManager::GEOMETRY_SHADER);
        mShader[1]->loadShaderFromFile("../shader/inspect/inspect_frag.glsl", glShaderManager::FRAGMENT_SHADER);


        GL_CHECK_ERROR();
        if(!mShader[1]->compile(true)) {
                std::cerr << "ERROR: Couldn't initialize shader" << std::endl;
                exit(1);
        }

        glBindFragDataLocation(mShader[1]->getShaderProgram(), 0, "resultCol");
        glBindAttribLocation(mShader[1]->getShaderProgram(),mAssimpScene.vertexLoc,"position");
        glBindAttribLocation(mShader[1]->getShaderProgram(),mAssimpScene.normalLoc,"normal");
        glBindAttribLocation(mShader[1]->getShaderProgram(),mAssimpScene.texCoordLoc,"texCoord");

        mShader[1]->link();
        mShader[1]->validate();

        GL_CHECK_ERROR();
}

void genBuffers() {
        glGenBuffers(1,&matricesUniBuffer);
}

public:
AssimpDataVBORenderer(AssimpVBOScene &pAssimpReader) : mAssimpScene(pAssimpReader), matricesUniLoc(1), texUnit(0), active_shader(0)  {
        mShader[0] = NULL;
        mShader[1] = NULL;
        loadDefaultShader();
        genBuffers();
        initializeShader();
}
/*
   AssimpDataVBORenderer(AssimpVBOScene &pAssimpReader, glShaderManager *pShader) : mAssimpScene(pAssimpReader), mShader(pShader), matricesUniLoc(1), texUnit(00), active_shader(0)  {
        genBuffers();
        initializeShader();
   }
 */

virtual ~AssimpDataVBORenderer() {

        //if(mShader)
        //        delete mShader;

        // delete buffers
        glDeleteBuffers(1,&matricesUniBuffer);

}


void setActiveShader(uint a) {
    active_shader = a;
}


void initializeShaderStorageBuffer(GLuint &bufferID, GLsizeiptr size, void* data, GLenum usage = GL_DYNAMIC_DRAW){
        glGenBuffers(1, &(bufferID));
        GL_CHECK_ERROR();
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, bufferID);
        GL_CHECK_ERROR();

        GL_CHECK_ERROR();;
        glBufferData(GL_SHADER_STORAGE_BUFFER, size, data, usage);
        GL_CHECK_ERROR();

}


void initializeShader() {

        for(uint i=0;i<2; i++) {
        mShader[i]->begin();
        GL_CHECK_ERROR();
        GLuint k = glGetUniformBlockIndex(mShader[i]->getShaderProgram(),"Matrices");
        GL_CHECK_ERROR();
        glUniformBlockBinding(mShader[i]->getShaderProgram(), k, matricesUniLoc);
        GL_CHECK_ERROR();
        k = glGetUniformBlockIndex(mShader[i]->getShaderProgram(),"Material");

        if(k == GL_INVALID_INDEX) {
                std::cerr << "Invalid index" << std::endl;
        }
        GL_CHECK_ERROR();
        glUniformBlockBinding(mShader[i]->getShaderProgram(), k, mAssimpScene.materialUniLoc);
        GL_CHECK_ERROR();
        mShader[i]->end();
        }


        //
        // Uniform Block
        //
        glBindBuffer(GL_UNIFORM_BUFFER, matricesUniBuffer);
        glBufferData(GL_UNIFORM_BUFFER, MatricesUniBufferSize,NULL,GL_DYNAMIC_DRAW);
        glBindBufferRange(GL_UNIFORM_BUFFER, matricesUniLoc, matricesUniBuffer, 0, MatricesUniBufferSize);
        glBindBuffer(GL_UNIFORM_BUFFER,0);


        mShader[0]->begin();

        std::cout << "Initialize SubTable" << std::endl;
        size_t subtable_size = mAssimpScene.getNbrOfFaces() * sizeof(uint);
        sub_table_data = (uint*) malloc(subtable_size);
        memset(sub_table_data, 0, subtable_size);

        initializeShaderStorageBuffer(sub_table, mAssimpScene.getNbrOfFaces() * sizeof(uint), sub_table_data);
        glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 12, sub_table);

        gutil::BBox sceneBB = mAssimpScene.getUnitUniformBBox();
        std::cout << "Uniform Bounding Box" << std::endl;
        std::cout << sceneBB.min.x << ", " << sceneBB.min.y << ", " << sceneBB.min.z << std::endl;
        std::cout << sceneBB.max.x << ", " << sceneBB.max.y << ", " << sceneBB.max.z << std::endl;
        glUniform3fv(mShader[0]->getUniformLocation("scene_bb_min"), 1, (float*) &sceneBB.min);
        glUniform3fv(mShader[0]->getUniformLocation("scene_bb_max"), 1, (float*) &sceneBB.max);
        GL_CHECK_ERROR();

        glUniform1ui(mShader[0]->getUniformLocation("min_base_level"), MIN_BASE_LEVEL);
        glUniform1ui(mShader[0]->getUniformLocation("max_base_level"), MAX_BASE_LEVEL);
        
        GL_CHECK_ERROR();

        glUniform1f(mShader[0]->getUniformLocation("scalef_frag"), scalef_frag);
        glUniform1f(mShader[0]->getUniformLocation("scalef_geom"), scalef_geom);
        GL_CHECK_ERROR();

        glUniform1i(mShader[0]->getUniformLocation("primitiveIDStartIdx"), 0);
        GL_CHECK_ERROR();
        mShader[0]->end();

}


void shaderBegin() {
        mShader[active_shader]->begin();
}

void shaderEnd() {
        mShader[active_shader]->end();
}


glShaderManager* getShader() {
        return mShader[active_shader];
}

void render(uint from, uint to) {

        // we are only going to use texture unit 0
        // unfortunately samplers can't reside in uniform blocks
        // so we have set this uniform separately
        texUnit = mShader[active_shader]->getUniformLocation("Texture0");
        glUniform1iARB(texUnit, 0);

        modelMatrix = glm::mat4(1.0);
        //glPushMatrix();
        /*    TransformationMatrix trans;
            trans.makeIdentity();


            trans.scale(gutil::Vec4f(mAssimpScene.getLargestExtendRecip(), mAssimpScene.getLargestExtendRecip(),mAssimpScene.getLargestExtendRecip()));


            trans.translate(-mAssimpScene.getScenesBBox().getMin());

            modelMatrix = trans;
         */
        size_t facesProcessed=0;
        recursiveRender(mAssimpScene.getScene()->mRootNode, from, to, facesProcessed);
        GL_CHECK_ERROR();

/*   glBindBuffer(GL_UNIFORM_BUFFER, matricesUniBuffer);

     float *bufferData = (float*) malloc(MatricesUniBufferSize);
     glGetBufferSubData(GL_UNIFORM_BUFFER, 0, MatricesUniBufferSize, (void*) bufferData);

     LOG_OUTL("Projection Matrix");
     for(int i=0;i<16;i++) {
        LOG_OUT(""<<std::dec << bufferData[i] << ", ");
     }
    LOG_OUTL("");

     LOG_OUTL("View Matrix");
     for(int i=16;i<32;i++) {
        LOG_OUT(""<< bufferData[i] << ", ");
     }
    LOG_OUTL("");

    LOG_OUTL("Model Matrix");
    for(int i=32;i<48;i++) {
      LOG_OUT(""<< bufferData[i] << ", ");
    }
    LOG_OUTL("");

    glBindBuffer(GL_UNIFORM_BUFFER, 0);
 */
}



void render() {

        render(0, mAssimpScene.getNbrOfFaces());
}


const GLint getGLSubTable() const {
        return sub_table;
}

AssimpVBOScene& getScene() {
        return mAssimpScene;
}

void setModelMatrix() {

        glBindBuffer(GL_UNIFORM_BUFFER, matricesUniBuffer);

        glBufferSubData(GL_UNIFORM_BUFFER,
                        ModelMatrixOffset, MatrixSize, &(modelMatrix));

        glBindBuffer(GL_UNIFORM_BUFFER,0);
}


void setProjectionMatrix(const glm::mat4 &pProjMatrix) {
        glBindBuffer(GL_UNIFORM_BUFFER,matricesUniBuffer);

        glBufferSubData(GL_UNIFORM_BUFFER, ProjMatrixOffset, MatrixSize, &(pProjMatrix));
        GL_CHECK_ERROR();
        //glCheckError();
        glBindBuffer(GL_UNIFORM_BUFFER,0);


}

void setViewMatrix(const glm::mat4 &pViewMatrix) {
        glBindBuffer(GL_UNIFORM_BUFFER, matricesUniBuffer);

        glBufferSubData(GL_UNIFORM_BUFFER, ViewMatrixOffset, MatrixSize, &(pViewMatrix));
        GL_CHECK_ERROR();

        glBindBuffer(GL_UNIFORM_BUFFER,0);
}


};

#endif
