// FIFAR: File Interpolator For Animation Rendering

#pragma once


#include <glm/glm.hpp>

// #include "VecMath.h"

#include <vector>
#include <iostream>
#include <fstream>

#include <string>
#include <sstream>
#include <utility>


// using namespace trd;
using namespace std;
using namespace glm;


namespace trd {


	class FrameDescriptor {
	public:
		vec3 pos;
		vec3 dir;
		vec3 up;
		float distanceToNext;
		float frameShare;
		int numFrames;
	};

	class Fifar {
	private:

		std::vector<std::string> explode(std::string const & s, char delim) {
			std::vector<std::string> result;
			std::istringstream iss(s);

			for (std::string token; std::getline(iss, token, delim);)
			{
				result.push_back(std::move(token));
			}

			return result;
		}

		vector<FrameDescriptor> __m_keyFrames;
		vector<FrameDescriptor> __m_frames;
		int __m_numFrames = 0;

	public:


		Fifar() {

		}


		void addKeyFrame(FrameDescriptor p_kf) {
			p_kf.pos = p_kf.pos;
			normalize(p_kf.dir);
			normalize(p_kf.up);


			__m_keyFrames.push_back(p_kf);
		}


		FrameDescriptor getKeyFrame(int p_index) {
			return __m_keyFrames[p_index];
		}


		FrameDescriptor getFrame(int p_index) {
			return __m_frames[p_index];
		}


		int getNumKeyFrames() {
			return __m_keyFrames.size();
		}

		int getNumFrames() {
			return __m_frames.size();
		}

		void computeFrameDescriptors(int p_numFrames) {
			__m_numFrames = p_numFrames;

			p_numFrames = p_numFrames - 1;
			__m_frames.reserve(p_numFrames);

			float totalDistance = 0.f;

			// first compute distances between keyframes
			for (int i = 0; i < __m_keyFrames.size() - 1; ++i) {
				vec3 posDist = __m_keyFrames[i].pos - __m_keyFrames[i + 1].pos;
				float dist = sqrtf(posDist[0] * posDist[0] + posDist[1] * posDist[1] + posDist[2] * posDist[2]);
				__m_keyFrames[i].distanceToNext = dist;
				totalDistance += dist;

				std::cout << "distance from " << i << " to " << (i + 1) << ": " << dist << std::endl;
			}






			// normalize distances
			for (int i = 0; i < __m_keyFrames.size() - 1; ++i) {
				__m_keyFrames[i].frameShare = __m_keyFrames[i].distanceToNext / totalDistance;
				__m_keyFrames[i].numFrames = roundf(__m_keyFrames[i].frameShare * p_numFrames);
				std::cout << "Frame share from " << i << " to " << (i + 1) << ": " << __m_keyFrames[i].frameShare << " (" << __m_keyFrames[i].numFrames << ")" << std::endl;
			}

			// ensure that the last frame is actually rendered
			__m_keyFrames[__m_keyFrames.size() - 1].numFrames = 1;


			/*** optimize numFrames to fit requested number of frames ***/
			// first compute sum of frames
			int numFramesTotal = 0;
			for (int i = 0; i < __m_keyFrames.size() - 1; ++i) {
				numFramesTotal += __m_keyFrames[i].numFrames;
			}

			do {


				if (numFramesTotal > p_numFrames) {
					std::cout << "Too many frames; ";
					int maxNumFrames = 0;
					int maxIndex = 0;
					// find keyframe with max frames
					for (int i = 0; i < __m_keyFrames.size() - 1; ++i) {
						if (__m_keyFrames[i].numFrames > maxNumFrames) {
							maxNumFrames = fmaxf(__m_keyFrames[i].numFrames, maxNumFrames);
							maxIndex = i;
						}
					}
					std::cout << "Max number of frames: kf " << maxIndex << " (" << maxNumFrames << ") => sub 1" << std::endl;
					--(__m_keyFrames[maxIndex].numFrames);
					--numFramesTotal;


				}
				else if (numFramesTotal < p_numFrames) {
					std::cout << "Not enough frames; ";
					int minNumFrames = 0;
					int minIndex = 0;
					// find keyframe with max frames
					for (int i = 0; i < __m_keyFrames.size() - 1; ++i) {
						if (__m_keyFrames[i].numFrames > minNumFrames) {
							minNumFrames = fmaxf(__m_keyFrames[i].numFrames, minNumFrames);
							minIndex = i;
						}
					}
					std::cout << "Min number of frames: kf " << minIndex << " (" << minNumFrames << ") => add 1" << std::endl;
					++(__m_keyFrames[minIndex].numFrames);
					++numFramesTotal;
				}

			} while (numFramesTotal != p_numFrames);





			// generate frame descriptors
			int curFrame = 0;

			for (int curKeyFrame = 0; curKeyFrame < __m_keyFrames.size() - 1; ++curKeyFrame) {

				int numFrames = __m_keyFrames[curKeyFrame].numFrames;
				FrameDescriptor startDescriptor = __m_keyFrames[curKeyFrame];
				FrameDescriptor endDescriptor = __m_keyFrames[curKeyFrame + 1];

				int progress = 0;

				// std::cout << "curKeyFrame: " << curKeyFrame << std::endl;
				// std::cout << "numFrames: " << numFrames << std::endl;

				int upperLimit = curFrame + numFrames;

				for (curFrame; curFrame < upperLimit; ++curFrame) {

					// std::cout << curFrame << std::endl;

					float t = (float)progress / (float)numFrames;
					FrameDescriptor curDescriptor;
					// std::cout << "startDescriptor.pos: " << startDescriptor.pos[0] << ", " << startDescriptor.pos[1] << ", " << startDescriptor.pos[2] << std::endl;
					// std::cout << "endDescriptor.pos: " << endDescriptor.pos[0] << ", " << endDescriptor.pos[1] << ", " << endDescriptor.pos[2] << std::endl;
					curDescriptor.pos = (1 - t) * startDescriptor.pos + t * endDescriptor.pos;
					curDescriptor.dir = (1 - t) * startDescriptor.dir + t * endDescriptor.dir;
					curDescriptor.up = (1 - t) * startDescriptor.up + t * endDescriptor.up;

					normalize(curDescriptor.dir);
					normalize(curDescriptor.up);

					// std::cout << "Frame " << curFrame << " pos: " << curDescriptor.pos[0] << ", " << curDescriptor.pos[1] << ", " << curDescriptor.pos[2] << std::endl;
					// std::cout << "Frame " << curFrame << " dir: " << curDescriptor.dir[0] << ", " << curDescriptor.dir[1] << ", " << curDescriptor.dir[2] << std::endl;
					// std::cout << "Frame " << curFrame << " up: " << curDescriptor.up[0] << ", " << curDescriptor.up[1] << ", " << curDescriptor.up[2] << std::endl;

					__m_frames.push_back(curDescriptor);

					++progress;

				}

				// std::cout << curKeyFrame << std::endl;

			}


		}


		void reset() {
			__m_keyFrames.clear();
			__m_frames.clear();
			__m_numFrames = 0;
		}


		/**
		* file format:
		* - number of keyframes
		* - n frameDescriptors
		*/
		void writeKeyFrameDescriptorsToFile(const char* p_file) {
			ofstream file;
			file.open(p_file);
			file << __m_keyFrames.size() << std::endl;

			// write
			for (int i = 0; i < __m_keyFrames.size(); ++i) {

				//pos
				file << __m_keyFrames[i].pos[0] << " " << __m_keyFrames[i].pos[1] << " " << __m_keyFrames[i].pos[2] << std::endl;
				//dir
				file << __m_keyFrames[i].dir[0] << " " << __m_keyFrames[i].dir[1] << " " << __m_keyFrames[i].dir[2] << std::endl;
				//up
				file << __m_keyFrames[i].up[0] << " " << __m_keyFrames[i].up[1] << " " << __m_keyFrames[i].up[2] << std::endl;

			}

			file.close();
		}


		// void writeAllFrameDescriptorsToFile(string& p_file) {

		// }


		void readFromFile(const char* p_file) {
			reset();

			ifstream file;
			file.open(p_file);

			string line;
			getline(file, line);
			int numKeyFrames = atof(line.c_str());

			for (int i = 0; i < numKeyFrames; ++i) {

				FrameDescriptor fd;

				//pos
				getline(file, line);
				vector<string> expString = explode(line, ' ');
				fd.pos[0] = atof(expString[0].c_str());
				fd.pos[1] = atof(expString[1].c_str());
				fd.pos[2] = atof(expString[2].c_str());

				//dir
				getline(file, line);
				expString = explode(line, ' ');
				fd.dir[0] = atof(expString[0].c_str());
				fd.dir[1] = atof(expString[1].c_str());
				fd.dir[2] = atof(expString[2].c_str());


				//up
				getline(file, line);
				expString = explode(line, ' ');
				fd.up[0] = atof(expString[0].c_str());
				fd.up[1] = atof(expString[1].c_str());
				fd.up[2] = atof(expString[2].c_str());

				addKeyFrame(fd);

			}
			file.close();
		}





	};

}
