#ifndef _CUDA_ERROR_HANDLER_H_
#define _CUDA_ERROR_HANDLER_H_

#include <cuda_runtime_api.h>
#include <cuda.h>
#include <iostream>

#ifdef CU_CHECK_ERRORS
#define CU_CHECK_ERROR(FUNC_NAME, CU_RES) checkCUError(__LINE__, __FILE__, FUNC_NAME, CU_RES)
#else
#define CU_CHECK_ERROR(FUNC_NAME, CU_RES)
#endif

#ifdef CUDA_CHECK_ERRORS
#define CUDA_CHECK_ERROR(FUNC_NAME, CUDA_ERROR) checkCUDAError(__LINE__, __FILE__, FUNC_NAME, CUDA_ERROR)
#define CUDA_LAST_ERROR()	lastCUDAError(__LINE__, __FILE__)
#else
#define CUDA_CHECK_ERROR(FUNC_NAME, CUDA_ERROR)
#define CUDA_LAST_ERROR()
#endif



inline const char* decodeCUError(CUresult res)
{
    const char* error;
    cuGetErrorName(res, &error);
//    cuGetErrorString(res, &error);
    return error;
}

inline const char* decodeCUDAError(cudaError_t cerror)
{
    const char* error = cudaGetErrorString(cerror);
//    cuGetErrorString(res, &error);
    return error;
}

inline void checkCUError(int line, const char* file, const char* funcName, CUresult res)
{
    if (res != CUDA_SUCCESS)
        std::cerr << "<" << file << ": " << std::dec << line << ">" << " cuError: (" << funcName << "() failed, " << decodeCUError(res) << std::endl;
}


inline void checkCUDAError(int line, const char* file, const char* funcName, cudaError_t error)
{
    if (error != cudaSuccess)
        std::cerr << "<" << file << ": " << std::dec << line << ">" << " cudaError: (" << funcName << "() failed, " << decodeCUDAError(error) << std::endl;
}

inline void lastCUDAError(int line, const char* file)
{
    cudaError_t error = cudaGetLastError ();
    if (error != cudaSuccess)
        std::cerr << "<" << file << ": " << std::dec << line << ">" << " Cuda failed: " << decodeCUDAError(error) << std::endl;

}

#endif

