#version 450

layout (std140) uniform Material {
    vec4 diffuse;
    vec4 ambient;
    vec4 specular;
    vec4 emissive;
    float shininess;
    int texCount;
};

uniform	sampler2D Texture0;

in fData {
  vec4 vertexPosWS;
  vec2 TexCoord;
  vec3 Normal;
  float min_depth;
  flat uint triId;
  flat uint base_level;
} frag;

layout(std430, binding=12) buffer SubTable {
    uint sub_table[];
};


uniform float scalef_frag;

uniform vec3 scene_bb_min;
uniform vec3 scene_bb_max;

uniform uint min_base_level;
uniform uint max_base_level;

out vec4 resultCol;

layout(binding = 4) uniform atomic_uint dbg_count;

void main()
{
    vec4 color;
    vec4 amb;
    float intensity;
    vec3 lightDir;
    vec3 n;

    lightDir = normalize(vec3(1.0,1.0,1.0));
    n = normalize(frag.Normal);
    intensity = max(dot(lightDir,n),0.0);

    if (texCount == 0) {
        color = diffuse;
        amb = ambient;
    } else {
        color = texture(Texture0, frag.TexCoord);
        amb = color * 0.33;
    }

    //resultCol = (color * intensity) + amb;
    //float fac = 1.f/20.f;
    //resultCol = vec4(frag.min_depth * fac, frag.min_depth * fac, frag.min_depth * fac, 1);

    float FLT_MAX = 1.0 / 0.0;
    vec3 scene_bb_extend = scene_bb_max - scene_bb_min;

    vec3 dx = dFdx(frag.vertexPosWS).xyz;
    vec3 dy = dFdy(frag.vertexPosWS).xyz;

/*
    vec4 px_fp_min = vec4(FLT_MAX, FLT_MAX, FLT_MAX, FLT_MAX);
    vec4 px_fp_max = vec4(-FLT_MAX, -FLT_MAX, -FLT_MAX, -FLT_MAX);

    px_fp_min = min(px_fp_min, dx);
    px_fp_max = max(px_fp_max, dx);

    px_fp_min = min(px_fp_min, dy);
    px_fp_max = max(px_fp_max, dy);
*/

    float max_ws_d = min(dot(dx, dx), dot(dy,dy));
    float scene_bb_ext_max = max(scene_bb_extend.x, max(scene_bb_extend.y, scene_bb_extend.z));
    //float scene_bb_ext_max = length(scene_bb_extend);
    float max_level_fac = abs(max_ws_d / scene_bb_ext_max);

    //vec3 bb_level_facs = abs(px_fp_min.xyz) / scene_bb_extend;
    //float max_level_fac = min(bb_level_facs.x, min(bb_level_facs.y, bb_level_facs.z));
    uint base_level = uint(floor(abs(log2(max_level_fac)* scalef_frag)));
    base_level = clamp(base_level, min_base_level, max_base_level);

    uint out_data = atomicMax(sub_table[frag.triId], base_level);

    float fac = 1.0/float(max_base_level);
    //resultCol = vec4(out_data * fac, out_data * fac, out_data * fac, 1);
    resultCol = color;

    atomicCounterIncrement(dbg_count);

}
