//GEOMETRY SHADER
#version 450

layout (triangles) in;
layout (triangle_strip, max_vertices=3) out;
//layout (line_strip, max_vertices=9) out;

layout (std140) uniform Matrices {
    mat4 projMatrix;
    mat4 viewMatrix;
    mat4 modelMatrix;
};


layout(std430, binding=12) buffer SubTable {
    uint sub_table[];
};


in vData {
  vec4 vertexPos;
  vec2 TexCoord;
  vec3 Normal;
} vertices[];

out fData {
  vec4 vertexPosWS;
  vec2 TexCoord;
  vec3 Normal;
  float min_depth;
  flat uint triId;
  flat uint base_level;
} frag;

uniform float scalef_geom;

uniform vec3 scene_bb_min;
uniform vec3 scene_bb_max;

uniform uint min_base_level;
uniform uint max_base_level;

uniform int primitiveIDStartIdx;

uniform int fixedSubLevel;

//After calling ftTransform in the vertex shader the positions in gl_Position are in eyespace/clipspace
void main()
{
    float FLT_MAX = 1.0 / 0.0;



    //Compute the primitive ID for that triangle
    frag.triId = (primitiveIDStartIdx + gl_PrimitiveIDIn);

    //Compute BBox of Triangle
    vec3 tri_bb_min = vec3(FLT_MAX, FLT_MAX, FLT_MAX);
    vec3 tri_bb_max = vec3(-FLT_MAX, -FLT_MAX, -FLT_MAX);

/*
		vMVpda[0] = mv1 * vertices[0].mMV * gl_in[0].gl_Position;
		vMVpda[1] = mv1 * vertices[1].mMV * gl_in[1].gl_Position;
		vMVpda[2] = mv1 * vertices[2].mMV * gl_in[2].gl_Position;
*/
/*
    gl_Position = gl_in[0].gl_Position
    EmitVertex();
    gl_Position = gl_in[1].gl_Position
    EmitVertex();
    gl_Position = gl_in[2].gl_Position
    EmitVertex();
    EndPrimitive();
*/

  for(int i = 0;i < gl_in.length();i++)
  {
    tri_bb_min = min(tri_bb_min, gl_in[i].gl_Position.xyz);
    tri_bb_max = max(tri_bb_max, gl_in[i].gl_Position.xyz);
  }




  //Initial setup for vertices
  vec3 scene_bb_extend = scene_bb_max - scene_bb_min;

  //Alternative A: Length of bb extend
  //vec3 tri_bb_extend = tri_bb_max - tri_bb_min;
  //float tri_bb_l = length(tri_bb_extend);

  //Alternative B: Length of shortes edge of triangle
  float l1=length(gl_in[1].gl_Position.xyz - gl_in[0].gl_Position.xyz);
  float l2=length(gl_in[2].gl_Position.xyz - gl_in[0].gl_Position.xyz);
  float l3=length(gl_in[2].gl_Position.xyz - gl_in[1].gl_Position.xyz);
  float tri_bb_l = min(l1, min(l2,l3));


  //float max_scene_bb_ext = max(scene_bb_extend.x, max(scene_bb_extend.y, scene_bb_extend.z));
  float max_scene_bb_ext = length(scene_bb_extend);

  float max_level_fac = tri_bb_l / max_scene_bb_ext;
  //vec3 bb_level_facs = tri_bb_extend / max_scene_bb_ext;
  //float max_level_fac = min(bb_level_facs.x, min(bb_level_facs.y, bb_level_facs.z));
  frag.base_level = uint(floor(abs(log2(max_level_fac)) * scalef_geom));
  frag.base_level = clamp(frag.base_level, min_base_level, max_base_level);

  atomicMax(sub_table[frag.triId], frag.base_level);

  //atomicMax(sub_table[frag.triId], floatBitsToInt(scene_bb_min.x));


    /////////
    for(int i = 0;i < gl_in.length();i++)
    {
      frag.Normal = vertices[i].Normal;
      frag.TexCoord = vertices[i].TexCoord;
      frag.min_depth = min(abs(gl_in[i].gl_Position.z), frag.min_depth);
      frag.vertexPosWS = gl_in[i].gl_Position;

      //frag.color = vertices[i].color;
      gl_Position = projMatrix * viewMatrix *  gl_in[i].gl_Position;
      EmitVertex();
    }


    EndPrimitive();
}
